package cz.nic.tablexia;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.graphics.GL30;
import com.badlogic.gdx.scenes.scene2d.InputEvent;
import com.badlogic.gdx.scenes.scene2d.actions.Actions;
import com.badlogic.gdx.scenes.scene2d.utils.ClickListener;

import net.engio.mbassy.bus.error.IPublicationErrorHandler;
import net.engio.mbassy.bus.error.PublicationError;
import net.engio.mbassy.listener.Handler;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.ArrayList;
import java.util.Locale;
import java.util.Map;

import cz.nic.tablexia.bus.ApplicationBus;
import cz.nic.tablexia.bus.ApplicationBus.ApplicationEvent;
import cz.nic.tablexia.loader.application.ApplicationAtlasManager;
import cz.nic.tablexia.loader.application.ApplicationExternalSoundManager;
import cz.nic.tablexia.loader.application.ApplicationFontManager;
import cz.nic.tablexia.loader.application.ApplicationInternalSoundManager;
import cz.nic.tablexia.loader.application.ApplicationInternalTextureManager;
import cz.nic.tablexia.loader.application.ApplicationTextManager;
import cz.nic.tablexia.loader.zip.ZipAssetLoader;
import cz.nic.tablexia.menu.MenuController;
import cz.nic.tablexia.screen.AbstractTablexiaScreen;
import cz.nic.tablexia.screen.loader.IConnectionManager;
import cz.nic.tablexia.util.Log;
import cz.nic.tablexia.util.Utility;
import cz.nic.tablexia.util.ui.button.TablexiaButton;
import cz.nic.tablexia.util.ui.dialog.TablexiaComponentDialog;
import cz.nic.tablexia.util.ui.dialog.TablexiaComponentDialogFactory;
import cz.nic.tablexia.util.ui.dialog.components.AdaptiveSizeDialogComponent;
import cz.nic.tablexia.util.ui.dialog.components.BackButtonHideComponent;
import cz.nic.tablexia.util.ui.dialog.components.CenterPositionDialogComponent;
import cz.nic.tablexia.util.ui.dialog.components.DimmerDialogComponent;
import cz.nic.tablexia.util.ui.dialog.components.FixedSpaceContentDialogComponent;
import cz.nic.tablexia.util.ui.dialog.components.PositiveNegativeButtonContentDialogComponent;
import cz.nic.tablexia.util.ui.dialog.components.ResizableSpaceContentDialogComponent;
import cz.nic.tablexia.util.ui.dialog.components.TablexiaDialogComponentAdapter;
import cz.nic.tablexia.util.ui.dialog.components.TextContentDialogComponent;

/**
 *
 * @author Matyáš Latner
 */
public class Tablexia extends TablexiaApplication {

    public static final String  INTERNAL_GFX_ASSETS_PATH                = "gfx/";
    public static final String  ERROR_DIALOG_BACKGROUND_INTERNAL        = INTERNAL_GFX_ASSETS_PATH + "dialog_square_borderlines.9.png";
    public static final int     ERROR_DIALOG_WIDTH                      = 460;
    public static final int     ERROR_DIALOG_HEIGHT                     = 300;
    public static final float   ERROR_DIALOG_DIMMER                     = 0.7f;

    public static final int     BACK_BUTTON_SIZE                        = 50;
    public static final float   BACK_BUTTON_LEFT_OFFSET_RATIO           = 0.045f;
    public static final float   BACK_BUTTON_ADAPTIVE_SIZE_RATIO         = 1.7f;
    public static final float   BACK_BUTTON_BOTTOM_OFFSET               = 0.04f;
    public static final boolean BACK_BUTTON_ADAPTIVE_SIZE_POSITION_FIX  = true;
    public static final float   SHOW_BACK_BUTTON_DURATION               = 1f;
    public static final float   HIDE_BACK_BUTTON_DURATION               = 1f;

    private static final boolean LOADER_ERROR_DIALOG_HIDE_ANIMATION     = false;

    private static IConnectionManager connectionManager;

    private final SQLConnectionType     sqlConnectionType;
    private MenuController              menuController;
    private TablexiaButton              backButton;
    private boolean                     backButtonVisibility;
    private ZipAssetLoader              zipAssetLoader;
    private ScreenshotListener          screenshotListener;
    private boolean                     reset;
    private boolean                     loadingComplete = false;
    private boolean                     errorDialogShown = false;
    private TablexiaComponentDialog     dialog;

    public static class SQLConnectionType {

        private String driver;
        private String connectionString;

        public SQLConnectionType(String driver, String connectionString) {
            this.driver = driver;
            this.connectionString = connectionString;
        }

        public String getDriver() {
            return driver;
        }

        public String getConnectionString() {
            return connectionString;
        }
    }

    public Tablexia(String buildType, Locale systemLocale, SQLConnectionType sqlConnectionType, IConnectionManager connManager, String sentryDSN, boolean hasSoftBackButton, boolean reset) {
        this.reset = reset;
		connectionManager = validateConnectionManager(connManager);
        this.sqlConnectionType = sqlConnectionType;

        TablexiaSettings.init(buildType, systemLocale, hasSoftBackButton);

		// register error handler for exceptions in event bus handler methods
		ApplicationBus.getInstance().addErrorHandler(new IPublicationErrorHandler() {
			@Override
			public void handleError(PublicationError error) {
				Log.err(ApplicationBus.class, error.getMessage(), error.getCause());
			}
		});

        if(TablexiaSettings.getInstance().getBuildType().isBugReport()) {
            @SuppressWarnings("ConstantConditions")
            String DSN = (TablexiaBuildConfig.SENTRY_DSN_FALLBACK == null || sentryDSN == null || TablexiaBuildConfig.SENTRY_DSN_FALLBACK.equals(sentryDSN)) ? null : sentryDSN;
            TablexiaRaven.start(DSN);
        }
    }

    private void loadingComplete() {
        if (!loadingComplete) {
            loadingComplete = true;
            ApplicationBus.getInstance().publishAsync(new ApplicationLoadingCompleteEvent());
        }
    }

    private Class<? extends AbstractTablexiaScreen> showLastOrInitialScreen() {
        AbstractTablexiaScreen<?> screenToChange = TablexiaStorage.getInstance().getSavedScreen(TablexiaSettings.getInstance().getSelectedUser());
        if (screenToChange != null) {
            setScreenIfIsDifferent(screenToChange, ScreenTransaction.FADE);
        } else {
            ScreenTransaction screenTransaction = ScreenTransaction.FADE;
            if (TablexiaSettings.LOADER_SCREEN.equals(TablexiaApplication.getActualScreenClass())) {
                screenTransaction = ScreenTransaction.MOVE_LEFT;
            }
            screenToChange = Utility.getScreenForScreenClass(TablexiaSettings.INITIAL_SCREEN, null);
            setScreenIfIsDifferent(screenToChange, screenTransaction);
        }
        return screenToChange.getClass();
    }

    private void prepareMenu() {
        menuController = new MenuController(TablexiaSettings.getViewportWidth(getStage()), TablexiaSettings.getViewportHeight(getStage()));
        setMenuControllerBounds();
        addMenuController(menuController);
    }

    private void setMenuControllerBounds() {
        if(menuController != null) {
            menuController.setPosition(TablexiaSettings.getViewportLeftX(getStage()), TablexiaSettings.getViewportBottomY(getStage()));
            menuController.setSize(TablexiaSettings.getViewportWidth(getStage()), TablexiaSettings.getViewportHeight(getStage()));
        }
    }

    private void prepareBackButton() {
        backButton = new TablexiaButton(
                "",
                false,
                ApplicationAtlasManager.getInstance().getTextureRegion(ApplicationAtlasManager.BACK_BUTTON),
                ApplicationAtlasManager.getInstance().getTextureRegion(ApplicationAtlasManager.BACK_BUTTON_PRESSED),
                null,
                null)
        .setButtonBounds(
                TablexiaSettings.getViewportLeftX(getStage()) + TablexiaSettings.getViewportWidth(getStage()) * BACK_BUTTON_LEFT_OFFSET_RATIO,
                TablexiaSettings.getViewportBottomY(getStage()) + TablexiaSettings.getViewportHeight(getStage()) * BACK_BUTTON_BOTTOM_OFFSET,
                BACK_BUTTON_SIZE,
                BACK_BUTTON_SIZE)
        .adaptiveSizeRatio(BACK_BUTTON_ADAPTIVE_SIZE_RATIO)
        .adaptiveSizePositionFix(BACK_BUTTON_ADAPTIVE_SIZE_POSITION_FIX);

        backButton.setVisible(false);
        backButton.setInputListener(new ClickListener() {
            @Override
            public void clicked(InputEvent event, float x, float y) {
                ApplicationBus.getInstance().post(new BackButtonPressed()).asynchronously();
            }
        });

        setBackButtonVisibility(backButtonVisibility);
        setBackButton(backButton);
    }

    private void setBackButtonVisibility(boolean hasBackButton) {
        backButtonVisibility = hasBackButton;
        if (backButton != null) {
            if (hasBackButton) {
                if (!backButton.isVisible()) {
                    backButton.addAction(Actions.alpha(0f));
                    backButton.setVisible(true);
                    backButton.addAction(Actions.sequence(Actions.fadeIn(SHOW_BACK_BUTTON_DURATION)));
                }
            } else{
                if (backButton.isVisible()) {
                    backButton.addAction(Actions.sequence(Actions.fadeOut(HIDE_BACK_BUTTON_DURATION), Actions.run(new Runnable() {
                        @Override
                        public void run() {
                            backButton.setVisible(false);
                        }
                    })));
                }
            }
        }
    }

    @Override
    protected void processNewScreen(AbstractTablexiaScreen<?> newScreen) {
        super.processNewScreen(newScreen);
        setBackButtonVisibility(newScreen.hasSoftBackButton());
    }

	@Override
    public void resize(int width, int height) {
        super.resize(width, height);
        setMenuControllerBounds();
    }

    private void startLoading(Locale locale) {
        // Prepare fonts (Fonts need to be loaded before any screen is active)
        ApplicationFontManager.getInstance().load();

        // sync loaded screen with loader image
        setScreen(Utility.getScreenForScreenClass(TablexiaSettings.LOADER_SCREEN, null));

        // async internal assets loading
		ApplicationTextManager.getInstance().load(locale);
        ApplicationInternalSoundManager.getInstance().load();
        ApplicationInternalTextureManager.getInstance().load();

		// async zip extraction
		zipAssetLoader = new ZipAssetLoader(getStage());
		zipAssetLoader.load(locale, Utility.createChecksumMapFromString(TablexiaBuildConfig.ASSETS_CHECKSUM));

		// async external assets loading
		ApplicationExternalSoundManager.getInstance().load();
		ApplicationAtlasManager.getInstance().load();

        loadingComplete = false;
    }

    private void clearLoadedData() {
        // clear main menu
        if (menuController != null) {
            menuController.dispose();
            menuController.remove();
            menuController = null;
        }

        // dispose all loaders
        ApplicationFontManager.getInstance().dispose();
        ApplicationTextManager.getInstance().dispose();
		ApplicationExternalSoundManager.getInstance().dispose();
		zipAssetLoader.dispose();
		zipAssetLoader = null;
		ApplicationAtlasManager.getInstance().dispose();
		ApplicationInternalSoundManager.getInstance().dispose();
        ApplicationInternalTextureManager.getInstance().dispose();
    }


//////////////////////////// APPLICATION FLOW HANDLERS

    @Handler
    public void handleApplicationLoadingCompleteEvent(Tablexia.ApplicationLoadingCompleteEvent applicationLoadingCompleteEvent) {
        if (TablexiaSettings.getInstance().getSelectedUser() != null) {
            Gdx.app.postRunnable(new Runnable() {

                @Override
                public void run() {
                    showLastOrInitialScreen();
                }
            });
        }
    }

    @Handler
    public void handleSelectedUserEvent(final TablexiaSettings.SelectedUserEvent selectedUserEvent) {
        Gdx.app.postRunnable(new Runnable() {
            @Override
            public void run() {
                if (selectedUserEvent.isUserSelected()) {
                    showLastOrInitialScreen();
                }
            }
        });
    }


//////////////////////////// LOCALE CHANGED HANDLER

    @Handler
    public void handleLocaleChangedEvent(TablexiaSettings.LocaleChangedEvent localeChangedEvent) {
        Gdx.app.postRunnable(new Runnable() {

            @Override
            public void run() {
                clearLoadedData();
                startLoading(TablexiaSettings.getInstance().getLocale());
            }
        });
    }


//////////////////////////// LIBGDX LIFECYCLE

    @Override
    public void create() {
        super.create();

        TablexiaRaven.sendSavedReports();

        // init data storage
        TablexiaStorage.init(sqlConnectionType, reset);

        // init dialog factory
        TablexiaComponentDialogFactory.init(getStage());

        TablexiaSettings.getInstance().loadPreferences(reset);
        Log.setLoglevel(TablexiaSettings.getInstance().getLogLevel());
		
		// init event bus handlers
		ApplicationBus.getInstance().subscribe(this);

        // start loading application scope data
        startLoading(TablexiaSettings.getInstance().getLocale());
    }

    @Override
    public void render() {
        // render other screens
        super.render();

        if (screenshotListener != null) {
            obtainScreenshot();
            screenshotListener = null;
        }

        // process loading
        if (!loadingComplete) {
            // load internal assets
			if (!ApplicationFontManager.getInstance().update()) return;
			if (!ApplicationTextManager.getInstance().update()) return;
			if (!ApplicationInternalSoundManager.getInstance().update()) return;
            if (!ApplicationInternalTextureManager.getInstance().update()) return;

			// load external assets
			if (!zipAssetLoader.update()) return;
			if (processLoaderError(zipAssetLoader.getError()))   return;

			if (!ApplicationExternalSoundManager.getInstance().update()) return;
			if (!ApplicationAtlasManager.getInstance().update()) return;

            prepareBackButton();
            prepareMenu();
            loadingComplete();
        }
    }

    /**
     * Queues request for screenshot with this listener, it is called after obtaining screenshot.
     *
     * @param listener
     */
    public void takeScreenShot(ScreenshotListener listener) {
        this.screenshotListener = listener;
    }

    /**
     * Listener interface, that gets called after creating screenshot array
     */
    public interface ScreenshotListener {
        void screenshotTaken(int[] screen);
    }

    /**
     * Method launched during render phase before cleaning the buffer, creates bytearray of screen pixels and passes it to the listener
     */
    private void obtainScreenshot() {
        Gdx.gl.glFinish();
        int width = Gdx.graphics.getWidth();
        int height = Gdx.graphics.getHeight();
        int bt[] = new int[width * height];

        ByteBuffer pbuf = ByteBuffer.allocateDirect(width * height * 4);
        pbuf.order(ByteOrder.nativeOrder());
        Gdx.gl.glReadPixels(0, 0, width, height, GL30.GL_RGBA, GL30.GL_UNSIGNED_BYTE, pbuf);
        pbuf.asIntBuffer().get(bt);
        if (screenshotListener != null) {
            screenshotListener.screenshotTaken(bt);
        }
    }

    @Override
    public void dispose() {
        super.dispose();
        TablexiaStorage.getInstance().dispose();
        TablexiaSettings.getInstance().dispose();
        TablexiaComponentDialogFactory.getInstance().dispose();
        clearLoadedData();
    }

	private static IConnectionManager validateConnectionManager(IConnectionManager connectionManager) {
		if(connectionManager == null)
			throw new RuntimeException(Tablexia.class.getName() + ": Connection Manager equals to " + connectionManager);

		return connectionManager;
	}

	public static IConnectionManager getConnectionManager() {
		return connectionManager;
	}


//////////////////////////// APPLICATION LOADERS

    public void showLoaderErrorDialog(int dialogWidth, int dialogHeight, String loaderError) {
        ArrayList<TablexiaDialogComponentAdapter> components = new ArrayList<TablexiaDialogComponentAdapter>();

        components.add(new CenterPositionDialogComponent());
        components.add(new AdaptiveSizeDialogComponent());
        components.add(new DimmerDialogComponent(ERROR_DIALOG_DIMMER));
        components.add(new BackButtonHideComponent(LOADER_ERROR_DIALOG_HIDE_ANIMATION) {
            @Override
            public void onBackButtonPress(BackButtonPressed event) {
                super.onBackButtonPress(event);
                returnToLastLocale();
            }
        });

        components.add(new FixedSpaceContentDialogComponent());
        components.add(new ResizableSpaceContentDialogComponent());

        components.add(new TextContentDialogComponent(loaderError));

        components.add(new FixedSpaceContentDialogComponent());
        components.add(new ResizableSpaceContentDialogComponent());
        components.add(new PositiveNegativeButtonContentDialogComponent(
				new ClickListener() {
					@Override
					public void clicked(InputEvent event, float x, float y) {
						zipAssetLoader.reset(true);
						errorDialogShown = false;
					}
				},
                new ClickListener() {
                    @Override
                    public void clicked(InputEvent event, float x, float y) {
                        returnToLastLocale();
                    }
                },
				TablexiaSettings.getInstance().hasLocaleChanged() ? PositiveNegativeButtonContentDialogComponent.PositiveNegativeButtonType.AGAIN_BACK : PositiveNegativeButtonContentDialogComponent.PositiveNegativeButtonType.AGAIN_EXIT
        ) {
            @Override
            protected boolean hasHideAnimation() {
                return LOADER_ERROR_DIALOG_HIDE_ANIMATION;
            }
        });
        components.add(new FixedSpaceContentDialogComponent());
        dialog = TablexiaComponentDialogFactory.getInstance().createDialog(components.toArray(new TablexiaDialogComponentAdapter[]{}));
        dialog.show(dialogWidth, dialogHeight);
    }

    private void returnToLastLocale() {
        if (TablexiaSettings.getInstance().hasLocaleChanged()) {
            TablexiaSettings.getInstance().returnToLastLocale();
            zipAssetLoader.reset(false);
            errorDialogShown = false;
        } else {
            Gdx.app.exit();
        }
    }

    private boolean processLoaderError(String loaderError) {
        if (loaderError != null && !errorDialogShown) {
            showLoaderErrorDialog(ERROR_DIALOG_WIDTH, ERROR_DIALOG_HEIGHT, loaderError);
            errorDialogShown = true;
        } else if (loaderError == null && dialog != null) {
			dialog.hide();
		}
        return errorDialogShown;
    }

    public static class ApplicationLoadingCompleteEvent implements ApplicationEvent {
        private ApplicationLoadingCompleteEvent(){}
    }


//////////////////////////// CHANGE SCREEN EVENT

    public static class ChangeScreenEvent implements ApplicationEvent {

        private Class<? extends AbstractTablexiaScreen<?>>  screen;
        private ScreenTransaction                           screenTransaction;
        private Map<String, String>                         screenInitialState;

        public ChangeScreenEvent(Class<? extends AbstractTablexiaScreen<?>> screen, ScreenTransaction screenTransaction) {
            this(screen, screenTransaction, null);
        }

        public ChangeScreenEvent(Class<? extends AbstractTablexiaScreen<?>> screen, ScreenTransaction screenTransaction, Map<String, String> screenInitialState) {
            this.screen = screen;
            this.screenTransaction = screenTransaction;
            this.screenInitialState = screenInitialState;
        }

        public Class<? extends AbstractTablexiaScreen<?>> getScreen() {
            return screen;
        }

        public ScreenTransaction getScreenTransaction() {
            return screenTransaction;
        }

        public Map<String, String> getScreenInitialState() {
            return screenInitialState;
        }
    }

    @Handler
    public void handleChangeScreenEvent(final ChangeScreenEvent changeScreenEvent) {
        final Class<? extends AbstractTablexiaScreen<?>> screenClass = changeScreenEvent.getScreen();
        if (!loadingComplete) {
            Log.err(((Object) this).getClass(), "Cannot change screen -> Application loading not complete!");
            return;
        }
        if (screenClass == null) {
            Log.err(((Object) this).getClass(), "Cannot change screen -> Received empty screen class!");
            return;
        }
        // create new screen on GL thread
        Gdx.app.postRunnable(new Runnable() {

            @Override
            public void run() {
                setScreenIfIsDifferent(Utility.getScreenForScreenClass(screenClass, changeScreenEvent.getScreenInitialState()), changeScreenEvent.getScreenTransaction());
            }
        });
    }
}