package cz.nic.tablexia;

import com.badlogic.gdx.Application;
import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.Preferences;
import com.badlogic.gdx.graphics.Color;
import com.badlogic.gdx.scenes.scene2d.Actor;
import com.badlogic.gdx.scenes.scene2d.InputEvent;
import com.badlogic.gdx.scenes.scene2d.Stage;
import com.badlogic.gdx.scenes.scene2d.utils.ClickListener;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.List;
import java.util.Locale;

import cz.nic.tablexia.bus.ApplicationBus;
import cz.nic.tablexia.loader.application.ApplicationTextManager;
import cz.nic.tablexia.model.UserDAO;
import cz.nic.tablexia.screen.AbstractTablexiaScreen;
import cz.nic.tablexia.screen.gamemenu.GameMenuScreen;
import cz.nic.tablexia.screen.loader.LoaderScreen;
import cz.nic.tablexia.shared.model.User;
import cz.nic.tablexia.util.Log;
import cz.nic.tablexia.util.ui.dialog.TablexiaComponentDialogFactory;

public class TablexiaSettings {

    private static final int    DEFAULT_SCREEN_WIDTH     = 1000;
    private static final double MAXIMUM_RATIO            = 9.0 / 16.0;
    private static final int    MIN_SCREEN_HEIGHT        = (int) (DEFAULT_SCREEN_WIDTH * MAXIMUM_RATIO);
    private static final Color  DEFAULT_BACKGROUND_COLOR = new Color(0.99f, 0.88f, 0.71f, 1);
    private static final Color  DEFAULT_FONT_COLOR       = Color.BLACK;

    public static final Class<? extends AbstractTablexiaScreen<?>> LOADER_SCREEN     = LoaderScreen.class;
    public static final Class<? extends AbstractTablexiaScreen<?>> INITIAL_SCREEN    = GameMenuScreen.class;
    public static final Long                                       GAMES_RANDOM_SEED = null;
    public static final int                                        AVATAR_COUNT      = 12;

    private static final String  PREFERENCES_KEY         = "cz.nic.tablexia.";
    private static final String  LAST_LOCALE_KEY         = "last_locale";
    public  static final String  LOCALE_KEY              = "locale";
    private static final String  SELECTED_USER           = "selected_user";

    public  static final int     EXPERIENCE_POINTS_TUTORIAL = 0;
    public  static final int     EXPERIENCE_POINTS_EASY     = 1;
    public  static final int     EXPERIENCE_POINTS_MEDIUM   = 2;
    public  static final int     EXPERIENCE_POINTS_HARD     = 3;

	//Interval <start month; end month>
	private static final int     WINTER_MODE_START_MONTH = 10;
	private static final int     WINTER_MODE_END_MONTH	 = 2;
	private static       boolean winterMode = false;

    private static final    String      APP_NAME        = TablexiaBuildConfig.APP_NAME;
    private final           String      VERSION_NAME    = TablexiaBuildConfig.VERSION_NAME;
    private final           BuildType   BUILD_TYPE;
    private final           boolean     HAS_SOFT_BACK_BUTTON;

    private Preferences      preferences;
    private LocaleDefinition systemLocale;
    private LocaleDefinition selectedLocale;
	private LocaleDefinition lastLocale;
    private User             selectedUser;

    private boolean          soundMuted = false;


    public enum BuildType {

        //                  boxes   aInfo   sInfo   dMenu   btName  bReport  debugBtn
        RELEASE ("release", false,  false,  false,  false,  false,  true,    false,   Log.TablexiaLogLevel.ERROR),
        DEBUG   ("debug",   false,  true,   true,   true,   true,   true,    true,    Log.TablexiaLogLevel.DEBUG),
        DEVEL   ("devel",   true,   true,   true,   true,   true,   false,   true,    Log.TablexiaLogLevel.DEBUG);

        private final static BuildType FALLBACK_VARIANT = BuildType.DEVEL;

        private final String               key;
        private       boolean              boundingBoxes;
        private       boolean              applicationInfo;
        private       boolean              screenInfo;
        private       boolean              debugMenu;
        private       boolean              showBuildTypeInName;
        private       boolean              bugReport;
        private       boolean              showDebugButtons;
        private final Log.TablexiaLogLevel logLevel;

        BuildType(String key, boolean boundingBoxes, boolean applicationInfo, boolean screenInfo, boolean debugMenu, boolean showBuildTypeInName, boolean bugReport, boolean showDebugButtons, Log.TablexiaLogLevel logLevel) {
            this.key = key;
            this.boundingBoxes = boundingBoxes;
            this.applicationInfo = applicationInfo;
            this.screenInfo = screenInfo;
            this.debugMenu = debugMenu;
            this.showBuildTypeInName = showBuildTypeInName;
            this.bugReport = bugReport;
            this.showDebugButtons = showDebugButtons;
            this.logLevel = logLevel;
        }

        public String getKey() {
            return key;
        }

        public String getAppName() {
            return showBuildTypeInName ? APP_NAME + " " + key.toUpperCase() : APP_NAME;
        }

        public boolean isBoundingBoxes() {
            return boundingBoxes;
        }

        public boolean isApplicationInfo() {
            return applicationInfo;
        }

        public boolean isScreenInfo() {
            return screenInfo;
        }

        public boolean isDebugMenu() {
            return debugMenu;
        }

        public boolean isBugReport() {
            return bugReport;
        }

        public boolean isShowDebugButtons() {
            return showDebugButtons;
        }

        public Log.TablexiaLogLevel getLogLevel() {
            return logLevel;
        }

        public static BuildType getBuildTypeForKey(String key) {
            for (BuildType buildType : BuildType.values()) {
                if (buildType.key.equals(key)) {
                    return buildType;
                }
            }
            return FALLBACK_VARIANT;
        }
    }


//////////////////////////// SETTINGS SINGLETON

    private static TablexiaSettings instance;

    private TablexiaSettings(BuildType buildType, Locale systemLocale, boolean hasSoftBackButton) {
        BUILD_TYPE = buildType;
        HAS_SOFT_BACK_BUTTON = hasSoftBackButton;
        this.systemLocale = LocaleDefinition.getLocaleDefinitionForLocale(systemLocale);
    }

    public static TablexiaSettings getInstance() {
        if (instance == null) {
            String exceptionMessage = "Tablexia settings is not initialized!";
            Log.err(TablexiaSettings.class, exceptionMessage);
            throw new IllegalStateException(exceptionMessage);
        }
        return instance;
    }

    static void init(String buildTypeKey, Locale systemLocale, boolean hasSoftBackButton) {
        TablexiaSettings.init(BuildType.getBuildTypeForKey(buildTypeKey), systemLocale, hasSoftBackButton);
    }

    static void init(BuildType buildType, Locale systemLocale, boolean hasSoftBackButton) {
        if (instance != null) {
            String exceptionMessage = "Tablexia settings already initialized!";
            Log.err(TablexiaSettings.class, exceptionMessage);
            return;
        }

        instance = new TablexiaSettings(buildType, systemLocale, hasSoftBackButton);
    }

    void dispose() {
        instance = null;
    }


//////////////////////////// LIBGDX PREFERENCES

    public void loadPreferences(boolean reset) {
        preferences = Gdx.app.getPreferences(PREFERENCES_KEY + BUILD_TYPE.getKey());
        lastLocale = LocaleDefinition.getLocaleDefinitionForKey(preferences.getString(LAST_LOCALE_KEY));
        selectedLocale = LocaleDefinition.getLocaleDefinitionForKey(preferences.getString(LOCALE_KEY));
        selectedUser = reset ? null : UserDAO.selectUser(preferences.getLong(SELECTED_USER));

		Calendar calendar = GregorianCalendar.getInstance();
		winterMode = (calendar.get(Calendar.MONTH) > WINTER_MODE_START_MONTH) || (calendar.get(Calendar.MONTH) < WINTER_MODE_END_MONTH);
	}


//////////////////////////// SETTINGS ACCESS

    public String getFullName() {
        return getAppName() + " " + getVersionName();
    }

    public String getAppName() {
        return getBuildType().getAppName();
    }

    public String getVersionName() {
        return VERSION_NAME;
    }

    public boolean hasSoftBackButton() {
        return HAS_SOFT_BACK_BUTTON;
    }

    public boolean isWinterMode() { return winterMode; }

    public boolean isShowBoundingBoxes() {
        return getBuildType().isBoundingBoxes();
    }

    public Log.TablexiaLogLevel getLogLevel() {
        return BUILD_TYPE.getLogLevel();
    }

    public BuildType getBuildType() {
        return BUILD_TYPE;
    }

    public static Color getDefaultBackgroundColor() {
        return DEFAULT_BACKGROUND_COLOR;
    }

    public static Color getDefaultFontColor() {
        return DEFAULT_FONT_COLOR;
    }

//////////////////////////// SOUND MUTE

    public boolean isSoundMuted() { return soundMuted; }

    public void toggleSoundMute() {
        this.soundMuted = !this.soundMuted;
        ApplicationBus.getInstance().publishAsync(new SoundMuteEvent(soundMuted));
    }

    public static class SoundMuteEvent implements ApplicationBus.ApplicationEvent {
        private boolean soundMuted;

        private SoundMuteEvent(boolean soundMuted) {
            this.soundMuted = soundMuted;
        }

        public boolean isSoundMuted() {
            return soundMuted;
        }
    }

//////////////////////////// USER SETTINGS

    public static class SelectedUserEvent implements ApplicationBus.ApplicationEvent {

        private boolean userSelected;

        private SelectedUserEvent(boolean userSelected) {
            this.userSelected = userSelected;
        }

        public boolean isUserSelected() {
            return userSelected;
        }
    }

    public void changeUser(User user, LogoutAcceptListener logoutAcceptListener) {
        if(this.selectedUser == null)
            setSelectedUser(user);

        if (!this.selectedUser.equals(user)) {
            showLogoutUserDialog(user, logoutAcceptListener);
        }
    }

    public void changeUser(User user) {
        changeUser(user, null);
    }

    private void setSelectedUser(User user) {
        selectedUser = user;

        if (selectedUser == null) {
            preferences.remove(SELECTED_USER);
        } else {
            preferences.putLong(SELECTED_USER, selectedUser.getId());
        }
        preferences.flush();
        ApplicationBus.getInstance().publishAsync(new SelectedUserEvent(user != null));
    }

    private void setLastLocale(LocaleDefinition lastLocale) {
        this.lastLocale = lastLocale;

        if (lastLocale == null) {
            preferences.remove(LAST_LOCALE_KEY);
        } else {
            preferences.putString(LAST_LOCALE_KEY, lastLocale.getLocaleKey());
        }
        preferences.flush();
    }

    public interface LogoutAcceptListener  {
        void onLogoutAccepted();
    }

    public void showLogoutUserDialog(final User user, final LogoutAcceptListener logoutAcceptListener) {
        TablexiaComponentDialogFactory.getInstance().createWarningYesNoDialog(
                ApplicationTextManager.getInstance().getText(ApplicationTextManager.ApplicationTextsAssets.USER_LOGOUT_MESSAGE),
                new ClickListener() {
                    @Override
                    public void clicked(InputEvent event, float x, float y) {
                        setSelectedUser(user);

                        if (logoutAcceptListener != null)
                            logoutAcceptListener.onLogoutAccepted();
                    }
                }, true).show(TablexiaComponentDialogFactory.WARNING_DIALOG_WIDTH, TablexiaComponentDialogFactory.WARNING_DIALOG_HEIGHT);
    }

    public User getSelectedUser() {
        return selectedUser;
    }


//////////////////////////// LOCALE SETTINGS

    public static class LocaleChangedEvent implements ApplicationBus.ApplicationEvent {

        private TablexiaSettings.LocaleDefinition localeDefinition;

        private LocaleChangedEvent(TablexiaSettings.LocaleDefinition localeDefinition) {
            this.localeDefinition = localeDefinition;
        }

        public TablexiaSettings.LocaleDefinition getLocaleDefinition() {
            return localeDefinition;
        }
    }

    public enum LocaleDefinition {

        SYSTEM  (null,                      "system",   ApplicationTextManager.ApplicationTextsAssets.LANGUAGE_SYSTEM,  true),
        cs_CZ   (new Locale("cs", "CZ"),    "cs_CZ",    ApplicationTextManager.ApplicationTextsAssets.LANGUAGE_CZECH,   true),
        sk_SK   (new Locale("sk", "SK"),    "sk_SK",    ApplicationTextManager.ApplicationTextsAssets.LANGUAGE_SLOVAK,  true),
        de_DE   (new Locale("de", "DE"),    "de_DE",    ApplicationTextManager.ApplicationTextsAssets.LANGUAGE_GERMAN,  true);

        public final static  LocaleDefinition DEFAULT_LOCALE   = LocaleDefinition.cs_CZ;
        private final static LocaleDefinition FALLBACK_VARIANT = LocaleDefinition.SYSTEM;

        private final Locale    locale;
        private final String    localeKey;
        private final String    descriptionKey;
        private boolean         enabled;

        private static List<LocaleDefinition> enabledLocaleDefinitions;

        LocaleDefinition(Locale locale, String localeKey, String descriptionKey, boolean enabled) {
            this.locale = locale;
            this.localeKey = localeKey;
            this.descriptionKey = descriptionKey;
            this.enabled = enabled;
        }

        public String getLocaleKey() {
            return localeKey;
        }

        public String getDescriptionKey() {
            return descriptionKey;
        }

        public Locale getLocale() {
            return locale != null ? locale : instance.systemLocale.locale;
        }

        @Override
        public String toString() {
            if (ApplicationTextManager.getInstance().update()) {
                return ApplicationTextManager.getInstance().getText(descriptionKey);
            } else {
                return name();
            }
        }

        public static LocaleDefinition getLocaleDefinitionForKey(String key) {
            for (LocaleDefinition localeDefinition : getEnabledLocaleDefinitions()) {
                if (localeDefinition.localeKey.equals(key)) {
                    return localeDefinition;
                }
            }
            return FALLBACK_VARIANT;
        }

        public static LocaleDefinition getLocaleDefinitionForLocale(Locale locale) {
            for (LocaleDefinition localeDefinition : getEnabledLocaleDefinitions()) {
                if (localeDefinition.locale != null && localeDefinition.locale.equals(locale)) {
                    return localeDefinition;
                }
            }
            return DEFAULT_LOCALE;
        }

        public static List<LocaleDefinition> getEnabledLocaleDefinitions() {
            if (enabledLocaleDefinitions == null) {
                enabledLocaleDefinitions = new ArrayList<LocaleDefinition>();
                for (LocaleDefinition localeDefinition : LocaleDefinition.values()) {
                    if (localeDefinition.enabled) {
                        enabledLocaleDefinitions.add(localeDefinition);
                    }
                }
            }
            return enabledLocaleDefinitions;
        }
    }

    public void setLocale(LocaleDefinition localeDefinition) {
        setLastLocale(selectedLocale);
        selectedLocale = localeDefinition;
        preferences.putString(LOCALE_KEY, localeDefinition.getLocaleKey());
        preferences.flush();
        if (hasLocaleChanged()) {
            ApplicationBus.getInstance().post(new LocaleChangedEvent(localeDefinition)).asynchronously();
        }
    }

	public boolean hasLocaleChanged() {
		return lastLocale == null || !lastLocale.getLocale().getLanguage().equals(selectedLocale.getLocale().getLanguage());
	}

    public LocaleDefinition getLocaleDefinition() {
        return selectedLocale;
    }

	public void returnToLastLocale() {
		if (hasLocaleChanged()) {
			setLocale(getLastLocale());
		}
	}

	public LocaleDefinition getLastLocale() {
		return lastLocale;
	}

	public Locale getLocale() {
        return selectedLocale.getLocale();
    }


//////////////////////////// SCREEN SIZE

    public static int getWorldSize() {
        return DEFAULT_SCREEN_WIDTH;
    }

    public static double getMaximumRatio() {
        return MAXIMUM_RATIO;
    }

    public static int getMinWorldHeight() {
        return MIN_SCREEN_HEIGHT;
    }

    private static float calculateWordSizeOffset() {
        return (getWorldSize() - getMinWorldHeight()) / 2f;
    }


//////////////////////////// SCENE SIZE

    public static float getSceneLeftX(Stage stage) {
        return 0;
    }

    public static float getSceneRightX(Stage stage) {
        return stage.getWidth();
    }

    public static float getSceneOuterBottomY(Stage stage) {
        return -calculateWordSizeOffset();
    }

    public static float getSceneInnerBottomY(Stage stage) {
        return 0;
    }

    public static float getSceneInnerTopY(Stage stage) {
        return getSceneInnerBottomY(stage) + MIN_SCREEN_HEIGHT;
    }

    public static float getSceneOuterTopY(Stage stage) {
        return getSceneInnerTopY(stage) + calculateWordSizeOffset();
    }

    public static float getSceneWidth(Stage stage) {
        return getSceneRightX(stage) - getSceneLeftX(stage);
    }

    public static float getSceneOuterHeight(Stage stage) {
        return getSceneOuterTopY(stage) - getSceneOuterBottomY(stage);
    }

    public static float getSceneInnerHeight(Stage stage) {
        return getSceneInnerTopY(stage) - getSceneInnerBottomY(stage);
    }


//////////////////////////// SCENE VIEWPORT

    public static float getViewportLeftX(Stage stage) {
        return 0;
    }

    public static float getViewportRightX(Stage stage) {
        return getViewportLeftX(stage) + stage.getWidth();
    }

    public static float getViewportBottomY(Stage stage) {
        return stage.getCamera().position.y - stage.getHeight() / 2;
    }

    public static float getViewportTopY(Stage stage) {
        return getViewportBottomY(stage) + stage.getHeight();
    }

    public static float getViewportWidth(Stage stage) {
        return getViewportRightX(stage) - getViewportLeftX(stage);
    }

    public static float getViewportHeight(Stage stage) {
        return getViewportTopY(stage) - getViewportBottomY(stage);
    }

	public static float getActorSceneVerticalCenter(Stage stage, Actor actor) {
		return getViewportHeight(stage) / 2f - actor.getHeight() / 2f + getViewportBottomY(stage);
	}

//////////////////////////// DEVICE INFO

    public boolean isRunningOnMobileDevice() {
        return  Gdx.app.getType() == Application.ApplicationType.Android ||
                Gdx.app.getType() == Application.ApplicationType.iOS;
    }
}