/*******************************************************************************
 *     Tablexia
 * 
 *     Copyright (C) 2013  CZ NIC z.s.p.o. <podpora at nic dot cz>
 * 
 *     This program is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ******************************************************************************/

package cz.nic.tablexia.game.games.in_the_darkness.action.widget;

import com.badlogic.gdx.graphics.g2d.TextureRegion;
import com.badlogic.gdx.math.Vector2;
import com.badlogic.gdx.scenes.scene2d.Group;
import com.badlogic.gdx.scenes.scene2d.actions.RepeatAction;
import com.badlogic.gdx.scenes.scene2d.ui.Image;

import java.util.ArrayList;
import java.util.List;

import cz.nic.tablexia.game.games.in_the_darkness.InTheDarknessDifficulty;
import cz.nic.tablexia.game.games.in_the_darkness.InTheDarknessGame;
import cz.nic.tablexia.game.games.in_the_darkness.action.Action;
import cz.nic.tablexia.game.games.in_the_darkness.action.Action.ActionListener;
import cz.nic.tablexia.game.games.in_the_darkness.action.ActionType;
import cz.nic.tablexia.game.games.in_the_darkness.assets.InTheDarknessAssets;

import static com.badlogic.gdx.scenes.scene2d.actions.Actions.delay;
import static com.badlogic.gdx.scenes.scene2d.actions.Actions.fadeIn;
import static com.badlogic.gdx.scenes.scene2d.actions.Actions.fadeOut;
import static com.badlogic.gdx.scenes.scene2d.actions.Actions.forever;
import static com.badlogic.gdx.scenes.scene2d.actions.Actions.run;
import static com.badlogic.gdx.scenes.scene2d.actions.Actions.sequence;

/**
 * Widget for showing action icons
 * 
 * @author Matyáš Latner
 *
 */
public class ActionsWidget extends Group implements ActionListener {

	public enum ActionLayer {
		BACKGROUND_LAYER			(0),
        ACTIONS_LAYER				(1);

        private Group layerEntity;
        private int layerZIndex;

        ActionLayer(int layerZIndex) {
            this.layerZIndex = layerZIndex;
        }

        private void setLayerEntity(Group layerEntity) {
            this.layerEntity = layerEntity;
        }

        public Group getLayerEntity() {
            return layerEntity;
        }

        public int getLayerZIndex() {
            return layerZIndex;
        }

        public static void attachLayersToEntity(Group containerGroup) {
            for (ActionLayer actionLayer : ActionLayer.values()) {
                Group group = new Group();
                actionLayer.setLayerEntity(group);
                containerGroup.addActor(group);
                group.setZIndex(actionLayer.getLayerZIndex());
            }
        }
    }
	
	private static final float 	TUTORIAL_INFO_ARROW_BLINK_DURATION 	= 0.3f;
	private static final float 	START_ANIMATION_DELAY 				= 0.15f;
	private static final int 	NUMBER_OF_COLUMNS 					= 2;
	public  static final int 	ACTION_OFFSET 						= InTheDarknessGame.ACTION_SIZE_SMALLER / 10;
	private static final float 	START_ARROW_X_OFFSET 				= InTheDarknessGame.ACTION_SIZE_SMALLER / 15;
	private static final float 	START_ARROW_Y_OFFSET 				= -InTheDarknessGame.ACTION_SIZE_SMALLER / 20;
	private static final float 	START_ARROW_WIDTH 					= InTheDarknessGame.TILE_SIZE * 1.5f;
	private static final float 	START_ARROW_HEIGHT 					= InTheDarknessGame.TILE_SIZE;

	private ActionsStripWidget 			actionsStripWidget;
	private List<Action>				actions;
	private InTheDarknessDifficulty 	inTheDarknessDifficulty;
	private int 						currentTutorialStepNumber;
	private InTheDarknessGame 			inTheDarknessGame;
	private Image 						tutorialArrowImage;
	private RepeatAction 				blinkAction;


	public ActionsWidget(InTheDarknessDifficulty inTheDarknessDifficulty,
						 ActionsStripWidget actionsStripWidget,
						 InTheDarknessGame inTheDarknessGame) {

		this.inTheDarknessGame 			= inTheDarknessGame;
		this.inTheDarknessDifficulty 	= inTheDarknessDifficulty;
		this.actionsStripWidget 		= actionsStripWidget;
		this.actions 					= new ArrayList<Action>();
		this.currentTutorialStepNumber 	= 0;

		// SET ACTOR SIZE
		List<ActionType> actionTypes = ActionType.getActionTypesForGameDifficulty(inTheDarknessDifficulty);
		setSize((float) (NUMBER_OF_COLUMNS * (InTheDarknessGame.ACTION_SIZE_SMALLER + ACTION_OFFSET) - ACTION_OFFSET),
				(float) Math.ceil(Double.valueOf(actionTypes.size()) / NUMBER_OF_COLUMNS) * (InTheDarknessGame.ACTION_SIZE_SMALLER + ACTION_OFFSET) - ACTION_OFFSET);

		// PREPARE LAYERS
		ActionLayer.attachLayersToEntity(this);

		prepareBackground();
		prepareActions(actionTypes);
		prepareTutorialArrowImage();
	}

	private Action createAction(ActionType actionType, int orderNumber, boolean visible, InTheDarknessGame inTheDarknessGame) {
		int actionSize = InTheDarknessGame.ACTION_SIZE_SMALLER;
		Action action = new Action(actionType,
				   				   orderNumber,
								   actionSize,
				   				   (orderNumber % NUMBER_OF_COLUMNS) * (actionSize + ACTION_OFFSET),
								   getHeight() - actionSize - ((orderNumber / NUMBER_OF_COLUMNS) * (actionSize + ACTION_OFFSET)),
				   				   visible,
								   inTheDarknessGame);
		
		action.addActionListener(this);
		action.setClickable();
		ActionLayer.ACTIONS_LAYER.getLayerEntity().addActor(action);
		actions.add(action.getOrderNumber(), action);
		return action;
	}


//////////////////////////////////////////// SCREEN COMPONENTS

	private void prepareBackground() {
		TextureRegion backgroundTexture = inTheDarknessGame.getScreenTextureRegion(InTheDarknessAssets.BACKGROUND_ACTIONS);
		float backgroundHeight = getHeight();
		float backgroundWidth = backgroundTexture.getRegionWidth() * (backgroundHeight / backgroundTexture.getRegionHeight());
		Image actionsBackground = new Image(backgroundTexture);
		actionsBackground.setSize(backgroundWidth, backgroundHeight);
		actionsBackground.setPosition((getWidth() / 2) - (backgroundWidth / 2), 0);
		ActionLayer.BACKGROUND_LAYER.getLayerEntity().addActor(actionsBackground);
	}

	private void prepareActions(List<ActionType> actionTypes) {
		for (ActionType actionType: actionTypes) {
			createAction(actionType, actionType.ordinal(), false, inTheDarknessGame);
		}
	}

	private void prepareTutorialArrowImage() {
		if (inTheDarknessDifficulty == InTheDarknessDifficulty.TUTORIAL) {
			tutorialArrowImage = new Image(inTheDarknessGame.getScreenTextureRegion(InTheDarknessAssets.INFO_START_ARROW));
			tutorialArrowImage.setSize(START_ARROW_WIDTH, START_ARROW_HEIGHT);
			tutorialArrowImage.setVisible(false);
			blinkAction = forever(sequence(fadeIn(TUTORIAL_INFO_ARROW_BLINK_DURATION), fadeOut(TUTORIAL_INFO_ARROW_BLINK_DURATION)));
			tutorialArrowImage.addAction(blinkAction);
		}
	}

	
//////////////////////////////////////////// COLLISIONS
	
	private void prepareActionsReactionAreas(Action action) {
		action.setActionsStripWidget(actionsStripWidget);
	}
	
	
//////////////////////////////////////////// ANIMATIONS
	
	public void showActions() {
		tryDimAllActions();
		float actualDelay = 0;
		for (int i = 0; i < actions.size(); i++) {
			final Action action = actions.get(i);

			addAction(sequence(delay(actualDelay), run(new Runnable() {
				@Override
				public void run() {
					action.showWithAnimation();
					prepareActionsReactionAreas(action);
				}
			})));
			actualDelay = actualDelay + START_ANIMATION_DELAY;
		}

		addAction(sequence(delay(actualDelay + START_ANIMATION_DELAY), run(new Runnable() {
			@Override
			public void run() {
				tryToPerformNextTutorialStep();
			}
		})));
	}
	
	
//////////////////////////////////////////// ACTION STATE
	
	public void enableActions() {
		if (inTheDarknessDifficulty != InTheDarknessDifficulty.TUTORIAL) {
			for (int i = 0; i < actions.size(); i++) {
				Action action = actions.get(i);
				action.enable();
			}
		}
	}
	
	public void disableActions() {
		for (int i = 0; i < actions.size(); i++) {
			Action action = actions.get(i);
			action.disable();
		}
	}
	
	
//////////////////////////////////////////// ACTION LISTENER

	@Override
	public void onActionDrag(Action lastAction) {
		lastAction.setVisible(false);
		Vector2 stageCoordinates = lastAction.localToStageCoordinates(new Vector2());
		InTheDarknessGame.GameLayer.OVERLAY_LAYER.getLayerEntity().stageToLocalCoordinates(stageCoordinates);
		lastAction.setPosition(stageCoordinates.x, stageCoordinates.y);
		InTheDarknessGame.GameLayer.OVERLAY_LAYER.getLayerEntity().addActor(lastAction);
		lastAction.setVisible(true);

		actions.remove(lastAction);
		Action action = createAction(lastAction.getActionType(), lastAction.getOrderNumber(), false, inTheDarknessGame);
		prepareActionsReactionAreas(action);
		action.showWithAnimation();
		disableActions();
		tryHideTutorialArrowSprite();
	}
	
	@Override
	public void onActionDrop(Action action, int collidesWithNumber) {
		if (collidesWithNumber != Action.NO_COLLISION_NUMBER) {
			actionsStripWidget.addSelectedAction(action.getActionType(), collidesWithNumber, inTheDarknessGame);
			currentTutorialStepNumber++;
		}
		tryToPerformNextTutorialStep();
	}
	
	
//////////////////////////////////////////// TUTORIAL STEPS
	
	private void tryToPerformNextTutorialStep() {
		if (inTheDarknessDifficulty == InTheDarknessDifficulty.TUTORIAL) {
			if (currentTutorialStepNumber < InTheDarknessGame.TUTORIAL_STEPS.size()) {
        		highliteAction(InTheDarknessGame.TUTORIAL_STEPS.get(currentTutorialStepNumber));
        	} else {
        		InTheDarknessGame.GameLayer.BUTTON_LAYER.sendToFront();
				inTheDarknessGame.disableStartButton(false);
				inTheDarknessGame.highliteStartButton(true);
        	}
        } else {
        	enableActions();
        }
	}
	
	private void tryDimAllActions() {
		if (inTheDarknessDifficulty == InTheDarknessDifficulty.TUTORIAL) {
			inTheDarknessGame.showDimmer();
		}
	}
	
	private void highliteAction(ActionType actionType) {
		if (actionType.ordinal() < actions.size()) {
			Action action = actions.get(actionType.ordinal());
			action.enable();
			showArrowSprite(action.getX() + action.getWidth(), action.getY() - action.getHeight() / 2);
			action.setVisible(false);
			Vector2 stageCoordinates = action.localToStageCoordinates(new Vector2());
			InTheDarknessGame.GameLayer.OVERLAY_LAYER.getLayerEntity().stageToLocalCoordinates(stageCoordinates);
			action.setPosition(stageCoordinates.x, stageCoordinates.y);
			InTheDarknessGame.GameLayer.OVERLAY_LAYER.getLayerEntity().addActor(action);
			action.setVisible(true);
		}
	}
	
	private void showArrowSprite(float positionX, float positionY) {
		if (tutorialArrowImage != null) {
			addActor(tutorialArrowImage);
			tutorialArrowImage.setPosition(positionX + START_ARROW_X_OFFSET, positionY + START_ARROW_Y_OFFSET);
			Vector2 stageCoordinates = tutorialArrowImage.localToStageCoordinates(new Vector2());
			InTheDarknessGame.GameLayer.OVERLAY_LAYER.getLayerEntity().stageToLocalCoordinates(stageCoordinates);
			tutorialArrowImage.setPosition(stageCoordinates.x, stageCoordinates.y);
			InTheDarknessGame.GameLayer.OVERLAY_LAYER.getLayerEntity().addActor(tutorialArrowImage);
			tutorialArrowImage.setVisible(true);
		}
	}
	
	private void tryHideTutorialArrowSprite() {
		if (tutorialArrowImage != null) {
			tutorialArrowImage.remove();
		}
	}
}
