/**
 * ****************************************************************************
 * Tablexia
 * <p/>
 * Copyright (C) 2013  CZ NIC z.s.p.o. <podpora at nic dot cz>
 * <p/>
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p/>
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * <p/>
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************
 */

package cz.nic.tablexia.game.games.kidnapping.media;

import com.badlogic.gdx.audio.Music;

import java.io.Serializable;
import java.util.Locale;
import java.util.Random;

import cz.nic.tablexia.TablexiaSettings;
import cz.nic.tablexia.game.difficulty.GameDifficulty;
import cz.nic.tablexia.game.games.kidnapping.KidnappingGame;
import cz.nic.tablexia.game.games.kidnapping.model.Direction;
import cz.nic.tablexia.util.Log;

/**
 * @author lhoracek
 */
public class DirectionSounds {
    public static SoundPack getRandomSoundPack(Random random, GameDifficulty gameDifficulty, Locale locale, Direction fromDirection, Direction nextDirection) {
        if (gameDifficulty == GameDifficulty.EASY) {
            return Level1.getRandomSoundPack(random, fromDirection, nextDirection, locale);
        } else if (gameDifficulty == GameDifficulty.MEDIUM) {
            return Level2.getRandomSoundPack(random, fromDirection, nextDirection, locale);
        } else if (gameDifficulty == GameDifficulty.HARD) {
            return Level3.getRandomSoundPack(random, fromDirection, nextDirection, locale);
        }
        throw new IllegalArgumentException("Unsupported difficulty " + gameDifficulty.name());
    }

    public static boolean isCorrectDirection(Direction direction, Direction correctDirection) {
        return direction == correctDirection;
    }

    public static class SoundPack implements Serializable {

        private static final long serialVersionUID = -4529735716574493739L;

        private Music[] musics;
        private Music exampleMusic;

        public SoundPack(int soundNumber, String example, String correct, String wrong1, String wrong2) {
            super();
            this.soundNumber = soundNumber;
            this.example = example;
            this.correct = correct;
            this.wrong1 = wrong1;
            this.wrong2 = wrong2;
        }

        private Integer soundNumber;

        private String example, correct, wrong1, wrong2;

        public String getCorrectName() {
            return correct;
        }

        public String getExampleName() {
            return example;
        }

        public String getWrong1Name() {
            return wrong1;
        }

        public String getWrong2Name() {
            return wrong2;
        }

        public Music getCorrect(KidnappingGame kidnappingGame) {
            return getMusic(kidnappingGame, correct);
        }

        public Music getExample(KidnappingGame kidnappingGame) {
            exampleMusic = getMusic(kidnappingGame, example);
            return exampleMusic;
        }

        public Music getWrong1(KidnappingGame kidnappingGame) {
            return getMusic(kidnappingGame, wrong1);
        }

        public Music getWrong2(KidnappingGame kidnappingGame) {
            return getMusic(kidnappingGame, wrong2);
        }

        private Music getMusic(KidnappingGame kidnappingGame, String music) {
            return kidnappingGame.getMusic("common/mfx." + music + ".mp3");
        }

        public Music[] getSoundsInOrder(KidnappingGame kidnappingGame, Direction[] directions, Direction next) {
            String[] sounds = getSoundNamesInOrder(directions, next);
            musics = new Music[3];
            for (int i = 0; i < 3; i++) {
                musics[i] = getMusic(kidnappingGame, sounds[i]);
            }
            return musics;
        }

        public String[] getSoundNamesInOrder(Direction[] directions, Direction next) {
            String[] sounds = new String[3];
            sounds[0] = (directions[0] == next) ? correct : wrong1;
            sounds[1] = (directions[0] == next) ? wrong1 : ((directions[1] == next) ? correct : wrong2);
            sounds[2] = (directions[2] == next) ? correct : wrong2;

            return sounds;
        }

        public Integer getSoundNumber() {
            return soundNumber;
        }

        public void disposeMusic() {
            if (musics != null) {
                for (Music music : musics) {
                    music.dispose();
                }
            }

            if (exampleMusic != null) {
                exampleMusic.dispose();
            }
        }

        @Override
        public String toString() {
            return "SoundPack{" +
                    "example='" + example + '\'' +
                    ", correct='" + correct + '\'' +
                    ", wrong1='" + wrong1 + '\'' +
                    ", wrong2='" + wrong2 + '\'' +
                    ", soundNumber=" + soundNumber +
                    '}';
        }
    }

    private enum LocaleSoundsCount {
        cs_CZ(TablexiaSettings.LocaleDefinition.cs_CZ.getLocale(),  20, 19, 31),
        sk_SK(TablexiaSettings.LocaleDefinition.sk_SK.getLocale(), 20, 19, 20),
        de_DE(TablexiaSettings.LocaleDefinition.de_DE.getLocale(), 20, 19, 20);

        Locale locale;
        int level1, level2, level3;

        LocaleSoundsCount(Locale locale, int level1, int level2, int level3) {
            this.locale = locale;
            this.level1 = level1;
            this.level2 = level2;
            this.level3 = level3;
        }

        private static LocaleSoundsCount getLocaleSoundCount(Locale locale) {
            for (LocaleSoundsCount localeSoundsCount : values()) {
                if(localeSoundsCount.locale.equals(locale)) {
                    return localeSoundsCount;
                }
            }
            Log.err(KidnappingGame.class, ": Currently chosen language can't be find in LocaleSoundCount Enum!");
            return null;
        }

        public static int getLevel1Count(Locale locale) {
            return getLocaleSoundCount(locale).level1;
        }

        public static int getLevel2Count(Locale locale) {
            return getLocaleSoundCount(locale).level2;
        }

        public static int getLevel3Count(Locale locale) {
            return getLocaleSoundCount(locale).level3;
        }
    }

    public static class Level1 {

        private static final String FOLDER_PREFIX = "level1/";

        public static SoundPack getRandomSoundPack(Random random, Direction fromDirection, Direction nextDirection, Locale locale) {
            int sounds = LocaleSoundsCount.getLevel1Count(locale);
            int num = random.nextInt(sounds) + 1;

            String correct = FOLDER_PREFIX + num + "-" + getRandomLetter(random);

            String wrong1 = null;
            do {
                wrong1 = FOLDER_PREFIX + num + "-" + getRandomLetter(random);
            } while (wrong1.equals(correct));

            String wrong2 = null;
            do {
                wrong2 = FOLDER_PREFIX + num + "-" + getRandomLetter(random);
            } while (wrong2.equals(correct) || wrong2.equals(wrong1));

            return new SoundPack(num, correct, correct, wrong1, wrong2);
        }
    }

    public static class Level2 {

        private static final String FOLDER_PREFIX = "level2/";

        public static SoundPack getRandomSoundPack(Random random, Direction fromDirection, Direction nextDirection, Locale locale) {
            int sounds = LocaleSoundsCount.getLevel2Count(locale);
            int num = random.nextInt(sounds) + 1;

            String correct = FOLDER_PREFIX + num + "-" + getRandomLetter(random);

            String wrong1 = null;
            do {
                wrong1 = FOLDER_PREFIX + num + "-" + getRandomLetter(random);
            } while (wrong1.equals(correct));

            String wrong2 = null;
            do {
                wrong2 = FOLDER_PREFIX + num + "-" + getRandomLetter(random);
            } while (wrong2.equals(correct) || wrong2.equals(wrong1));

            boolean previewSpelled = random.nextBoolean();

            if (previewSpelled) {
                return new SoundPack(num, correct + "-HL", correct, wrong1, wrong2);
            }
            return new SoundPack(num, correct, correct + "-HL", wrong1 + "-HL", wrong2 + "-HL");
        }
    }

    public static class Level3 {
        private static final String FOLDER_PREFIX = "level3/";

        public static SoundPack getRandomSoundPack(Random random, Direction fromDirection, Direction nextDirection, Locale locale) {
            int sounds = LocaleSoundsCount.getLevel3Count(locale);
            int num = random.nextInt(sounds) + 1;

            String example = FOLDER_PREFIX + num + "-VZ";
            String correct = FOLDER_PREFIX + num + "-A";

            String wrong1 = null;
            do {
                wrong1 = FOLDER_PREFIX + num + "-" + getRandomLetter(random);
            } while (wrong1.equals(correct));

            String wrong2 = null;
            do {
                wrong2 = FOLDER_PREFIX + num + "-" + getRandomLetter(random);
            } while (wrong2.equals(correct) || wrong2.equals(wrong1));

            return new SoundPack(num, example, correct, wrong1, wrong2);
        }

    }

    protected static String getRandomLetter(Random random) {
        int rand = random.nextInt(3);
        switch (rand) {
            case 1:
                return "A";
            case 2:
                return "B";
            default:
                return "C";
        }
    }

}
