package cz.nic.tablexia.game.games.pursuit;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.audio.Music;
import com.badlogic.gdx.graphics.Texture;
import com.badlogic.gdx.math.CatmullRomSpline;
import com.badlogic.gdx.math.Vector2;
import com.badlogic.gdx.scenes.scene2d.Action;
import com.badlogic.gdx.scenes.scene2d.Actor;
import com.badlogic.gdx.scenes.scene2d.Group;
import com.badlogic.gdx.scenes.scene2d.InputEvent;
import com.badlogic.gdx.scenes.scene2d.actions.Actions;
import com.badlogic.gdx.scenes.scene2d.actions.RunnableAction;
import com.badlogic.gdx.scenes.scene2d.ui.Image;
import com.badlogic.gdx.utils.Align;
import com.badlogic.gdx.utils.Array;
import com.badlogic.gdx.utils.Scaling;
import com.badlogic.gdx.utils.TimeUtils;
import com.badlogic.gdx.utils.viewport.ExtendViewport;
import com.badlogic.gdx.utils.viewport.Viewport;

import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicBoolean;

import cz.nic.tablexia.TablexiaApplication;
import cz.nic.tablexia.TablexiaSettings;
import cz.nic.tablexia.game.AbstractTablexiaGame;
import cz.nic.tablexia.game.GameDefinition;
import cz.nic.tablexia.game.difficulty.GameDifficulty;
import cz.nic.tablexia.game.games.pursuit.action.RotateAndMovePieceInPosition;
import cz.nic.tablexia.game.games.pursuit.action.RotatePieceToClosestAngle;
import cz.nic.tablexia.game.games.pursuit.action.VehicleMoveAction;
import cz.nic.tablexia.game.games.pursuit.assets.PursuitAssets;
import cz.nic.tablexia.game.games.pursuit.helper.ArithmeticsHelper;
import cz.nic.tablexia.game.games.pursuit.helper.GameRulesHelper;
import cz.nic.tablexia.game.games.pursuit.helper.TextureHelper;
import cz.nic.tablexia.game.games.pursuit.model.Grid;
import cz.nic.tablexia.game.games.pursuit.model.Maps;
import cz.nic.tablexia.game.games.pursuit.model.PuzzlePiece;
import cz.nic.tablexia.game.games.pursuit.model.Vehicle;
import cz.nic.tablexia.game.games.pursuit.model.VehicleType;
import cz.nic.tablexia.loader.TablexiaAbstractFileManager;
import cz.nic.tablexia.loader.TablexiaTextureManager;
import cz.nic.tablexia.loader.application.ApplicationTextManager;
import cz.nic.tablexia.model.game.GameDAO;
import cz.nic.tablexia.shared.model.Game;
import cz.nic.tablexia.util.Log;
import cz.nic.tablexia.util.Point;
import cz.nic.tablexia.util.ScaleUtil;
import cz.nic.tablexia.util.Utility;
import cz.nic.tablexia.util.listener.DragAndRotateActorListener;
import cz.nic.tablexia.util.ui.AnimatedImage;
import cz.nic.tablexia.util.ui.TablexiaNoBlendingImage;
import cz.nic.tablexia.util.ui.dialog.components.AnimatedImageContentDialogComponent;
import cz.nic.tablexia.util.ui.dialog.components.TablexiaDialogComponentAdapter;
import cz.nic.tablexia.util.ui.dialog.components.TextContentDialogComponent;
import cz.nic.tablexia.util.ui.dialog.components.TwoColumnContentDialogComponent;

/**
 * Created by Václav Tarantík on 6.3.15.
 */
public class PursuitGame extends AbstractTablexiaGame<int[][]> {

    public static class GameResultResolver implements GameDefinition.GameResultResolver {
        @Override
        public GameResult getGameCupsResult(Game game) {
            return GameRulesHelper.getNumberOfStarsForTime(GameDifficulty.getGameDifficultyForDifficultyNumber(game.getGameDifficulty()), GameDAO.getGameDuration(game));
        }
        @Override
        public float getGameScoreResult(Game game) {
             return GameDAO.getGameDuration(game) != null ? (float)GameDAO.getGameDuration(game) : 0;
        }

	}
    private enum ResultMapping {

        NO_STAR_TEXT(GameResult.NO_STAR, PursuitAssets.VICTORYTEXT_NOSTAR, PursuitAssets.VICTORYSPEECH_NOSTAR),
        ONE_STAR_TEXT(GameResult.ONE_STAR, PursuitAssets.VICTORYTEXT_ONESTAR, PursuitAssets.VICTORYSPEECH_ONESTAR),
        TWO_STAR_TEXT(GameResult.TWO_STAR, PursuitAssets.VICTORYTEXT_TWOSTARS, PursuitAssets.VICTORYSPEECH_TWOSTAR),
        THREE_STAR_TEXT(GameResult.THREE_STAR, PursuitAssets.VICTORYTEXT_THREESTARS, PursuitAssets.VICTORYSPEECH_THREESTAR);

        private final GameResult gameResult;
        private final String textKey;
        private final String soundName;

        ResultMapping(GameResult gameResult, String textKey, String soundName) {
            this.gameResult = gameResult;
            this.textKey = textKey;
            this.soundName = soundName;
        }

        public String getTextKey() {
            return textKey;
        }

        public String getSoundName() {
            return soundName;
        }

        public static ResultMapping getResultTextMappingForGameResult(GameResult gameResult) {
            for (ResultMapping resultMapping : ResultMapping.values()) {
                if (resultMapping.gameResult.equals(gameResult)) {
                    return resultMapping;
                }
            }
            return null;
        }
    }


    private static final int GRID_PADDING = 10;

    private static final String SUMMARY_TEXT_TIME_HOURS = "victory_text_hours";
    private static final String SUMMARY_TEXT_TIME_MINUTES = "victory_text_minutes";
    private static final String SUMMARY_TEXT_TIME_SECONDS = "victory_text_seconds";
    private static final String SUMMARY_TEXT_MOVE_COUNT = "victory_text_moves";
    private static final String SUMMARY_TEXT_TIME = "victory_text_time";

    private static final String SCORE_KEY_MOVE_COUNT = "move_count";
	private static final String SCORE_KEY_LAST_MAP   = "last_map_index";

	private static final float OVERLAP_PIECE_SCALE	=	1.2f;
    
	private static final float ANIMATION_MOVING_DURATION =	5f;
	private static final float ANIMATION_DELAY_DURATION  =	1.5f;

    private static final int MAPS_COUNT = Maps.values().length;

	private static final int MIN_ANGLE_TO_PLAY_ROTATE_SOUND   = 45;  //degrees
	private static final int ROTATE_SOUND_MIN_INTERVAL        = 420; //ms

    private Group contentGroup;

    private Grid grid;
    private int gridColumnCount;
    private int overlappedPieceZIndex;
    private TablexiaNoBlendingImage backgroundImage;
    private DragAndRotateActorListener dragAndRotateActorListener;

	private int lastCirclePart = 0;
	private long lastTimeRotateSoundPlayed;

    private int movesCounter;
    private int mapNumber = 0;

    private AtomicBoolean gameFinished = new AtomicBoolean(false);

    private Vehicle vehicle;
    private Image finishFlag;
    private Texture roadMap;

    private TablexiaTextureManager textureManager;
    private String clickMapTextureName;
    private String mapTextureName;

    @Override
    protected Viewport createViewport() {
        return new ExtendViewport(0, TablexiaSettings.getMinWorldHeight());
    }

	@Override
	protected void prepareGameSoundAssetNames(List<String> soundsFileNames) {
		//Car Sounds
		for(String carSound : PursuitAssets.SOUNDS_CAR) {
			soundsFileNames.add(carSound);
		}
		//Train Sounds
		for(String trainSound : PursuitAssets.SOUNDS_TRAIN) {
			soundsFileNames.add(trainSound);
		}
		//Boat Sounds
		for(String boatSound : PursuitAssets.SOUNDS_BOAT) {
			soundsFileNames.add(boatSound);
		}
		//Motorcycle Sounds
		for(String motorcycleSound : PursuitAssets.SOUNDS_MOTORCYCLE) {
			soundsFileNames.add(motorcycleSound);
		}

		//Pieces Sounds
		soundsFileNames.add(PursuitAssets.SOUND_LIFT);
		soundsFileNames.add(PursuitAssets.SOUND_ROTATION);
		soundsFileNames.add(PursuitAssets.SOUND_SWITCH);
	}

    @Override
    protected int[][] prepareGameData(Map<String, String> gameState) {
        final Object loaderLock = new Object();

        updateMapIndex();
        mapTextureName = prepareScreenAssetsPath(prepareScreenName()) + TextureHelper.getMapFullName(mapNumber);
        clickMapTextureName = prepareScreenAssetsPath(prepareScreenName()) + TextureHelper.getClickMapPath(mapNumber);

        synchronized (loaderLock) {

            Gdx.app.postRunnable(new Runnable() {
                @Override
                public void run() {
                    synchronized (loaderLock) {
                        tryToDisposeTextureManager();
                        textureManager = new TablexiaTextureManager(TablexiaAbstractFileManager.AssetsStorageType.EXTERNAL);
                        textureManager.loadTexture(mapTextureName);
                        textureManager.loadTexture(clickMapTextureName);
                        textureManager.finishLoading();
                        loaderLock.notify();
                    }
                }
            });

            try {
                loaderLock.wait();
            } catch (InterruptedException e) {
                Log.err(getClass(), "Error while waiting to texture loader!", e);
            }

            roadMap = textureManager.getTexture(clickMapTextureName);
            if (!roadMap.getTextureData().isPrepared()) {
                roadMap.getTextureData().prepare();
            }
        }
        return Utility.createColorMap(roadMap);
    }

    private void tryToDisposeTextureManager() {
        if (textureManager != null) {
            textureManager.dispose();
            textureManager = null;
        }
    }

    @Override
    protected void screenResized(int width, int height) {
        ScaleUtil.setFullScreen(contentGroup, getStage());
        float minSide = Math.min(getViewportWidth(), getViewportHeight());
        grid.setSize(minSide - 2 * GRID_PADDING, minSide - 2 * GRID_PADDING);
        grid.setPosition(getViewportWidth() / 2 - grid.getWidth() / 2, (getViewportHeight() / 2 - grid.getHeight() / 2));
        grid.scaleGrid();

    }

    @Override
    protected void gameLoaded(Map<String, String> gameState) {
        gameFinished.set(false);
        movesCounter = 0;
        gridColumnCount = GameRulesHelper.getNumberOfColumns(getGameDifficulty());
        overlappedPieceZIndex = (int) Math.pow(gridColumnCount, 2) - 1;
        contentGroup = new Group();
        ScaleUtil.setFullScreen(contentGroup, getStage());
        getStage().addActor(contentGroup);

		prepareBackground();
		prepareGrid();
		prepareFinishFlag();
		prepareVehicle();

        dragAndRotateActorListener = new DragAndRotateActorListener(new DragAndRotateActorListener.IOnRotationFinished() {
            @Override
            public void onDragAndRotationFinished() {
                Log.info(getClass(), "DRAG AND ROTATION FINISHED");
                dragAndRotateActorListener.setPieceAnimationDone(true);
                setGameScore(SCORE_KEY_MOVE_COUNT, movesCounter++);
            }
        }) {
            @Override
            public void selected(float x, float y) {
                if (dragAndRotateActorListener != null) {
                    if (dragAndRotateActorListener.getDraggedActor() == null) {
						getSound(PursuitAssets.SOUND_LIFT).play();
                        dragAndRotateActorListener.setDraggedActor(grid.pieceAtPoint(x, y));
                    }
                }
            }

            @Override
            public void moved(float x, float y, Actor draggedPiece) {
                if (dragAndRotateActorListener != null) {
                    PuzzlePiece puzzlePiece = (PuzzlePiece) draggedPiece;
                    PuzzlePiece overlappedPiece = grid.pieceIntersectingWithDragged(puzzlePiece.getId());
                    if (overlappedPiece != null) {
                        overlappedPiece.setZIndex(overlappedPieceZIndex);
                        overlappedPiece.setScale(OVERLAP_PIECE_SCALE);
                        grid.unscalePieces(puzzlePiece.getId(), overlappedPiece.getId());
                    } else {
                        grid.unscalePieces(puzzlePiece.getId(), -1);
                    }
                }
            }

			@Override
			public void rotated(float angle, Actor rotatedPiece) {
				int circlePart = (int)(angle) / MIN_ANGLE_TO_PLAY_ROTATE_SOUND;

				if (lastCirclePart != circlePart && System.currentTimeMillis() - lastTimeRotateSoundPlayed > ROTATE_SOUND_MIN_INTERVAL) {
					getSound(PursuitAssets.SOUND_ROTATION).play();
					lastTimeRotateSoundPlayed = System.currentTimeMillis();
					lastCirclePart = circlePart;
				}
			}

			@Override
			public void touchUp(InputEvent event, float x, float y, int pointer, int button) {
				super.touchUp(event, x, y, pointer, button);
			}

			@Override
			public boolean touchDown(InputEvent event, float x, float y, int pointer, int button) {
				lastCirclePart = 0;
				return super.touchDown(event, x, y, pointer, button);
			}

			@Override
            public void dropped(float x, float y, final Actor draggedPiece) {
                if (dragAndRotateActorListener != null) {
                    if (draggedPiece != null) {
                        dragAndRotateActorListener.setPerformingAction(true);
                        final PuzzlePiece puzzlePiece = (PuzzlePiece) draggedPiece;
                        final PuzzlePiece overlappedPiece = grid.pieceIntersectingWithDragged(puzzlePiece.getId());
                        Action onPiecesSwitchFinishedAction = new Action() {
                            @Override
                            public boolean act(float delta) {
                                setGameScore(SCORE_KEY_MOVE_COUNT, movesCounter++);
                                dragAndRotateActorListener.setDraggedActor(null);
                                dragAndRotateActorListener.setPerformingAction(false);
                                grid.unscaleAllPieces();
                                //cancelling all actions on actor that would take place after piece was moved in correct position
                                Array<Action> gridActions = grid.getActions();
                                if (gridActions.size > 0) {
                                    for (Action action : gridActions) {
                                        if (action instanceof RotatePieceToClosestAngle || action instanceof RotateAndMovePieceInPosition) {
                                            grid.removeAction(action);
                                        }
                                    }
                                    //in case we canceled rotation animation, rotate piece to correct angle
                                    grid.addAction(Actions.sequence(new RotatePieceToClosestAngle((PuzzlePiece) draggedPiece), Actions.run(new Runnable() {
										@Override
										public void run() {
											// check if puzzle is solved after possible rotation animation
											checkEndGameCondition();
										}
									})));
                                }
								checkEndGameCondition();
                                return true;
                            }
                        };
                        if (grid.switchPieces(puzzlePiece, overlappedPiece, onPiecesSwitchFinishedAction)) {
							getSound(PursuitAssets.SOUND_SWITCH).play();
						}
                    }
                }
            }
        };
        dragAndRotateActorListener.setParentActor(grid);
        backgroundImage.addListener(dragAndRotateActorListener);
    }

	private void checkEndGameCondition() {
		if (grid.allPiecesInCorrectPositions() && !gameFinished.get()) {
			backgroundImage.removeListener(dragAndRotateActorListener);
			showCheckpoints();
            gameFinished.set(true);
		}
	}

    @Override
    protected void gameDisposed() {
        tryToDisposeTextureManager();
        super.gameDisposed();
    }

	private void updateMapIndex() {
		String lastMap = GameDAO.getLastGameScoreValueForGameAndKey(GameDefinition.PURSUIT.getGameNumber(), SCORE_KEY_LAST_MAP);

		//couldn't find the last map
		if (lastMap == null) {
			mapNumber = 0;
		} else {
            //To loop over maps
            mapNumber = Integer.parseInt(lastMap);
            mapNumber = (mapNumber + 1) % MAPS_COUNT;
		}

		//Update last map index in db
		setGameScore(SCORE_KEY_LAST_MAP, mapNumber);
	}

    private void prepareBackground() {
        backgroundImage = new TablexiaNoBlendingImage(getScreenTextureRegion(TextureHelper.getBgTexturePath()));
        backgroundImage.setFillParent(true);
        contentGroup.addActor(backgroundImage);

    }

    private void prepareGrid() {
        Texture mapTextureRegion = textureManager.getTexture(mapTextureName);
        mapTextureRegion.setFilter(Texture.TextureFilter.Linear, Texture.TextureFilter.Linear);
        float minSide = Math.min(getViewportWidth(),getViewportHeight());
        grid = new Grid(((getViewportWidth() / 2) - minSide / 2) + GRID_PADDING,(getViewportHeight() / 2 - minSide / 2) + GRID_PADDING, minSide- 2 * GRID_PADDING, gridColumnCount, mapTextureRegion, getRandom());
        contentGroup.addActor(grid);
    }

    private void prepareVehicle() {
        VehicleType vehicleType = TextureHelper.getVehicleTextureName(getRandom(), mapNumber);
        final Point vehiclePosition = new Point(0, 0);
        Point vehicleSize = new Point(vehicleType.getVehicleSize().x, vehicleType.getVehicleSize().y);
        vehicle = new Vehicle(getScreenTextureRegion(vehicleType.getTextureName()), vehiclePosition, vehicleSize);
        vehicle.setDebug(TablexiaSettings.getInstance().isShowBoundingBoxes());
        vehicle.setVisible(false);
        grid.addActor(vehicle);
	}

	private void prepareFinishFlag() {
        finishFlag = new Image(getScreenTextureRegion(TextureHelper.getFinishFlagTextureName()));
		finishFlag.setSize(TextureHelper.FINISH_FLAG_SIZE.x, TextureHelper.FINISH_FLAG_SIZE.y);
        finishFlag.setOrigin(finishFlag.getWidth() / 2, finishFlag.getHeight() / 2);
        finishFlag.setDebug(TablexiaSettings.getInstance().isShowBoundingBoxes());
        finishFlag.setVisible(false);
        grid.addActor(finishFlag);

    }

	/**
	 * Returns all Sounds (instances of Music) for vehicleType
	 * @param vehicleType
	 * @return Music vehicle sounds
	 */
	private Music[] getVehicleSounds(VehicleType vehicleType) {
		//Using music, so i can use OnCompletionListener
		Music[] result;
		String[] soundsToLoad;

		switch (vehicleType) {
			case CAR:
				soundsToLoad = PursuitAssets.SOUNDS_CAR;
				break;
			case MOTORCYCLE:
				soundsToLoad = PursuitAssets.SOUNDS_MOTORCYCLE;
				break;
			case TRAIN:
				soundsToLoad = PursuitAssets.SOUNDS_TRAIN;
				break;
			case BOAT:
				soundsToLoad = PursuitAssets.SOUNDS_BOAT;
				break;
			default:
				throw new IllegalArgumentException(this.getClass().getSimpleName() + ": Can't get sounds for vehicleType: " + vehicleType);
		}

		result = new Music[soundsToLoad.length];

		for(int i = 0; i < soundsToLoad.length; i++) {
			result[i] = getMusic(soundsToLoad[i]);
		}

		return result;
	}

	//sets the position according to completed map rotation
    private void showCheckpoints() {
        Point gridRotationCenter = new Point(grid.getWidth() / 2, grid.getHeight() / 2);
        Log.debug(getClass(),"grid rotation: " + grid.getRotation());
        //flag
        Point flagPosition = TextureHelper.getFinishFlagPosition(grid.getWidth(), mapNumber);
        Point flagPositonConsideringMapRotation = ArithmeticsHelper.translatePositionAccordingToRotation(flagPosition, grid.getRotation(), gridRotationCenter);
        //positioning both flag and vehicle by center
        //flag is not rotated
        Point flagPoleOrigin = new Point(finishFlag.getWidth() * TextureHelper.FLAG_BLACK_DOT_RELATIVE_POSITION.x, finishFlag.getHeight() * TextureHelper.FLAG_BLACK_DOT_RELATIVE_POSITION.y);
        finishFlag.setPosition(flagPositonConsideringMapRotation.x - flagPoleOrigin.x, flagPositonConsideringMapRotation.y - flagPoleOrigin.y);

        //vehicle
        Point vehiclePosition = TextureHelper.getVehicleStartPosition(grid.getWidth(), mapNumber);
        Point vehiclePositionConsideringRotation = ArithmeticsHelper.translatePositionAccordingToRotation(vehiclePosition, grid.getRotation(), gridRotationCenter);
        vehicle.setPosition(vehiclePositionConsideringRotation.x - vehicle.getWidth() / 2, vehiclePositionConsideringRotation.y - vehicle.getHeight() / 2);
        vehicle.setRotation(grid.getRotation());

        finishFlag.setVisible(true);

        final Music[] vehicleSounds = getVehicleSounds( TextureHelper.getVehicleTextureName(getRandom(), mapNumber));
        final CatmullRomSpline<Vector2> path = new CatmullRomSpline<Vector2>(Maps.values()[mapNumber].getPursuitPositionDefinition().getPointsWithRotation(grid.getRotation(), vehiclePosition.x,vehiclePosition.y),false);
        VehicleMoveAction vehicleMoveAlongAction = new VehicleMoveAction(path, ANIMATION_MOVING_DURATION, grid.getWidth(), vehicleSounds);
        vehicle.addAction(Actions.sequence(Actions.show(),Actions.delay(ANIMATION_DELAY_DURATION),vehicleMoveAlongAction,new RunnableAction() {
            @Override
            public void run() {
                for(Music snd : vehicleSounds) {
                    snd.stop();
                    snd.dispose();
                }
                endGame();
                showGameResultDialog();
                vehicle.clear();
                vehicle.remove();
                finishFlag.clear();
                finishFlag.remove();
            }
        }));
    }

    @Override
    protected List<SummaryMessage> getSummaryMessageForGameResult(Game game) {
        return Arrays.asList(new SummaryMessage(SummaryImage.TIME, getFormattedText(SUMMARY_TEXT_TIME, getOverallTimeText(GameDAO.getGameDuration(game)))),
                new SummaryMessage(SummaryImage.STATS, getFormattedText(SUMMARY_TEXT_MOVE_COUNT, game.getGameScore(SCORE_KEY_MOVE_COUNT, "0"))));
    }

    @Override
    protected String getTextKeyForGameResult(GameResult gameResult) {
        return ResultMapping.getResultTextMappingForGameResult(gameResult).getTextKey();
    }

    @Override
    protected String getSoundNameForGameResult(GameResult gameResult) {
        return ResultMapping.getResultTextMappingForGameResult(gameResult).getSoundName();
    }

    private String getOverallTimeText(float millis) {
        String time="";
        StringBuilder sb = new StringBuilder();
        int seconds = (int) (millis / 1000) % 60 ;
        int minutes = (int) ((millis / (1000*60)) % 60);
        int hours   = (int) ((millis / (1000*60*60)) % 24);
        Log.info(getClass(), "TIME: " + hours + "h, " + minutes + "m, " + seconds + "s");
        if (hours != 0) {
            Log.info(getClass(), "HOURS");
            sb.append(getFormattedText(SUMMARY_TEXT_TIME_HOURS, hours));
        }
        if (minutes != 0) {
            Log.info(getClass(), "MINUTES");
            sb.append(getFormattedText(SUMMARY_TEXT_TIME_MINUTES, minutes));
        }
        if (seconds != 0) {
            Log.info(getClass(), "SECONDS");
            sb.append(getFormattedText(SUMMARY_TEXT_TIME_SECONDS, seconds));
        }
        return sb.toString();
    }


//////////////////////////////////////////// PRELOADER

    private static final String                             PRELOADER_ANIM_IMAGE            = "preloader_anim";
    private static final int                                PRELOADER_ANIM_FRAMES           = 14;
    private static final float                              PRELOADER_ANIM_FRAME_DURATION   = 0.5f;
    private static final String                             PRELOADER_TEXT_KEY             	= ApplicationTextManager.ApplicationTextsAssets.GAME_PURSUIT_PRELOADER_TEXT;

    private static final Scaling        PRELOADER_IMAGE_SCALING 		    = Scaling.fit;
    private static final int 			PRELOADER_TEXT_ALIGN 				= Align.left;
    private static final float 			PRELOADER_TEXT_PADDING 				= 10f;
    private static final float 			PRELOADER_LEFT_COLUMN_WIDTH_RATIO 	= 1f / 4;
    private static final float 			PRELOADER_RIGHT_COLUMN_RATIO 		= 5f / 8;
    private static final float 			PRELOADER_ROW_HEIGHT 				= 1f / 3;

    @Override
    public void preparePreloaderContent(float width, float height, TablexiaApplication.PreloaderAssetsManager preloaderAssetsManager, List<TablexiaDialogComponentAdapter> components) {
        AnimatedImage preloaderImage = new AnimatedImage(preloaderAssetsManager.getAnimation(PRELOADER_ANIM_IMAGE, PRELOADER_ANIM_FRAMES, PRELOADER_ANIM_FRAME_DURATION), false);
        preloaderImage.startAnimationLoop();
        String preloaderText = preloaderAssetsManager.getText(PRELOADER_TEXT_KEY);
        components.add(new TwoColumnContentDialogComponent(
                new AnimatedImageContentDialogComponent(preloaderImage, PRELOADER_IMAGE_SCALING),
                new TextContentDialogComponent(preloaderText, Integer.valueOf(PRELOADER_TEXT_ALIGN), Float.valueOf(PRELOADER_TEXT_PADDING)),
                PRELOADER_LEFT_COLUMN_WIDTH_RATIO,
                PRELOADER_RIGHT_COLUMN_RATIO,
                PRELOADER_ROW_HEIGHT));
    }

//////////////////////////////////////////// DEBUG - FORCE GAME END

    @Override
    protected void forceGameEnd(GameResult result) {
        backgroundImage.removeListener(dragAndRotateActorListener);

        long duration = GameRulesHelper.getDurationForStars(GameDifficulty.getByGame(getGame()), result.getStarCount());
        GameDAO.startGame(getGame(), TimeUtils.millis() - duration);
        GameDAO.setGameScore(getGame(), SCORE_KEY_MOVE_COUNT, Integer.toString(30));
        waitForActualGameScore();

        endGame();
        gameFinished.set(true);

        showGameResultDialog();
    }
}
