package cz.nic.tablexia.loader;

import com.badlogic.gdx.assets.AssetManager;
import com.badlogic.gdx.assets.loaders.FileHandleResolver;
import com.badlogic.gdx.assets.loaders.resolvers.InternalFileHandleResolver;
import com.badlogic.gdx.assets.loaders.resolvers.LocalFileHandleResolver;
import com.badlogic.gdx.files.FileHandle;

import cz.nic.tablexia.TablexiaSettings;
import cz.nic.tablexia.util.Log;

public abstract class TablexiaAbstractFileManager extends AssetManager {

    private enum BuildTypeSuffix {

        DEBUG(TablexiaSettings.BuildType.DEBUG),
        DEVEL(TablexiaSettings.BuildType.DEVEL);

        private final TablexiaSettings.BuildType buildType;

        BuildTypeSuffix(TablexiaSettings.BuildType buildType) {
            this.buildType = buildType;
        }

        public String getSuffix() {
            return "_" + buildType.getKey() + "/";
        }

        public static String getSuffixForBuildType(TablexiaSettings.BuildType buildType) {
            for (BuildTypeSuffix buildTypeSuffix: BuildTypeSuffix.values()) {
                if (buildTypeSuffix.buildType == buildType) {
                    return buildTypeSuffix.getSuffix();
                }
            }
            return "/";
        }
    }

    public interface StorageType {
        String getStoragePath();
        FileHandleResolver getResolver();
    }

    public enum RootStorageType implements StorageType {

        INTERNAL("", false, new InternalFileHandleResolver()),
        LOCAL(RootStorageType.TABLEXIA_DIRECTORY, true, new LocalFileHandleResolver());

        public static final String TABLEXIA_DIRECTORY = ".tablexia";

        private String storagePath;
        private boolean useBuildTypeSuffix;
        private FileHandleResolver resolver;

        RootStorageType(String storagePath, boolean useBuildTypeSuffix, FileHandleResolver resolver) {
            this.storagePath = storagePath;
            this.useBuildTypeSuffix = useBuildTypeSuffix;
            this.resolver = resolver;
        }

        public String getStoragePath() {
            return storagePath + (useBuildTypeSuffix ? BuildTypeSuffix.getSuffixForBuildType(TablexiaSettings.getInstance().getBuildType()) : "");
        }

        public FileHandleResolver getResolver() {
            return resolver;
        }
    }

    public enum AssetsStorageType implements StorageType {

        INTERNAL(RootStorageType.INTERNAL, ""),
        EXTERNAL(RootStorageType.LOCAL, AssetsStorageType.ASSETS_DIRECTORY);

        public static final String ASSETS_DIRECTORY = "assets/";

        private String storagePath;
        private FileHandleResolver resolver;

        AssetsStorageType(RootStorageType rootStorageType, String assetsPath) {
            this.storagePath = rootStorageType.getStoragePath() + assetsPath;
            this.resolver = rootStorageType.getResolver();
        }

        public String getStoragePath() {
            return storagePath;
        }

        public FileHandleResolver getResolver() {
            return resolver;
        }
    }

    public enum DataStorageType implements StorageType {

        INTERNAL(RootStorageType.INTERNAL, DataStorageType.DATA_DIRECTORY),
        EXTERNAL(RootStorageType.LOCAL, DataStorageType.DATA_DIRECTORY);

        public static final String DATA_DIRECTORY = "data/";

        private String storagePath;
        private FileHandleResolver resolver;

        DataStorageType(RootStorageType rootStorageType, String assetsPath) {
            this.storagePath = rootStorageType.getStoragePath() + assetsPath;
            this.resolver = rootStorageType.getResolver();
        }

        public String getStoragePath() {
            return storagePath;
        }

        public FileHandleResolver getResolver() {
            return resolver;
        }
    }

    public enum DownloadStorageType implements StorageType {

        INTERNAL(RootStorageType.INTERNAL, DownloadStorageType.DOWNLOAD_DIRECTORY),
        EXTERNAL(RootStorageType.LOCAL, DownloadStorageType.DOWNLOAD_DIRECTORY);

        public static final String DOWNLOAD_DIRECTORY = "download/";

        private String storagePath;
        private FileHandleResolver resolver;

        DownloadStorageType(RootStorageType rootStorageType, String assetsPath) {
            this.storagePath = rootStorageType.getStoragePath() + assetsPath;
            this.resolver = rootStorageType.getResolver();
        }

        public String getStoragePath() {
            return storagePath;
        }

        public FileHandleResolver getResolver() {
            return resolver;
        }
    }

    public enum KeystoreStorageType implements StorageType {

        INTERNAL(RootStorageType.INTERNAL, KeystoreStorageType.KEYSTORE_DIRECTORY),
        EXTERNAL(RootStorageType.LOCAL, KeystoreStorageType.KEYSTORE_DIRECTORY);

        public static final String KEYSTORE_DIRECTORY = "keystore/";

        private String storagePath;
        private FileHandleResolver resolver;

        KeystoreStorageType(RootStorageType rootStorageType, String assetsPath) {
            this.storagePath = rootStorageType.getStoragePath() + assetsPath;
            this.resolver = rootStorageType.getResolver();
        }

        public String getStoragePath() {
            return storagePath;
        }

        public FileHandleResolver getResolver() {
            return resolver;
        }
    }

    public enum ReportStorageType implements StorageType {

        INTERNAL(RootStorageType.INTERNAL, ReportStorageType.REPORTS_DIRECTORY),
        EXTERNAL(RootStorageType.LOCAL, ReportStorageType.REPORTS_DIRECTORY);

        public static final String REPORTS_DIRECTORY = "reports/";

        private String storagePath;
        private FileHandleResolver resolver;

        ReportStorageType(RootStorageType rootStorageType, String assetsPath) {
            this.storagePath = rootStorageType.getStoragePath() + assetsPath;
            this.resolver = rootStorageType.getResolver();
        }

        public String getStoragePath() {
            return storagePath;
        }

        public FileHandleResolver getResolver() {
            return resolver;
        }
    }

    public static FileHandle getFileStoragePathFileHandle(StorageType storageType, String additionalPath) {
        return storageType.getResolver().resolve(storageType.getStoragePath() + additionalPath);
    }

    public static FileHandle getFileStoragePathFileHandle(StorageType storageType) {
        return getFileStoragePathFileHandle(storageType, "");
    }

    public static String getFileStoragePath(StorageType storageType, String additionalPath) {
        return getFileStoragePathFileHandle(storageType, additionalPath).file().getAbsolutePath();
    }

    public static String getFileStoragePath(StorageType storageType) {
        return getFileStoragePathFileHandle(storageType).file().getAbsolutePath();
    }

    private final StorageType storageType;

    public TablexiaAbstractFileManager(StorageType storageType) {
        super(storageType.getResolver());
        this.storageType = storageType;
    }

    public <T> void loadAsset(String assetFileName, Class<T> clazz) {
        loadAsset(assetFileName, clazz, true);
    }

    public <T> void loadAsset(String assetFileName, Class<T> clazz, boolean checkFileExistence) {
        if (assetFileName != null) {
            String filePath = storageType.getStoragePath() + assetFileName;
            if (!checkFileExistence || storageType.getResolver().resolve(filePath).exists()) {
                load(filePath, clazz);
                return;
            }
        }
        Log.info(getClass(), "Asset file for name: " + assetFileName + " was not found!");
    }

    public <T> T getAsset(String assetName, Class<T> clazz) {
        return get(storageType.getStoragePath() + assetName, clazz);
    }

    @Override
    public synchronized boolean isLoaded (String fileName, Class type) {
        return super.isLoaded(storageType.getStoragePath() + fileName, type);
    }
}
