package cz.nic.tablexia.loader.application;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.graphics.Texture;
import com.badlogic.gdx.graphics.g2d.DistanceFieldFont;
import com.badlogic.gdx.graphics.g2d.TextureRegion;
import com.badlogic.gdx.graphics.glutils.ShaderProgram;

import cz.nic.tablexia.loader.IApplicationLoader;

/**
 * Font loader and manager class for application context
 * 
 * @author Matyáš Latner
 *
 */
public class ApplicationFontManager implements IApplicationLoader {

	public enum FontType_NEW {
		REGULAR_10  (10, false),
		REGULAR_12  (12, false),
		REGULAR_14  (14, false),
		REGULAR_16  (16, false),
		REGULAR_18  (18, false),
		REGULAR_20  (20, false),
		REGULAR_26  (26, false),
		REGULAR_30  (30, false),
		REGULAR_35  (35, false),
		REGULAR_40  (40, false),
		REGULAR_45  (45, false),
		REGULAR_50  (50, false),
		REGULAR_55  (55, false),

		BOLD_10  (10, true),
		BOLD_12  (12, true),
		BOLD_14  (14, true),
		BOLD_16  (16, true),
		BOLD_18  (18, true),
		BOLD_20  (20, true),
		BOLD_26  (26, true),
		BOLD_30  (30, true),
		BOLD_35  (35, true),
		BOLD_40  (40, true),
		BOLD_45  (45, true),
		BOLD_50  (50, true),
		BOLD_55  (55, true),
		BOLD_65  (65, true);

		int size;
		boolean bold;

		FontType_NEW(int size, boolean bold) {
			this.size = size;
			this.bold = bold;
		}

		public static FontType_NEW getProximateFontTypeForSize(int size, boolean bold) {
			FontType_NEW result = null;
			for (FontType_NEW fontType : FontType_NEW.values()) {
				if (result == null || (fontType.bold == bold && Math.abs(fontType.size - size) < Math.abs(result.size - size))) {
					result = fontType;
				}
			}
			return result;
		}

		public int getSize() {
			return size;
		}

		public boolean isBold() {
			return bold;
		}

		public float getScale() {
			return this.size/DISTANCE_FIELD_FONT_SIZE;
		}
	}

	private static final String		ROBOTO_REGULAR_FONT_FILE = "font/roboto-regular";
	private static final String		ROBOTO_BOLD_FONT_FILE    = "font/roboto-bold";
	private static final String 	FONT_FILE_EXTENSION		 = ".fnt";
	private static final String 	FONT_BITMAP_EXTENSION	 = ".png";

	public static final float		DISTANCE_FIELD_FONT_SIZE = 32;

	//According to (https://github.com/libgdx/libgdx/wiki/Distance-field-fonts) it should be 8/scale... I feel like breaking the rules a little bit :/
	public static final float		DISTANCE_FIELD_FONT_DEFAULT_SMOOTHING = 6;

	private static ApplicationFontManager instance;

	private DistanceFieldFont robotoRegularFont;
	private DistanceFieldFont robotoBoldFont;

	private ShaderProgram distanceFieldShader;
	

	private ApplicationFontManager() {
	}
	
	public static ApplicationFontManager getInstance() {
		if (instance == null) {
			instance = new ApplicationFontManager();
		}
		return instance;
	}

	public ShaderProgram getDistanceFieldShader() {
		return distanceFieldShader;
	}

	public DistanceFieldFont getDistanceFieldFont(FontType_NEW fontType) {
		return fontType.isBold() ? robotoBoldFont : robotoRegularFont;
	}
	
	public void load() {
		Texture robotoRegular = new Texture(Gdx.files.internal(ROBOTO_REGULAR_FONT_FILE + FONT_BITMAP_EXTENSION), true);
		robotoRegular.setFilter(Texture.TextureFilter.MipMapLinearNearest, Texture.TextureFilter.MipMapLinearNearest);
		robotoRegularFont = new DistanceFieldFont(Gdx.files.internal(ROBOTO_REGULAR_FONT_FILE + FONT_FILE_EXTENSION), new TextureRegion(robotoRegular));
		robotoRegularFont.setUseIntegerPositions(false);
		robotoRegularFont.setDistanceFieldSmoothing(ApplicationFontManager.DISTANCE_FIELD_FONT_SIZE);
		robotoRegularFont.getData().markupEnabled = false;

		Texture robotoBold = new Texture(Gdx.files.internal(ROBOTO_BOLD_FONT_FILE + FONT_BITMAP_EXTENSION), true);
		robotoBold.setFilter(Texture.TextureFilter.MipMapLinearNearest, Texture.TextureFilter.MipMapLinearNearest);
		robotoBoldFont = new DistanceFieldFont(Gdx.files.internal(ROBOTO_BOLD_FONT_FILE + FONT_FILE_EXTENSION), new TextureRegion(robotoBold));
		robotoBoldFont.setUseIntegerPositions(false);
		robotoBoldFont.setDistanceFieldSmoothing(ApplicationFontManager.DISTANCE_FIELD_FONT_SIZE);
		robotoBoldFont.getData().markupEnabled = false;

		distanceFieldShader = DistanceFieldFont.createDistanceFieldShader();
//		pedantic flag indicating whether attributes & uniforms must be present at all times
//		distanceFieldShader.pedantic = false;
	}


//////////////////////////// IApplicationLoader

	@Override
	public boolean update() {
		return true;
	}

	@Override
	public synchronized void dispose() {
		instance = null;
		robotoRegularFont.dispose();
		robotoBoldFont.dispose();
		distanceFieldShader.dispose();
	}
}