package cz.nic.tablexia.loader.zip;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.Net;
import com.badlogic.gdx.files.FileHandle;
import com.badlogic.gdx.scenes.scene2d.InputEvent;
import com.badlogic.gdx.scenes.scene2d.Stage;
import com.badlogic.gdx.scenes.scene2d.utils.ClickListener;
import com.badlogic.gdx.utils.async.AsyncTask;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.security.NoSuchAlgorithmException;
import java.util.Locale;
import java.util.Map;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import cz.nic.tablexia.Tablexia;
import cz.nic.tablexia.TablexiaBuildConfig;
import cz.nic.tablexia.bus.ApplicationBus;
import cz.nic.tablexia.checksum.Checksum;
import cz.nic.tablexia.loader.IApplicationLoader;
import cz.nic.tablexia.loader.TablexiaAbstractFileManager;
import cz.nic.tablexia.loader.TablexiaDataManager;
import cz.nic.tablexia.loader.application.ApplicationTextManager;
import cz.nic.tablexia.screen.AbstractTablexiaScreen;
import cz.nic.tablexia.util.Log;
import cz.nic.tablexia.util.ui.dialog.TablexiaComponentDialogFactory;

/**
 * Zip assets content loader
 *
 * @author Matyáš Latner
 *
 */
public class ZipAssetLoader extends TablexiaDataManager<Void> implements IApplicationLoader {

    public static final TablexiaAbstractFileManager.StorageType     ZIP_FILES_STORAGE_TYPE              = TablexiaAbstractFileManager.DownloadStorageType.EXTERNAL;
    public static final String                                      ZIP_FILE_EXTENSION                  = ".zip";
    public static final String                                      CHECKSUM_FILE_EXTENSION             = ".checksum";
    private static final String                                     TABLEXIA_ASSETS_DOWNLOAD_BASE_PATH  = TablexiaBuildConfig.TABLEXIA_ASSETS_URL + "/";
    private static final int                                        TABLEXIA_ASSETS_DOWNLOAD_TIMEOUT    = 2500;
    private static final String                                     TABLEXIA_TRUST_KEYSTORE_NAME        = "tablexiaTrustKeystore";
    private static final String                                     TABLEXIA_TRUST_KEYSTORE_PASSWORD    = "tablexia";

    public static final int										    STATUS_CODE_SUCCESS                 = 200;

    private static final String                                     SNAPSHOT_PACKAGE_NAME               = "SNAPSHOT";
	public static final int                                         DOWNLOAD_TRY_COUNT                  = 3;
	private static final Object                                     LOCK                                = new Object();

    private static final Object                                     REQUEST_DOWNLOAD_LOCK               = new Object();


    private static boolean                                          downloadResult                      = false;
    private static boolean                                          downloadHasResult                   = false;
    private static Long                                             startTime                           = null;
    private static boolean                                          error                               = false;

    private static boolean                                          usingMobileData = false;

    private static boolean                                          requestShown = false;
    private static boolean                                          requestAccepted = false;
    private static boolean                                          downloadPaused = false;

    private Locale                                                  locale;
    private Map<String, String>                                     buildChecksum;
    private Stage                                                   stage;

    public ZipAssetLoader(Stage stage) {
        this.stage = stage;
    }

    public void load(Locale locale, Map<String, String> buildChecksum) {
        this.locale = locale;
        this.buildChecksum = buildChecksum;
        reset(true);
    }

    public void reset(boolean withSyncTask) {
        error = false;
        downloadResult = false;
        downloadHasResult = false;
        startTime = null;

        requestShown = false;
        requestAccepted = false;
        downloadPaused = false;

		usingMobileData = Tablexia.getConnectionManager().isUsingMobileData();

        if (withSyncTask) {
			setAsyncTask(new ZipAssetLoaderTask(this, locale, buildChecksum));
		}
    }

    @Override
    public boolean update() {
        //show download request dialog ?
        if (usingMobileData && !requestShown && downloadPaused) {
            showRequestDialog();
        }

		return super.update();
	}

    /**
     * Shows Download Request Dialog
     */
    private void showRequestDialog() {
        requestShown = true;
        Log.info(ZipAssetLoader.class, "Download request dialog has been shown");

        TablexiaComponentDialogFactory.getInstance().createYesNoDialog(
                ApplicationTextManager.getInstance().getText(ApplicationTextManager.ApplicationTextsAssets.ZIPASSETLOADER_DOWNLOAD_REQUEST),
                new ClickListener() {
                    @Override
                    public void clicked(InputEvent event, float x, float y) {
                        super.clicked(event, x, y);
                        acceptRequest();
                    }
                },
                new ClickListener() {
                    @Override
                    public void clicked(InputEvent event, float x, float y) {
                        super.clicked(event, x, y);
                        Gdx.app.exit();
                    }
                }).show(Tablexia.ERROR_DIALOG_WIDTH, Tablexia.ERROR_DIALOG_HEIGHT);
    }

    /**
     * Accepts the download request and notifies the downloading thread
     */
    public void acceptRequest() {
        synchronized (REQUEST_DOWNLOAD_LOCK) {
            requestAccepted = true;

            REQUEST_DOWNLOAD_LOCK.notify();
        }
    }

	/**
	 * Event that is emitted when loader starts to check assets files
	 */
	public static class AssetsCheckStartEvent implements ApplicationBus.ApplicationEvent {}

	/**
	 * Event that is emitted when loader starts extracting zip archive
	 */
	public static class AssetsExtractingEvent implements ApplicationBus.ApplicationEvent {}

    /**
     * Event representing status of the download
     */
    public static class FileDownloadEvent implements ApplicationBus.ApplicationEvent{
        public enum FileDownloadEventType {
            Started,
            Update,
            Finished,
            Failed
        }

        private FileDownloadEventType type;
        private String fileName;
        private float progress;

        private FileDownloadEvent(FileDownloadEventType type, String fileName, float progress) {
            this.type = type;
            this.fileName = fileName;
            this.progress = progress;
        }

        public FileDownloadEventType getType() {
            return type;
        }
        public float getProgress() {
            return progress;
        }
        public String getFileName() {
            return fileName;
        }
    }

    private static class ZipAssetLoaderTask implements AsyncTask<Void> {

        private Locale locale;
        private Map<String, String> buildChecksums;
		private ZipAssetLoader zipAssetLoader;

        public ZipAssetLoaderTask(ZipAssetLoader zipAssetLoader, Locale locale, Map<String, String> buildChecksums) {
			this.zipAssetLoader = zipAssetLoader;
			this.locale = locale;
            this.buildChecksums = buildChecksums;
        }

        public void deleteOldContent(String extractDestinationDirectory) {
            // delete modified or old content
            File eddFile = new File(extractDestinationDirectory);
            if (eddFile.exists()) {
                Log.info(ZipAssetLoader.class, "Found invalid assets content!");
                Log.info(ZipAssetLoader.class, String.format("Deleting content under: %s", extractDestinationDirectory));
                startTimer();
                deleteDirectory(eddFile);
                resultTimer("Delete delay");
            }
        }

        public void extractNewContent(FileHandle assetsPackageFileHandle, String extractDestinationDirectory) throws IOException {
			ApplicationBus.getInstance().post(new AssetsExtractingEvent()).asynchronously();

            //extract new content
            startTimer();
            Log.info(ZipAssetLoader.class, String.format("Extracting new content to: %s", extractDestinationDirectory));
            unzip(assetsPackageFileHandle, extractDestinationDirectory);
            resultTimer("Extract delay");
        }

		@Override
		public Void call() throws Exception {
            String language = locale.getLanguage();
            String extractDestinationDirectory = TablexiaAbstractFileManager.getFileStoragePath(TablexiaAbstractFileManager.AssetsStorageType.EXTERNAL);

            if (buildChecksums != null) {
                String buildChecksum = buildChecksums.get(language);
                if (buildChecksum != null) {

                    // check current content
                    Log.info(ZipAssetLoader.class, "Assets check started...");
					ApplicationBus.getInstance().post(new AssetsCheckStartEvent()).asynchronously();
                    startTimer();
                    boolean checksumResult = checkAssets((buildChecksum), extractDestinationDirectory);
                    resultTimer("Assets check delay");

                    if (checksumResult) {
                        Log.info(ZipAssetLoader.class, "Assets check OK!");
                        Log.info(ZipAssetLoader.class, "Continue to loading");
                        return null;
                    }
                    Log.info(ZipAssetLoader.class, "Assets check FAILED!");

                    boolean downloadResult = true;

                    String assetsPackageName = language + "_" + buildChecksum;
                    String zipAssetsPackageName = assetsPackageName + ZIP_FILE_EXTENSION;
                    FileHandle assetsPackageFileHandle = TablexiaAbstractFileManager.getFileStoragePathFileHandle(ZIP_FILES_STORAGE_TYPE, zipAssetsPackageName);

                    // check current package file and download it if it is necessary
                    prepareKeyStore();
                    if (!checkAndDownload(assetsPackageName)) {
                        Log.err(getClass(), String.format("Cannot download assets package: %s", zipAssetsPackageName));

                        Log.info(ZipAssetLoader.class, "Devel mode --> try to use SNAPSHOT assets package");
                        String snapshotAssetsPackageName = language + "_" + SNAPSHOT_PACKAGE_NAME;
                        String snapshotZipAssetsPackageName = snapshotAssetsPackageName + ZIP_FILE_EXTENSION;

                        if (checkAndDownload(snapshotAssetsPackageName)) {
                            assetsPackageFileHandle = TablexiaAbstractFileManager.getFileStoragePathFileHandle(ZIP_FILES_STORAGE_TYPE, snapshotZipAssetsPackageName);
                        } else {
                            error(getClass(), String.format("Cannot download assets package: %s", snapshotZipAssetsPackageName));
                            downloadResult = false;
                        }
                    }

                    if(downloadResult) {
                        deleteOldContent(extractDestinationDirectory);
                        extractNewContent(assetsPackageFileHandle, extractDestinationDirectory);
                    }
                } else {
                    error(getClass(), "BUILD CHECKSUM: no build checksum specified for language " + language);
                }
            } else {
                error(getClass(), "BUILD CHECKSUM: no build checksum specified");
            }

			return null;
		}


//////////////////////////// CHECK

        private boolean checkAssets(String buildChecksum, String extractDestinationDirectory) {
            File file = new File(extractDestinationDirectory);
            if (file.exists() && file.isDirectory()) {
                try {
                    String runtimeChecksum = Checksum.getMd5OfDir(file);
                    Log.debug(ZipAssetLoader.class, "Comparing assets checksums: [BUILD: " + buildChecksum + "] - [RUNTIME: " + runtimeChecksum + "]");
                    return runtimeChecksum.equals(buildChecksum);
                } catch (NoSuchAlgorithmException e) {
                    Log.err(getClass(), "Cannot get checksum for assets!", e);
                } catch (IOException e) {
                    Log.err(getClass(), "Cannot get checksum for assets!", e);
                }
            }
            return false;
        }

        private boolean checkAssetsPackage(FileHandle downloadedChecksum, FileHandle downloadedFile) {
            if (downloadedFile.exists() && !downloadedFile.isDirectory()) {
                try {
                    String fileChecksum = Checksum.getMd5OfFile(downloadedFile.file());
                    String controlChecksum = downloadedChecksum.readString();
                    Log.debug(ZipAssetLoader.class, "Comparing assets package checksum: [PACKAGE: " + fileChecksum + "] - [CONTROL: " + controlChecksum + "]");
                    return fileChecksum.equals(controlChecksum);
                } catch (NoSuchAlgorithmException e) {
                    Log.err(getClass(), "Cannot get checksum for assets package!", e);
                } catch (IOException e) {
                    Log.err(getClass(), "Cannot get checksum for assets package!", e);
                }
            }
            return false;
        }

//////////////////////////// DOWNLOAD

        private void prepareKeyStore() {
            FileHandle keystoreExternal = TablexiaAbstractFileManager.getFileStoragePathFileHandle(TablexiaAbstractFileManager.KeystoreStorageType.EXTERNAL, TABLEXIA_TRUST_KEYSTORE_NAME);
            FileHandle keystoreInternal = TablexiaAbstractFileManager.getFileStoragePathFileHandle(TablexiaAbstractFileManager.KeystoreStorageType.INTERNAL, TABLEXIA_TRUST_KEYSTORE_NAME);
            keystoreExternal.write(keystoreInternal.read(), false);

            System.setProperty("javax.net.ssl.trustStore", keystoreExternal.file().getAbsolutePath());
            System.setProperty("javax.net.ssl.trustStorePassword", TABLEXIA_TRUST_KEYSTORE_PASSWORD);
        }

        private boolean checkAndDownload(final String fileName) {
            int tryCounter = 0;
            String zipFileName = fileName + ZIP_FILE_EXTENSION;
            String checksumFileName = fileName + CHECKSUM_FILE_EXTENSION;

            // load checksum and package file
            FileHandle zipFileHandle = TablexiaAbstractFileManager.getFileStoragePathFileHandle(TablexiaAbstractFileManager.DownloadStorageType.EXTERNAL, zipFileName);
            FileHandle checksumFileHandle = TablexiaAbstractFileManager.getFileStoragePathFileHandle(TablexiaAbstractFileManager.DownloadStorageType.EXTERNAL, checksumFileName);

            while (true) {
                // check if checksum and package file exists
                if (checksumFileHandle.exists() && zipFileHandle.exists()) {
                    if (checkAssetsPackage(checksumFileHandle, zipFileHandle)) {
                        Log.info(ZipAssetLoader.class, String.format("Valid assets package file: %s", zipFileName));
                        return true;
                    } else {
                        Log.info(ZipAssetLoader.class, String.format("Malformed assets package file: %s", zipFileName));
                        removeUnnecessaryAssetFiles();
                    }
                } else {
                    Log.info(ZipAssetLoader.class, String.format("Checksum or assets package file not found!", zipFileName));
                    removeUnnecessaryAssetFiles();
                }

                /*
                 * If user is using mobile data
                 * Stop the thread and wait for the user to accept Request Download Dialog
                 * before continuing
                 */
                synchronized (REQUEST_DOWNLOAD_LOCK) {
                    try {
                        if(usingMobileData && !requestAccepted) {
                            downloadPaused = true;
                            REQUEST_DOWNLOAD_LOCK.wait();
                        }
                    } catch (InterruptedException e) {
                        e.printStackTrace();
                    }
                }

                if (tryCounter >= DOWNLOAD_TRY_COUNT) {
                    return false;
                }

				// download assets package checksum
				boolean checksumDownloadSuccess = false;
                boolean assetsDownloadSuccess   = false;

                Log.info(ZipAssetLoader.class, String.format("Downloading checksum file: %s try %d/%d", checksumFileName, tryCounter + 1, DOWNLOAD_TRY_COUNT));
				startTimer();
				checksumDownloadSuccess = downloadToOutputStream(fileName + CHECKSUM_FILE_EXTENSION, checksumFileHandle.write(false));
				resultTimer("Download delay of " + checksumFileName);

                if(checksumDownloadSuccess) {
                    Log.info(ZipAssetLoader.class, String.format("Downloading assets package file: %s try %d/%d", zipFileName, tryCounter + 1, DOWNLOAD_TRY_COUNT));
                    startTimer();
                    assetsDownloadSuccess = downloadToOutputStream(fileName + ZIP_FILE_EXTENSION, zipFileHandle.write(false));
                    resultTimer("Download delay of " + zipFileName);
                }

                if(!checksumDownloadSuccess)
                    deleteFile(checksumFileHandle.file());
                if(!assetsDownloadSuccess)
                    deleteFile(zipFileHandle.file());

				tryCounter++;
            }
        }

        private boolean downloadToOutputStream(final String fileName, final OutputStream os) {
            downloadHasResult = false;

            Net.HttpRequest request = new Net.HttpRequest(Net.HttpMethods.GET);
            request.setTimeOut(TABLEXIA_ASSETS_DOWNLOAD_TIMEOUT);
            request.setUrl(TABLEXIA_ASSETS_DOWNLOAD_BASE_PATH + fileName);

            Gdx.net.sendHttpRequest(request, new Net.HttpResponseListener() {

                @Override
                public void handleHttpResponse(Net.HttpResponse httpResponse) {
                    if(httpResponse.getStatus().getStatusCode() != STATUS_CODE_SUCCESS) {
                        Log.info(ZipAssetLoader.class, String.format("Download failed. Returned status code: %d", httpResponse.getStatus().getStatusCode()));
                        downloadResult = false;
                        notifyDownload();
                        return;
                    }

                    long length = Long.parseLong(httpResponse.getHeader("Content-Length"));

                    InputStream is = httpResponse.getResultAsStream();

                    byte[] bytes = new byte[1024];
                    int count = -1;
                    long read = 0;

                    ApplicationBus.getInstance().post(new FileDownloadEvent(FileDownloadEvent.FileDownloadEventType.Started, fileName, 0)).asynchronously();

                    try {
                        while ((count = is.read(bytes, 0, bytes.length)) != -1) {
                            os.write(bytes, 0, count);
                            read += count;

                            final double done = read / (double) length;
                            final int progress = ((int) ((done) * 100));

                            ApplicationBus.getInstance().post(new AbstractTablexiaScreen.ScreenInfoEvent("Downloading: ", fileName + " [ " + progress + "% ]")).asynchronously();
                            ApplicationBus.getInstance().post(new FileDownloadEvent(FileDownloadEvent.FileDownloadEventType.Update, fileName, (float) (done))).asynchronously();
                        }

						ApplicationBus.getInstance().post(new FileDownloadEvent(FileDownloadEvent.FileDownloadEventType.Finished, fileName, 1)).asynchronously();
                        downloadResult = true;
					} catch (IOException e) {
						Log.err(ZipAssetLoader.class, "Cannot download file: " + fileName + " !", e);
						downloadResult = false;
                    }
					notifyDownload();
                }

                @Override
                public void failed(Throwable t) {
                    Log.err(ZipAssetLoader.class, "Downloading of file: " + fileName + " Failed!", t);
                    ApplicationBus.getInstance().post(new FileDownloadEvent(FileDownloadEvent.FileDownloadEventType.Failed, fileName, 0)).asynchronously();
                    downloadResult = false;
                    notifyDownload();
                }

                @Override
                public void cancelled() {
                    Log.info(ZipAssetLoader.class, "Downloading of file: " + fileName + " Canceled!");
                    ApplicationBus.getInstance().post(new FileDownloadEvent(FileDownloadEvent.FileDownloadEventType.Failed, fileName, 0)).asynchronously();
                    downloadResult = false;
                    notifyDownload();
                }
            });

            // wait to download complete
            synchronized (LOCK) {
                if (!downloadHasResult) {
                    try {
                        LOCK.wait();
                    } catch (InterruptedException e) {
                        Log.err(ZipAssetLoader.class, "Cannot wait to download end!", e);
                    }
                }
            }

            return downloadResult;
        }

        private void notifyDownload() {
            synchronized (LOCK) {
                downloadHasResult = true;
                LOCK.notify();
            }
        }

        private void removeUnnecessaryAssetFiles() {
            File dir = new File(TablexiaAbstractFileManager.getFileStoragePath(TablexiaAbstractFileManager.DownloadStorageType.EXTERNAL));
            File [] files = dir.listFiles(new FilenameFilter() {
                @Override
                public boolean accept(File dir, String name) {
                    return name.startsWith(locale.getLanguage()) && !name.contains(SNAPSHOT_PACKAGE_NAME);
                }
            });

            if(files == null) return;

            for (File assetFile : files) {
                Log.info(ZipAssetLoader.class, "Removing unnecessary asset file: " + assetFile.getName());
                deleteFile(assetFile);
            }
        }


//////////////////////////// EXTRACT

        private static void unzip(FileHandle zipFile, String extractDestinationDirectory) throws IOException {
            ZipInputStream zipInputStream = new ZipInputStream(new BufferedInputStream(zipFile.read()));
            try {
                ZipEntry zipEntry;
                int count;
                byte[] buffer = new byte[8192];
                while ((zipEntry = zipInputStream.getNextEntry()) != null) {

                    if (zipEntry.isDirectory()) {
                        continue;
                    }

                    File file = new File(extractDestinationDirectory, zipEntry.getName());

                    File dir = zipEntry.isDirectory() ? file : file.getParentFile();
                    if (!dir.isDirectory() && !dir.mkdirs()) {
                        throw new FileNotFoundException("Failed to ensure directory: " + dir.getAbsolutePath());
                    }
                    if (zipEntry.isDirectory()) {
                        continue;
                    }
                    FileOutputStream fout = new FileOutputStream(file);
                    try {
                        while ((count = zipInputStream.read(buffer)) != -1) {
                            fout.write(buffer, 0, count);
                        }
                    } finally {
                        fout.close();
                    }
                }
            } finally {
                zipInputStream.close();
            }
        }

        private static boolean deleteDirectory(File directory) {
            if(directory.exists()) {
                File[] files = directory.listFiles();

                if(files != null) {
                    for(int i=0; i < files.length; i++) {
                        if(files[i].isDirectory()) {
                            deleteDirectory(files[i]);
                        } else {
                            deleteFile(files[i]);
                        }
                    }
                }
            }
            return deleteFile(directory);
        }
	}

    private static boolean deleteFile(File file) {
        // workaround for android devices filesystem lock
        final File renamedFile = new File(file.getPath() + System.currentTimeMillis());
        file.renameTo(renamedFile);
        return renamedFile.delete();
    }


//////////////////////////// TIMER

    private static void startTimer() {
        startTime = System.nanoTime();
    }

    private static void resultTimer(String message) {
        Log.debug(ZipAssetLoader.class, String.format("%s: %05f s", message, (System.nanoTime() - startTime) / 1000000000.0));
    }

//////////////////////////// ERROR

    private static void error(Class clazz, String message) {
        Log.err(clazz, message);
        ApplicationBus.getInstance().post(new AbstractTablexiaScreen.ScreenInfoEvent("Downloading: ", "ERROR")).asynchronously();
        error = true;
    }

    public String getError() {
        return error ? ApplicationTextManager.getInstance().getText(ApplicationTextManager.ApplicationTextsAssets.ZIPASSETLOADER_ERROR) : null;
    }
}