package cz.nic.tablexia.model.game;

import com.badlogic.gdx.utils.TimeUtils;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.atomic.AtomicInteger;

import cz.nic.tablexia.TablexiaSettings;
import cz.nic.tablexia.TablexiaStorage;
import cz.nic.tablexia.game.AbstractTablexiaGame;
import cz.nic.tablexia.game.GameDefinition;
import cz.nic.tablexia.game.common.TablexiaRandom;
import cz.nic.tablexia.game.difficulty.GameDifficulty;
import cz.nic.tablexia.model.UserDAO;
import cz.nic.tablexia.shared.model.Game;
import cz.nic.tablexia.shared.model.GamePause;
import cz.nic.tablexia.shared.model.GameScore;
import cz.nic.tablexia.shared.model.User;
import cz.nic.tablexia.util.Log;

/**
 * Created by Matyáš Latner
 */
public class GameDAO {

    private static final Object ATOMIC_LOCK = new Object();
    private static final AtomicInteger gameScoreSetCounter = new AtomicInteger(0);

//////////////////////////// Game API

    public static Game createGame(User user, GameDifficulty difficulty, GameDefinition gameDefinition, TablexiaRandom random) {
        Game game = null;
        Long id = insertNewGame(user, difficulty, gameDefinition, random);
        if (id != null) {
            game = new Game(id, user, difficulty.getDifficultyNumber(), gameDefinition.getGameNumber(), random.getSeed());
            game.setGamePauses(GamePauseDAO.selectGamePausesForGame(game));
        }
        return game;
    }

    public static Game getGameForId(long id) {
        return selectGameForId(id);
    }

	public static String getLastGameScoreValueForGameAndKey(int gameNumber, String key) {
		return selectLastGameScoreValue(TablexiaSettings.getInstance().getSelectedUser().getId(), gameNumber, key);
	}

    public static List<Game> getGamesForUserAndDefinition(Long userId, GameDefinition definition) {
        return selectGamesForUserAndDefinition(userId, definition);
    }

    public static void startGame(Game game) {
        game.setStartTime(gameUpdateStart(game.getId(), TimeUtils.millis()));
    }

    public static void startGame(Game game, long startTime) {
        game.setStartTime(gameUpdateStart(game.getId(), startTime));
    }

    public static void endGame(Game game) {
        game.setEndTime(gameUpdateEnd(game.getId(), TimeUtils.millis()));
    }

    public static Long pauseGame(Game game) {
        if (!game.isFinished()) {
            return GamePauseDAO.createGamePause(game);
        }
        return null;
    }

    public static Long resumeGame(Game game) {
        if (!game.isFinished()) {
            return GamePauseDAO.updateGamePauseEnd(game);
        }
        return null;
    }

	public static Long getAllPausesDuration(Game game) {
		long pausesDuration = 0;
		List<GamePause> gamePauses = GamePauseDAO.selectGamePausesForGame(game);

		for (GamePause gamePause : gamePauses) {
			if (gamePause.hasStartTime() && gamePause.hasEndTime()) {
				pausesDuration = pausesDuration + (gamePause.getEndTime() - gamePause.getStartTime());
			}
		}
		return pausesDuration;
	}

	public static Long getGameDuration(Game game) {
		if (game.getEndTime() == null || game.getGameStartTime() == null) {
			return null;
		}
		return game.getEndTime() - game.getGameStartTime() - getAllPausesDuration(game);
	}

    public static void setGameScore(final Game game, final String key, final String value) {
        gameScoreSetCounter.getAndIncrement();
        new Thread(new Runnable() {
            @Override
            public void run() {
                synchronized (ATOMIC_LOCK) {
                    final GameScore storedScore = game.getGameScore(key);
                    if (storedScore == null) {
                        if (GameScoreDAO.insertGameScore(game.getId(), key, value)) {
                            game.getGameScoreMap().add(new GameScore(key, value));
                        }
                    } else {
                        if (!storedScore.getValue().equals(value) && GameScoreDAO.updateGameScore(game.getId(), key, value)) {
                            storedScore.setValue(value);
                        }
                    }
                    gameScoreSetCounter.getAndDecrement();
                }
            }
        }).start();
    }

    /**
     * Is there any background thread setting up game score values ?
     * @return
     */
    public static boolean isGameScoreActual() {
        return gameScoreSetCounter.get() == 0;
    }

    public static AbstractTablexiaGame.GameResult getGameResult(Game game) {
        return GameDefinition.getGameDefinitionForGameNumber(game.getGameNumber()).getGameResultResolver().getGameCupsResult(game);
    }

    public static Integer getNumberOfCompleteGamesForGameDefinitionAndDifficulty(GameDefinition gameDefinition, GameDifficulty gameDifficulty, User user) {
        return selectCompleteGameCountForGameNumberAndDifficultyNumber(gameDefinition.getGameNumber(), gameDifficulty.getDifficultyNumber(), user.getId());
    }

    public static Integer getNumberOfIncompleteAndCompleteGamesForGameDefinition(GameDefinition gameDefinition, User user) {
        return selectIncompleteAndCompleteGameCountForGameNumber(gameDefinition.getGameNumber(), user.getId());
    }


//////////////////////////// DB ACCESS

    // prepared statements
    public static final String CREATE_TABLE = "CREATE TABLE IF NOT EXISTS game (id INTEGER PRIMARY KEY AUTOINCREMENT, user_id INTEGER NOT NULL, start_time INTEGER, end_time INTEGER, difficulty_number INTEGER NOT NULL, game_number INTEGER NOT NULL, random_seed INTEGER NOT NULL, sync_at INTEGER, FOREIGN KEY(user_id) REFERENCES user(id))";
    public static final String NEW_GAME_INSERT = "INSERT INTO game (user_id, difficulty_number, game_number, random_seed) VALUES (?, ?, ?, ?)";
    public static final String GAME_UPDATE_START = "UPDATE game SET start_time = ? WHERE id = ?";
    public static final String GAME_UPDATE_END = "UPDATE game SET end_time = ? WHERE id = ?";
    public static final String GAME_UPDATE_SYNC_AT = "UPDATE game SET sync_at = ? WHERE id = ?";
    public static final String GAME_SELECT_FOR_ID = "SELECT id, user_id, difficulty_number, game_number, random_seed, start_time, end_time FROM game WHERE id = ?";
    public static final String GAME_SELECT_LAST_ID = "SELECT max(id) FROM game";
    public static final String GAME_SELECT_COUNT_FOR_GAME_AND_DIFFICULTY = "SELECT count(id) FROM game WHERE game_number = ? AND difficulty_number = ? AND user_id = ? AND end_time IS NOT NULL";
    public static final String GAME_SELECT_FOR_USER_AND_DEFINITION = "SELECT id, user_id, difficulty_number, game_number, random_seed, start_time, end_time FROM game WHERE user_id = ? AND game_number = ? AND end_time IS NOT NULL AND end_time != 0 ORDER BY start_time ASC";
    public static final String GAME_SELECT_COUNT_FOR_GAME = "SELECT count(id) FROM game WHERE game_number = ? AND user_id = ?";
    public static final String GAME_SELECT_ALL_FOR_USER_SYNC = "SELECT id, user_id, difficulty_number, game_number, random_seed, start_time, end_time FROM game where user_id = ? AND sync_at is null";

    public static final String GAME_SELECT_BY_START_AND_END = "SELECT id FROM game WHERE user_id = ? AND start_time = ? AND end_time = ?";
    public static final String GAME_SELECT_BY_START_AND_NULL_END = "SELECT id FROM game WHERE user_id = ? AND start_time = ? AND end_time IS NULL";


    public static final String GAME_SELECT_LAST_SCORES_FOR_USER_AND_GAME = "SELECT value FROM game_score INNER JOIN game ON game_score.game_id=game.id AND game.user_id=? AND game.game_number=? WHERE game_score.key=? AND end_time IS NOT NULL ORDER BY game_id DESC LIMIT 1";
    public static final String GAME_SELECT_FOR_RANK_MANAGER = "SELECT id, difficulty_number, game_number, start_time, end_time FROM game WHERE user_id = ? AND start_time IS NOT NULL AND start_time != 0 AND end_time IS NOT NULL AND end_time != 0 ORDER BY end_time ASC";

    // classic statements
    public static final String IMPORT_GAME_INSERT = "INSERT INTO game (user_id, difficulty_number, game_number, random_seed, start_time, end_time, sync_at) VALUES (%d, %d, %d, %d, %d, %s, %d)";


    private static Long insertNewGame(User user, GameDifficulty difficulty, GameDefinition gameDefinition, TablexiaRandom random) {
        try {
            PreparedStatement insertStatement = TablexiaStorage.getInstance().prepareStatement(NEW_GAME_INSERT);
            insertStatement.setLong(1, user.getId());
            insertStatement.setInt(2, difficulty.getDifficultyNumber());
            insertStatement.setInt(3, gameDefinition.getGameNumber());
            insertStatement.setLong(4, random.getSeed());
            insertStatement.executeUpdate();
            insertStatement.close();
            // insertStatement.getGeneratedKeys().getLong(1) cannot be used -> getGeneratedKeys() is not implemented in SQLDroid
            return selectLastGameId();
        } catch (SQLException e) {
            Log.err(Game.class, "Cannot insert new game record into DB!", e);
        }
        return null;
    }

	private static String selectLastGameScoreValue(long userId, int gameNumber, String key) {
		String result = null;

		PreparedStatement statement = TablexiaStorage.getInstance().prepareStatement(GAME_SELECT_LAST_SCORES_FOR_USER_AND_GAME);
		try {
			//Set userId, gameNumber and key
			statement.setLong(1, userId);
			statement.setLong(2, gameNumber);
			statement.setString(3, key);
			ResultSet resultSet = statement.executeQuery();

			while(resultSet.next()) {
				result = resultSet.getString(1);
				resultSet.close();
			}
		} catch (SQLException e) {
			e.printStackTrace();
		}

		return result;
	}

    public static Long importGame(User user, int difficulty, int gameNumber, long randomSeed, long startTime, long endTime) {
        try {
            Long gameId = null;
            Statement st = TablexiaStorage.getInstance().createStatement();
            // workaround for sync error -> serialized JSON from server contains 0 instead of NULL
//            String sql = String.format(IMPORT_GAME_INSERT, user.getId(), difficulty, gameNumber, randomSeed, startTime, endTime, System.currentTimeMillis());
            String sql = String.format(IMPORT_GAME_INSERT, user.getId(), difficulty, gameNumber, randomSeed, startTime, endTime != 0 ? String.valueOf(endTime) : "NULL", System.currentTimeMillis());
            st.executeUpdate(sql);
            gameId = selectLastGameId();
            st.close();

            return gameId;
        } catch (SQLException ex) {
            Log.err(GameDAO.class, "Cannot import game record into DB!", ex);
        }

        return null;
    }

    public static Long selectGameByTimes(long userId, long startTime, long endTime) throws SQLException {
        // workaround for sync error -> serialized JSON from server contains 0 instead of NULL
//        PreparedStatement statement = TablexiaStorage.getInstance().prepareStatement(GAME_SELECT_BY_START_AND_END);
        PreparedStatement statement = TablexiaStorage.getInstance().prepareStatement(endTime != 0 ? GAME_SELECT_BY_START_AND_END : GAME_SELECT_BY_START_AND_NULL_END);
        statement.setLong(1, userId);
        statement.setLong(2, startTime);

        // workaround for sync error -> serialized JSON from server contains 0 instead of NULL
//        statement.setLong(3, endTime);
        if (endTime != 0) {
            statement.setLong(3, endTime);
        }

        Long gameId = null;
        ResultSet resultSet = statement.executeQuery();
        if (resultSet.next()) {
            gameId = resultSet.getLong("id");
        }
        resultSet.close();
        statement.close();

        return gameId;
    }

    private static Game selectGameForId(long id) {
        Game selectedGame = null;
        try {
            PreparedStatement statement = TablexiaStorage.getInstance().prepareStatement(GAME_SELECT_FOR_ID);
            statement.setLong(1, id);
            try {
                ResultSet resultSet = statement.executeQuery();
                if (resultSet.next()) {
                    String startTimeStr = resultSet.getString(6);
                    String endTimeStr = resultSet.getString(7);
                    selectedGame = new Game(resultSet.getLong(1),
                            UserDAO.selectUser(resultSet.getLong(2)),
                            resultSet.getInt(3),
                            resultSet.getInt(4),
                            resultSet.getLong(5),
                            startTimeStr != null ? Long.valueOf(startTimeStr) : null,
                            endTimeStr != null ? Long.valueOf(endTimeStr) : null);

                    selectedGame.setGamePauses(GamePauseDAO.selectGamePausesForGame(selectedGame));
                    selectedGame.getGameScoreMap().addAll(GameScoreDAO.selectGameScores(selectedGame.getId()));
                }
                resultSet.close();
            } catch (SQLException e) {
                Log.err(User.class, "Cannot select game with id: " + id, e);
            }
            statement.close();
        } catch (SQLException e) {
            Log.err(User.class, "Cannot select game with id: " + id, e);
        }
        return selectedGame;
    }

    public static List<Game> selectUsersGamesForSync(long userId) {
        List<Game> games = new ArrayList<Game>();
        try {
            PreparedStatement statement = TablexiaStorage.getInstance().prepareStatement(GAME_SELECT_ALL_FOR_USER_SYNC);
            statement.setLong(1, userId);
            try {
                ResultSet resultSet = statement.executeQuery();
                while (resultSet.next()) {
                    String startTimeStr = resultSet.getString(6);
                    String endTimeStr = resultSet.getString(7);
                    Game game = new Game(resultSet.getLong(1),
                            UserDAO.selectUser(resultSet.getLong(2)),
                            resultSet.getInt(3),
                            resultSet.getInt(4),
                            resultSet.getLong(5),
                            startTimeStr != null ? Long.valueOf(startTimeStr) : null,
                            endTimeStr != null ? Long.valueOf(endTimeStr) : null);

                    game.setGamePauses(GamePauseDAO.selectGamePausesForGame(game));
                    game.getGameScoreMap().addAll(GameScoreDAO.selectGameScores(game.getId()));
                    games.add(game);
                }
                resultSet.close();
            } catch (SQLException e) {
                Log.err(User.class, "Cannot select games for user_id: " + userId, e);
            }
            statement.close();
        } catch (SQLException e) {
            Log.err(User.class, "Cannot select games for user_id: " + userId, e);
        }
        return games;
    }

    public static List<Game> selectUserGamesForRankManager(long userId) {
        List<Game> games = new ArrayList<Game>();
        try {
            PreparedStatement statement = TablexiaStorage.getInstance().prepareStatement(GAME_SELECT_FOR_RANK_MANAGER);
            statement.setLong(1, userId);
            try {

                ResultSet resultSet = statement.executeQuery();
                while (resultSet.next()) {
                    long id  = resultSet.getLong(1);
                    int diff = resultSet.getInt(2);
                    int gameNum = resultSet.getInt(3);
                    String startTime = resultSet.getString(4);
                    String endTime   = resultSet.getString(5);

                    Long startTimeLong = startTime != null ? Long.valueOf(startTime) : null;
                    Long endTimeLong   = endTime   != null ? Long.valueOf(endTime)   : null;

                    if(startTime == null || endTime == null) continue; //only game that were finished

                    Game game = new Game(
                        id,
                        UserDAO.selectUser(userId),
                        diff,
                        gameNum,
                        0L,
                        startTimeLong,
                        endTimeLong
                    );

                    game.setGamePauses(GamePauseDAO.selectGamePausesForGame(game));
                    game.getGameScoreMap().addAll(GameScoreDAO.selectGameScores(game.getId()));
                    games.add(game);
                }
                resultSet.close();
            } catch (SQLException e) {
                Log.err(User.class, "Cannot select games for user_id: " + userId, e);
            }
            statement.close();
        } catch (SQLException e) {
            Log.err(User.class, "Cannot select games for user_id: " + userId, e);
        }
        return games;
    }

    private static List<Game> selectGamesForUserAndDefinition(long userId, GameDefinition definition) {

        List<Game> selectedGames = new ArrayList<Game>();

        try {
            PreparedStatement statement = TablexiaStorage.getInstance().prepareStatement(GAME_SELECT_FOR_USER_AND_DEFINITION);
            statement.setLong(1, userId);
            statement.setInt(2, definition.getGameNumber());

            try {
                ResultSet resultSet = statement.executeQuery();
                boolean hasNextGames = true;

                while (hasNextGames) {
                    if (resultSet.next()) {
                        Game game;

                        String startTimeStr = resultSet.getString(6);
                        String endTimeStr = resultSet.getString(7);
                        game = new Game(resultSet.getLong(1),
                                UserDAO.selectUser(resultSet.getLong(2)),
                                resultSet.getInt(3),
                                resultSet.getInt(4),
                                resultSet.getLong(5),
                                startTimeStr != null ? Long.valueOf(startTimeStr) : null,
                                endTimeStr != null ? Long.valueOf(endTimeStr) : null);

                        game.setGamePauses(GamePauseDAO.selectGamePausesForGame(game));
                        game.getGameScoreMap().addAll(GameScoreDAO.selectGameScores(game.getId()));
                        selectedGames.add(game);

                    } else {
                        hasNextGames = false;
                    }
                }

                resultSet.close();
            } catch (SQLException e) {
                Log.err(User.class, "Cannot select game with user id : " + userId, e);
            }

            statement.close();
        } catch (SQLException e) {
            Log.err(User.class, "Cannot select game with user id : " + userId, e);
        }

        return selectedGames;
    }

    private static Long selectLastGameId() {
        Long id = null;
        try {
            PreparedStatement statement = TablexiaStorage.getInstance().prepareStatement(GAME_SELECT_LAST_ID);
            try {
                ResultSet resultSet = statement.executeQuery();
                if (resultSet.next()) {
                    id = resultSet.getLong(1);
                }
                resultSet.close();
            } catch (SQLException e) {
                Log.err(User.class, "Cannot select game with id: " + id, e);
            }
            statement.close();
        } catch (SQLException e) {
            Log.err(User.class, "Cannot select game with id: " + id, e);
        }
        return id;
    }

    private static int selectCompleteGameCountForGameNumberAndDifficultyNumber(int gameNumber, int difficultyNumber, long userId) {
        Integer count = null;
        try {
            PreparedStatement statement = TablexiaStorage.getInstance().prepareStatement(GAME_SELECT_COUNT_FOR_GAME_AND_DIFFICULTY);
            try {
                statement.setInt(1, gameNumber);
                statement.setInt(2, difficultyNumber);
                statement.setLong(3, userId);
                ResultSet resultSet = statement.executeQuery();
                if (resultSet.next()) {
                    count = resultSet.getInt(1);
                }
                resultSet.close();
            } catch (SQLException e) {
                Log.err(User.class, "Cannot select games count for game definition number: " + gameNumber + " and difficulty number: " + difficultyNumber, e);
            }
            statement.close();
        } catch (SQLException e) {
            Log.err(User.class, "Cannot select games count for game definition number: " + gameNumber + " and difficulty number: " + difficultyNumber, e);
        }
        return count;
    }

    private static int selectIncompleteAndCompleteGameCountForGameNumber(int gameNumber, long userId) {
        Integer count = null;
        try {
            PreparedStatement statement = TablexiaStorage.getInstance().prepareStatement(GAME_SELECT_COUNT_FOR_GAME);
            try {
                statement.setInt(1, gameNumber);
                statement.setLong(2, userId);
                ResultSet resultSet = statement.executeQuery();
                if (resultSet.next()) {
                    count = resultSet.getInt(1);
                }
                resultSet.close();
            } catch (SQLException e) {
                Log.err(User.class, "Cannot select games count for game definition number: " + gameNumber, e);
            }
            statement.close();
        } catch (SQLException e) {
            Log.err(User.class, "Cannot select games count for game definition number: " + gameNumber, e);
        }
        return count;
    }

    public static void markGamesAsSync(List<Game> games) {
        if (games == null) {
            return;
        }

        PreparedStatement statement = TablexiaStorage.getInstance().prepareStatement(GAME_UPDATE_SYNC_AT);
        try {
            try {
                TablexiaStorage.getInstance().setAutoCommit(false);
                for (Game game : games) {
                    statement.setLong(1, System.currentTimeMillis());
                    statement.setLong(2, game.getId());
                    statement.addBatch();
                }
                statement.executeBatch();
                statement.close();
                TablexiaStorage.getInstance().commit();
            } catch (SQLException ex) {
                TablexiaStorage.getInstance().rollback();
                Log.err(GameDAO.class, "Can't mark games as synchronized: " + ex.getMessage(), ex);
            } finally {
                TablexiaStorage.getInstance().setAutoCommit(true);
            }

        } catch (SQLException ex) {
            Log.err(GameDAO.class, "Can't mark games as synchronized: " + ex.getMessage(), ex);
        }
    }

    private static Long gameUpdateStart(long id, long startTime) {
        try {
            PreparedStatement updateStatement = TablexiaStorage.getInstance().prepareStatement(GAME_UPDATE_START);
            updateStatement.setLong(1, startTime);
            updateStatement.setLong(2, id);
            float result = updateStatement.executeUpdate();
            updateStatement.close();
            if (result > 0) {
                return startTime;
            }
        } catch (SQLException e) {
            Log.err(Game.class, "Cannot update game start time in DB!", e);
        }
        return null;
    }

    private static Long gameUpdateEnd(long id, long endTime) {
        try {
            PreparedStatement updateStatement = TablexiaStorage.getInstance().prepareStatement(GAME_UPDATE_END);
            updateStatement.setLong(1, endTime);
            updateStatement.setLong(2, id);
            float result = updateStatement.executeUpdate();
            updateStatement.close();
            if (result > 0) {
                return endTime;
            }
        } catch (SQLException e) {
            Log.err(Game.class, "Cannot update game end time in DB!", e);
        }
        return null;
    }
}
