package cz.nic.tablexia.model.game;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import cz.nic.tablexia.TablexiaStorage;
import cz.nic.tablexia.game.AbstractTablexiaGame;
import cz.nic.tablexia.game.difficulty.GameDifficulty;
import cz.nic.tablexia.game.trophy.GameTrophyDefinition;
import cz.nic.tablexia.model.UserDAO;
import cz.nic.tablexia.shared.model.User;
import cz.nic.tablexia.shared.model.Game;
import cz.nic.tablexia.util.Log;

/**
 * Created by frantisek on 3.11.15.
 */
public class GameTrophy {

    public interface GameTrophyResolver {
        boolean hasGameTrophy(User user, GameTrophyDefinition trophyDef);
    }

    public static class NumberOfTimesFinished implements GameTrophyResolver {

        public static final String GAME_PLAYED_SELECT = "SELECT COUNT(*) FROM game WHERE user_id = ? AND game_number = ? AND difficulty_number != ? AND end_time IS NOT NULL GROUP BY user_id, game_number";

        @Override
        public boolean hasGameTrophy(User user, GameTrophyDefinition trophyDef) {
            int playedCount = 0;
            try {
                PreparedStatement statement = TablexiaStorage.getInstance().prepareStatement(GAME_PLAYED_SELECT);
                statement.setLong(1, user.getId());
                statement.setInt(2, trophyDef.getGameDefinition().getGameNumber());
                statement.setInt(3, GameDifficulty.TUTORIAL.getDifficultyNumber());
                try {
                    ResultSet resultSet = statement.executeQuery();
                    if (resultSet.next()) {
                        playedCount = resultSet.getInt(1);
                    }
                    resultSet.close();
                } catch (SQLException e) {
                    Log.err(GameTrophy.class, "Cannot select number of games from DB!", e);
                }
                statement.close();
            } catch (SQLException e) {
                Log.err(GameTrophy.class, "Cannot select number of games from DB!", e);
            }
            return playedCount >= trophyDef.getLimit();
        }
    }

    public static class ThreeStarsReceived implements GameTrophyResolver {

        public static final String GAME_SELECT_FOR_GAME_AND_DIFFICULTY = "SELECT id, user_id, difficulty_number, game_number, random_seed, start_time, end_time FROM game WHERE game_number = ? AND difficulty_number = ? AND user_id = ? AND end_time IS NOT NULL";

        private static List<Game> getGamesForGameAndDifficulty(User user, GameTrophyDefinition trophyDef) {
            List<Game> games = new ArrayList<Game>();
            try {
                PreparedStatement statement = TablexiaStorage.getInstance().prepareStatement(GAME_SELECT_FOR_GAME_AND_DIFFICULTY);
                statement.setInt(1, trophyDef.getGameDefinition().getGameNumber());
                statement.setInt(2, trophyDef.getLimit());
                statement.setLong(3, user.getId());
                try {
                    ResultSet resultSet = statement.executeQuery();
                    while (resultSet.next()) {
                        String startTimeStr = resultSet.getString(6);
                        String endTimeStr = resultSet.getString(7);
                        Game game = new Game(resultSet.getLong(1),
                                UserDAO.selectUser(resultSet.getLong(2)),
                                resultSet.getInt(3),
                                resultSet.getInt(4),
                                resultSet.getLong(5),
                                startTimeStr != null ? Long.valueOf(startTimeStr) : null,
                                endTimeStr != null ? Long.valueOf(endTimeStr) : null);
                        game.setGamePauses(GamePauseDAO.selectGamePausesForGame(game));
                        game.getGameScoreMap().addAll(GameScoreDAO.selectGameScores(game.getId()));
                        games.add(game);
                    }
                    resultSet.close();
                } catch (SQLException e) {
                    Log.err(GameTrophy.class, "Cannot select game with user_id: " + user.getId() + ", game_number: " + trophyDef.getGameDefinition().getGameNumber() + ", difficulty: " + trophyDef.getLimit(), e);
                }
                statement.close();
            } catch (SQLException e) {
                Log.err(GameTrophy.class, "Cannot select game with user_id: " + user.getId() + ", game_number: " + trophyDef.getGameDefinition().getGameNumber() + ", difficulty: " + trophyDef.getLimit(), e);
            }
            return games;
        }

        @Override
        public boolean hasGameTrophy(User user, GameTrophyDefinition trophyDef) {
            List<Game> games = getGamesForGameAndDifficulty(user, trophyDef);

            if (games.isEmpty()) {
                return false;
            }

            for (Game game : games) {
                if (GameDAO.getGameResult(game).getStarCount() == AbstractTablexiaGame.GameResult.THREE_STAR.getStarCount()) {
                    return true;
                }
            }
            return false;
        }
    }
}
