package cz.nic.tablexia.screen.gamemenu;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.audio.Music;
import com.badlogic.gdx.graphics.g2d.TextureRegion;
import com.badlogic.gdx.math.Vector2;
import com.badlogic.gdx.scenes.scene2d.Actor;
import com.badlogic.gdx.scenes.scene2d.Touchable;
import com.badlogic.gdx.scenes.scene2d.actions.Actions;
import com.badlogic.gdx.scenes.scene2d.actions.SequenceAction;
import com.badlogic.gdx.scenes.scene2d.ui.Image;

import net.engio.mbassy.listener.Handler;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import cz.nic.tablexia.bus.ApplicationBus;
import cz.nic.tablexia.game.GameDefinition;
import cz.nic.tablexia.menu.AbstractMenu;
import cz.nic.tablexia.menu.game.GameMenuDefinition;
import cz.nic.tablexia.menu.main.MainMenu;
import cz.nic.tablexia.model.screen.ScreenDAO;
import cz.nic.tablexia.screen.AbstractTablexiaScreen;
import cz.nic.tablexia.screen.gamemenu.actions.ScrollHomeAction;
import cz.nic.tablexia.screen.gamemenu.event.ShowGameMenuPageEvent;
import cz.nic.tablexia.screen.gamemenu.pages.GameMenuPage;
import cz.nic.tablexia.screen.gamemenu.pages.MenuPage;
import cz.nic.tablexia.screen.gamemenu.pages.OfficeMenuPage;
import cz.nic.tablexia.screen.gamemenu.pages.WallPage;
import cz.nic.tablexia.util.MusicUtil;
import cz.nic.tablexia.util.Utility;

/**
 * Created by lhoracek
 */
public class GameMenuScreen extends AbstractTablexiaScreen<int[][]> {

    public static final String SCREEN_STATE_PAGE_NUMBER = "pageNumber";
    public static final String INTRO_SPEECH             = MFX_PATH + "intro.mp3";
    public static final float  AMBIENT_SOUND_VOLUME     = 0.5f;

    private Music ambientMusic;

    private Music introSpeechMusic;
    private Music.OnCompletionListener introSpeechMusicCompleteListener;

    private SequenceAction initialScrollAction;
    private boolean initialScrollActive = false;

    public enum GameMenuPageDefinition {
        OFFICE_PAGE(0, null, null, null),
        ROBBERY_PAGE(        1, GameDefinition.ROBBERY,          GameMenuAssets.BASE_ASSETS_ROBBERY,         new Vector2(0.168f, 0.016f)),
        PURSUIT_PAGE(        2, GameDefinition.PURSUIT,          GameMenuAssets.BASE_ASSETS_PURSUIT,         new Vector2(0f, -0.12f)),
        KIDNAPPING_PAGE(     3, GameDefinition.KIDNAPPING,       GameMenuAssets.BASE_ASSETS_KIDNAPPING,      new Vector2(-0.080f, -0.219f)),
        NIGH_WATCH_PAGE(     4, GameDefinition.NIGHT_WATCH,      GameMenuAssets.BASE_ASSETS_NIGHT_WATCH,     new Vector2(-0.330f, -0.219f)),
        SHOOTING_RANGE_PAGE( 5, GameDefinition.SHOOTING_RANGE,   GameMenuAssets.BASE_ASSETS_SHOOTING_RANGE,  new Vector2(0.220f, -0.219f)),
        IN_THE_DARKNESS_PAGE(6, GameDefinition.IN_THE_DARKNESS,  GameMenuAssets.BASE_ASSETS_IN_THE_DARKNESS, new Vector2(-0.340f, 0.028f));

        private int             pageNumber;
        private GameDefinition  gameDefinition;

        //Offset for foregroundLayer in percents...
        private Vector2         foregroundOffset;

        //base path to assets without postfix(layer) like mid, back, title etc...
        private String          baseAssetsPath;

        GameMenuPageDefinition(int pageNumber, GameDefinition gameDefinition, String baseAssetsPath, Vector2 foregroundOffset) {
            this.pageNumber = pageNumber;
            this.gameDefinition = gameDefinition;
            this.baseAssetsPath = baseAssetsPath;
            this.foregroundOffset = foregroundOffset;
        }

        public String getBaseAssetsPath() {
            return baseAssetsPath;
        }

        public int getPageNumber() {
            return pageNumber;
        }

        public Vector2 getForegroundOffset() {
            return foregroundOffset;
        }

        public static String getResourcePath(GameDefinition gameDefinition, GameMenuAssets.AssetType assetType) {
            for(GameMenuPageDefinition game : GameMenuPageDefinition.values()) {
                if(gameDefinition.getGameNumber() == game.getPageNumber()) {
                       return assetType.getFolder() + "" + game.getBaseAssetsPath() + "_" + assetType.getPostFix();
                }
            }

            throw new IllegalArgumentException("Couldn't find game base assets for game definition: " + gameDefinition.name());
        }

        public static GameMenuPageDefinition getGameMenuPageForNumber(int gameMenuPageNumber) {
            for (GameMenuPageDefinition gameMenuPageDefinition : GameMenuPageDefinition.values()) {
                if (gameMenuPageDefinition.getPageNumber() == gameMenuPageNumber) {
                    return gameMenuPageDefinition;
                }
            }

            throw new IllegalArgumentException("Couldn't find game definition for page number: " + gameMenuPageNumber);
        }

        public static GameMenuPageDefinition getGameMenuPageForGameDefinition(GameDefinition gameDefinition) {
            for (GameMenuPageDefinition gameMenuPageDefinition : GameMenuPageDefinition.values()) {
                if (gameMenuPageDefinition.gameDefinition != null && gameMenuPageDefinition.gameDefinition == gameDefinition) {
                    return gameMenuPageDefinition;
                }
            }
            throw new IllegalArgumentException("Couldn't find game page number for game definition: " + gameDefinition.name());
        }
    }

    private PagedScrollPane pagedScrollPane;
    private Image vignette;

    @Override
    protected String prepareScreenTextResourcesAssetName() {
        return null;
    }

    @Override
    protected void screenLoaded(Map<String, String> screenState) {
        boolean isFirstLogIn = ScreenDAO.getScreenViewCount(getSelectedUser().getId(), getClass().getName()) == 0;

        // /game menu pages list, which will be later used for turning on sounds after first scroll
        final List<GameMenuPage> gameMenuPages = new ArrayList<GameMenuPage>();

        Image leftBorder = new Image(getScreenTextureRegion(GameMenuAssets.LEFT_OVERSCROLL));
        Image rightBorder = new Image(getScreenTextureRegion(GameMenuAssets.RIGHT_OVERSCROLL));

        pagedScrollPane = new PagedScrollPane(leftBorder, rightBorder);
        pagedScrollPane.setBounds(getSceneLeftX(), getSceneOuterBottomY(), getSceneWidth(), getSceneOuterHeight());

        vignette = new Image(getScreenPatch(GameMenuAssets.VIGNETTE));
        vignette.setTouchable(Touchable.disabled);

        prepareVignetteAndPagerSize();

        //Adds office
        OfficeMenuPage officeMenuPage = new OfficeMenuPage(this);
        officeMenuPage.setSize(getSceneWidth(), getSceneOuterHeight());
        pagedScrollPane.addPage(officeMenuPage);

        //Adds wall to separate office and street
        pagedScrollPane.addPage(new WallPage(this));

        //Adds other game menu pages
        for (GameMenuDefinition gd : GameMenuDefinition.values()) {
            GameMenuPage gameMenuPage =  new GameMenuPage(this, gd.getGameDefinition());
            if(isFirstLogIn) {
                gameMenuPages.add(gameMenuPage);
                gameMenuPage.setIsFirstLogin(isFirstLogIn);
            }
            gameMenuPage.setSize( getSceneWidth(), getSceneOuterHeight());
            pagedScrollPane.addPage(gameMenuPage);
        }

        pagedScrollPane.layout();
        getStage().addActor(pagedScrollPane);

        prepareVignetteAndPagerSize();
        getStage().addActor(vignette);

        ApplicationBus.getInstance().subscribe(this);

        if (screenState.containsKey(GameMenuScreen.SCREEN_STATE_PAGE_NUMBER)) {
            pagedScrollPane.scrollToPage(Integer.valueOf(screenState.get(GameMenuScreen.SCREEN_STATE_PAGE_NUMBER)), false);
        }

        // show intro scroll and play intro speech only first time user visits GameMenuScreen
        introSpeechMusic = null;
        if(isFirstLogIn) {
            introSpeechMusic = getMusic(INTRO_SPEECH);
            introSpeechMusic.play();

            pagedScrollPane.setTouchable(Touchable.disabled);
            pagedScrollPane.scrollToPage(pagedScrollPane.getChildren().size - 1, false);

            //TODO - Use scrollToPage
            pagedScrollPane.setScrollX(pagedScrollPane.getContent().getWidth());
            pagedScrollPane.updateVisualScroll();

            initialScrollAction = Actions.sequence(Actions.delay(1f), ScrollHomeAction.createInstace(), Actions.run(new Runnable() {
                @Override
                public void run() {
                    pagedScrollPane.setTouchable(Touchable.enabled);
                    initialScrollActive = false;
                }
            }));
            pagedScrollPane.addAction(initialScrollAction);
            initialScrollActive = true;
        }

        //sound effect of street is waiting, until intro speech is finished
        //game sound effects in menu are turned of after intro speech is finished
        if(isFirstLogIn) {
            if (introSpeechMusic != null) {
                introSpeechMusicCompleteListener = new Music.OnCompletionListener() {
                    @Override
                    public void onCompletion(Music music) {
                        music.dispose();
                        playAmbientMusic();
                        resetGameMenuSounds(gameMenuPages);
                    }
                };

                introSpeechMusic.setOnCompletionListener(introSpeechMusicCompleteListener);
            }
        } else {
            playAmbientMusic();
        }

        screenResized(0, 0);
    }

    private void cancelInitialScroll() {
        initialScrollActive = false;
        pagedScrollPane.removeAction(initialScrollAction);
        pagedScrollPane.cancel();
        pagedScrollPane.scrollToInitialPage();
        pagedScrollPane.setTouchable(Touchable.enabled);

        if(introSpeechMusic != null && introSpeechMusic.isPlaying()) {
            MusicUtil.fadeOut(introSpeechMusic, 1f, introSpeechMusicCompleteListener);
        }
    }

    private void playAmbientMusic() {
        ambientMusic = getMusic(GameMenuAssets.AMBIENT_MUSIC);
        ambientMusic.setVolume(AMBIENT_SOUND_VOLUME);
        ambientMusic.setLooping(true);
        ambientMusic.play();
    }

    private void resetGameMenuSounds(List<GameMenuPage> gameMenuPages) {
        for(GameMenuPage page : gameMenuPages) {
            page.setIsFirstLogin(false);
        }
    }

    public void prepareVignetteAndPagerSize() {
        vignette.setBounds(getViewportLeftX(), getViewportBottomY(), getViewportWidth(), getViewportHeight());
        setActorToFullScene(pagedScrollPane);
    }

    @Override
    protected void screenResized(int width, int height) {
        super.screenResized(width, height);
        prepareVignetteAndPagerSize();

        for(Actor a : pagedScrollPane.getContent().getChildren()) {
            MenuPage page = (MenuPage) a;
            page.onResize();
        }
    }

    @Handler
    public void handleShowGameMenuPageEvent(ShowGameMenuPageEvent showGameMenuPageEvent) {
        pagedScrollPane.scrollToPage(showGameMenuPageEvent.getGameMenuPageDefinition().getPageNumber(), showGameMenuPageEvent.isSmoothScroll());
    }

    @Override
    protected int[][] prepareScreenData(Map<String, String> screenState) {
        TextureRegion clickMap = getScreenTextureRegion(GameMenuAssets.OFFICE_CLICKMAP);
        if (!clickMap.getTexture().getTextureData().isPrepared()) {
            clickMap.getTexture().getTextureData().prepare();
        }
        return Utility.createColorMap(clickMap);
    }

    public static Map<String, String> createScreenState(GameMenuPageDefinition gameMenuPageDefinition) {
        HashMap<String, String> screenState = new HashMap<String, String>();
        screenState.put(GameMenuScreen.SCREEN_STATE_PAGE_NUMBER, String.valueOf(gameMenuPageDefinition.getPageNumber()));
        return screenState;
    }

    public PagedScrollPane getPagedScrollPane() {
        return pagedScrollPane;
    }

    @Override
    protected void screenDisposed() {
        if(pagedScrollPane != null) {
            for (Actor a : pagedScrollPane.getContent().getChildren()) {
                MenuPage page = (MenuPage) a;
                page.dispose();
            }
        }

        if(ambientMusic != null) {
            ambientMusic.stop();
            ambientMusic.dispose();
        }

        //Disposes sounds, which GameMenuPages use when changing difficulty
        getMusic(GameMenuAssets.STEP_HARD_MEDIUM).dispose();
        getMusic(GameMenuAssets.STEP_MEDIUM_HARD).dispose();
        getMusic(GameMenuAssets.STEP_EASY_MEDIUM).dispose();
        getMusic(GameMenuAssets.STEP_MEDIUM_EASY).dispose();

        super.screenDisposed();
    }

    @Override
    public void backButtonPressed() {
        if(initialScrollActive) {
            cancelInitialScroll();
            return;
        }

        if (!pagedScrollPane.getCurrentPage().backButtonPressed()) {
            if (!pagedScrollPane.scrollToInitialPage()) {
                Gdx.app.exit();
            }
        }
    }

    @Handler
    public void onMenuEvent(AbstractMenu.MenuEvent menuEvent) {
        if(initialScrollActive && menuEvent.getTargetMenu() == MainMenu.class) {
            cancelInitialScroll();
        }
    }
}