package cz.nic.tablexia.screen.gamemenu;

import com.badlogic.gdx.Input;
import com.badlogic.gdx.graphics.Color;
import com.badlogic.gdx.graphics.g2d.Batch;
import com.badlogic.gdx.math.MathUtils;
import com.badlogic.gdx.scenes.scene2d.Actor;
import com.badlogic.gdx.scenes.scene2d.InputEvent;
import com.badlogic.gdx.scenes.scene2d.InputListener;
import com.badlogic.gdx.scenes.scene2d.ui.HorizontalGroup;
import com.badlogic.gdx.scenes.scene2d.ui.Image;
import com.badlogic.gdx.utils.SnapshotArray;

import cz.nic.tablexia.screen.gamemenu.pages.MenuPage;
import cz.nic.tablexia.util.ui.ScrollPaneWithBorders;

/**
 * Represents Horizontal Paged Scroll Pane
 * Previous one (ViewPager) was a bit glitchy
 *
 * Created by Drahomir Karchnak on 01/12/15.
 */
public class PagedScrollPane extends ScrollPaneWithBorders {
	private static final int   BACKGROUND_SPACING_WORKAROUND = -1;

    //TODO - This is overscroll workaround... If you use int distance (like 50) it renders weird gap between menu pages and borders...
    private static final float OVERSCROLL_DISTANCE           = 49.5f;
    private static final float OVERSCROLL_MIN_SPEED          = 30f;
    private static final float OVERSCROLL_MAX_SPEED          = 200f;

    private static final int INITIAL_PAGE_NUMBER = GameMenuScreen.GameMenuPageDefinition.OFFICE_PAGE.getPageNumber();

	private HorizontalGroup content;

    private int currentPage;
    private int indexCurrentPage;

    private float distanceToChangePage = 120;
    private float lastScrollX = -1;

    private float lastTouchX  = -1;

	private int backgroundTotalWidth;

    public PagedScrollPane(Image leftBorder, Image rightBorder) {
        super(null, leftBorder, rightBorder);
        prepareScrollPane();
        prepareInput();

        setSmoothScrolling(true);
    }

    /**
     * Prepares ScrollPane and it's content
     */
    private void prepareScrollPane() {
        content = new HorizontalGroup();
		content.space(BACKGROUND_SPACING_WORKAROUND);
        super.setWidget(content);

        setFlingTime(0);
        setScrollingDisabled(false, true);
        setCancelTouchFocus(false);
        setupOverscroll(OVERSCROLL_DISTANCE, OVERSCROLL_MIN_SPEED, OVERSCROLL_MAX_SPEED);
    }

	@Override
	public void draw(Batch batch, float parentAlpha) {
		batch.setColor(Color.WHITE);

		batch.disableBlending();

		//Get width of the office page
		float officeWidth = content.getChildren().get(GameMenuScreen.GameMenuPageDefinition.OFFICE_PAGE.getPageNumber()).getWidth();

		//Width of all pages without the office page
		float pagesTotalWidth = content.getWidth() - officeWidth;

		float pagesScrollRange = pagesTotalWidth - getWidth();
		float visualScroll = getVisualScrollX() - officeWidth;

		//OverScrolling fix
		float overScroll = 0;
		if(visualScroll > pagesScrollRange)
			overScroll = visualScroll - pagesScrollRange;
		else if(visualScroll < 0)
			overScroll = visualScroll;

		visualScroll = MathUtils.clamp(visualScroll, 0, pagesScrollRange);

		float backgroundScrollRange = backgroundTotalWidth - getWidth();

		int x = 0;
		for (Actor a : content.getChildren()) {
			MenuPage page = (MenuPage) a;

			float backgroundWidth = page.getWidth()/2;

			if(page.getBackgroundTextureRegion() != null) {
				float posX = x - (visualScroll/pagesScrollRange) * backgroundScrollRange - overScroll;

				//Is page visible ?
				if(posX < getWidth() && posX + page.getWidth() > 0) {
					batch.draw(page.getBackgroundTextureRegion(),
							MathUtils.ceil(posX),
							getY() + getHeight() - page.getBackgroundTextureRegion().getRegionHeight(),
							backgroundWidth,
							page.getBackgroundTextureRegion().getRegionHeight());
				}
				x += backgroundWidth;
			}
		}

		super.draw(batch, parentAlpha);
	}

	/**
     * Prepares InputListener
     */
    private void prepareInput() {
        this.addListener(new InputListener() {
            @Override
            public boolean keyDown(InputEvent event, int keycode) {
                //TODO - Use TablexiaSettings for Key Bindings, and enable keydown events
                switch (keycode) {
                    case Input.Keys.LEFT:
                        scrollToPage(currentPage - 1, true);
                        break;

                    case Input.Keys.RIGHT:
                        scrollToPage(currentPage + 1, true);
                        break;
                }

                return super.keyDown(event, keycode);
            }

            @Override
            public boolean touchDown(InputEvent event, float x, float y, int pointer, int button) {
                lastTouchX = x;
                return true;
            }

            @Override
            public void touchUp(InputEvent event, float x, float y, int pointer, int button) {
                super.touchUp(event, x, y, pointer, button);
                onScroll(lastTouchX, x);
            }
        });
    }

    @Override
    public void setScrollX(float pixels) {
        super.setScrollX(pixels);

        SnapshotArray<Actor> pages = content.getChildren();
        int page = 0;
        for(int i = 0; i < pages.size; i++) {
            if(pages.get(i).getX() < getScrollX()) {
                page++;
            }
        }
        currentPage = page;
    }

    /**
     * Determinates if PagedScrollPane should change page or not
     * according to swipes X coordinate
     * @param touchDownX X coordinate of last touch down event
     * @param touchUpX X coordinate of last touch up event
     */
    private void onScroll(float touchDownX, float touchUpX) {
        setVelocityX(0);

        if(Math.abs(touchDownX - touchUpX) <= distanceToChangePage) {
            scrollToPage(currentPage, true);
            return;
        }
        else {
            if(touchDownX - touchUpX > 0) { //swiped right
                //scroll to next valid page
                scrollToPage(currentPage + 1, true);
            }
            else { //swiped left
                //scroll to previous valid page
                scrollToPage(currentPage - 1, true);
            }
        }
    }

    /**
     * Adds a page to PagedScrollPane's content
     * @param page
     */
    public void addPage(MenuPage page) {
		if(page != null) {
            content.addActor(page);

			backgroundTotalWidth = 0;

			for (Actor a : content.getChildren()) {
				MenuPage currPage = (MenuPage) a;

				if(currPage.getBackgroundTextureRegion() != null) {
					backgroundTotalWidth += currPage.getWidth()/2;
				}
			}
		}
    }

    /**
     * Updates PagedScrollPane
     * Had to override this method to add custom behaviour
     * @param delta
     */
    @Override
    public void act(float delta) {
        super.act(delta);

        //if scrollpane changed since the last frame
        if (lastScrollX != getVisualScrollX()) {
            if(content == null)
                return;

            float scrollX = getVisualScrollX();
            float pageX = 0;
            float pageWidth = 0;

            for (Actor a : content.getChildren()) {
                MenuPage menuPage = (MenuPage) a;

                pageX = menuPage.getX();
                pageWidth = menuPage.getWidth();

                //Calls children's onScoll method
                if (scrollX > pageX && scrollX < (pageX + pageWidth)) {
                    menuPage.onScroll(-(scrollX - pageX));
                }
                else if (scrollX < (pageX) && (scrollX + pageWidth) > pageX) {
                    menuPage.onScroll(pageX - scrollX);
                }
                if (pageX == scrollX) { //Scrolled to the page
                    menuPage.onScrolledToPage();
                    menuPage.onScroll(0);
                }
            }
        }
        lastScrollX = getVisualScrollX();
    }

    public boolean scrollToInitialPage() {
        if (currentPage == INITIAL_PAGE_NUMBER) {
            return false;
        } else {
            scrollToPage(INITIAL_PAGE_NUMBER, true);
            return true;
        }
    }

    /**
     * Tells PagedScrollPane to scroll to the page
     * page number isn't index in contents children list
     * but rather order from the left (includes only valid pages)
     * @param pageNumber
     * @param smooth
     */
    public void scrollToPage(int pageNumber, boolean smooth) {
        if(content == null) return;

        int validPagesCount = 0;
        //iterates through all children of contentTable
        for (int i = 0; i < content.getChildren().size; i++) {
            MenuPage page = (MenuPage) (content.getChildren().get(i));

            //tells if pagedScrollPane can snap to this child
            if (page.isSnapToAllowed()) {
                //have we reached the page we want to scroll to
                if (validPagesCount == pageNumber) {
                    setSmoothScrolling(smooth);
                    setScrollX(page.getX());

                    //Is changing page ?
                    if(indexCurrentPage != i) {
                        MenuPage leavePage = (MenuPage) content.getChildren().get(indexCurrentPage);
                        leavePage.onScrollPageExit();
                    }

                    //in case smooth scrolling is set to false, we need to update visual scroll
                    if (!smooth) {
                        updateVisualScroll();
                    }

                    //Saves index of current page, index doesnt include invalid pages
                    currentPage = validPagesCount;
                    indexCurrentPage = i;
                    break;
                }
                validPagesCount++;
            }
        }

        setSmoothScrolling(true);
    }

    /**
     * How much you have to scroll to change the page
     * Everyscroll above the distance value will be consider as a page changing scroll
     * @param distance
     */
    public void setDistanceToChangePage(float distance) {
        distanceToChangePage = Math.abs(distance);
    }

    public MenuPage getCurrentPage() {
        return (MenuPage) content.getChildren().get(currentPage);
    }

    public HorizontalGroup getContent() {
        return content;
    }
}
