package cz.nic.tablexia.screen.gamemenu.pages;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.audio.Music;
import com.badlogic.gdx.graphics.Color;
import com.badlogic.gdx.graphics.g2d.Batch;
import com.badlogic.gdx.graphics.g2d.TextureRegion;
import com.badlogic.gdx.math.MathUtils;
import com.badlogic.gdx.scenes.scene2d.Actor;
import com.badlogic.gdx.scenes.scene2d.InputEvent;
import com.badlogic.gdx.scenes.scene2d.InputListener;
import com.badlogic.gdx.scenes.scene2d.Touchable;
import com.badlogic.gdx.scenes.scene2d.actions.Actions;
import com.badlogic.gdx.scenes.scene2d.actions.SequenceAction;
import com.badlogic.gdx.scenes.scene2d.ui.Image;
import com.badlogic.gdx.scenes.scene2d.ui.Stack;
import com.badlogic.gdx.scenes.scene2d.utils.ClickListener;
import com.badlogic.gdx.utils.Align;

import net.engio.mbassy.listener.Handler;

import cz.nic.tablexia.Tablexia;
import cz.nic.tablexia.TablexiaApplication;
import cz.nic.tablexia.TablexiaSettings;
import cz.nic.tablexia.bus.ApplicationBus;
import cz.nic.tablexia.game.GameDefinition;
import cz.nic.tablexia.game.difficulty.GameDifficulty;
import cz.nic.tablexia.game.ranksystem.UserRankManager;
import cz.nic.tablexia.loader.application.ApplicationFontManager;
import cz.nic.tablexia.model.UserDifficultySettingsDAO;
import cz.nic.tablexia.screen.gamemenu.GameMenuAssets;
import cz.nic.tablexia.screen.gamemenu.GameMenuScreen;
import cz.nic.tablexia.screen.profile.ProfileScreen;
import cz.nic.tablexia.util.MusicUtil;
import cz.nic.tablexia.util.ui.ClickListenerWithSound;
import cz.nic.tablexia.util.ui.Switch;
import cz.nic.tablexia.util.ui.TablexiaLabel;
import cz.nic.tablexia.util.ui.TablexiaProgressBar;
import cz.nic.tablexia.util.ui.button.TablexiaButton;

/**
 * Represents a single page with game in Main Menu
 * Created by lhoracek on 3/4/15.
 * Edited by Drahomir Karchnak on 2/12/15.
 */
public class GameMenuPage extends MenuPage {
    /*
     *  SIZE OF ACTORS
     *  Size is represented as a percentage of the screens width
     *  For example 0.6 means that Actors width will be scaled to 60% of the screens width
     */
    private static final float TITLE_WIDTH              = 0.6f;
    private static final float START_BUTTON_WIDTH       = 0.15f;
    private static final float DIFFICULTY_THUMB_WIDTH   = 0.074f;
    private static final float DIFFICULTY_BAR_WIDTH     = 0.22f;
    /*
     * OFFSET OF ACTORS
     * Offset is represented as a percentage of the screens height
     * for example 0.05 means that actor will have offset of 5% of the screens height
     */
    private static final float TITLE_TOP_OFFSET             = 0.03f;
    private static final float START_BUTTON_BOTTOM_OFFSET   = 0.26f;
    private static final float DIFFICULTY_BOTTOM_OFFSET     = 0.1f;
    private static final float DIFFICULTY_BAR_BOTTOM_OFFSET = 0.04f;

	/*
	 * MIN and MAX Height to ensure gui elements are not
	 * too big or too small
	 */
    private static final float START_BUTTON_MIN_HEIGHT      = 0.33f;
    private static final float START_BUTTON_MAX_HEIGHT      = 0.36f;

	private static final float TITLE_MIN_HEIGHT		=	0.25f;
	private static final float TITLE_MAX_HEIGHT		=	0.37f;

    /*
     *  offset to the center of edge buttons on difficulty bar
     *  multiply this value with difficulty bars width to get offset for current screens coord system
     *
     *  14 is distance to center of the edge button
     *  251 is width of difficulty bar texture
     */
    private static final float DIFFICULTY_BAR_EDGE_OFFSET   = 14/251f;

    private static final float PROGRESS_BAR_WIDTH           = 0.6f;
    private static final float PROGRESS_BAR_BOTTOM_OFFSET   = 0.002f;

    private static final char  PROGRESS_STATUS_SEPARATOR    = '/';
    private static final float PROGRESS_STATUS_OFFSET_X     = 0.01f;

    private static final float INTRO_SOUND_DELAY    =   1.0f; //Delay in seconds
    
    private static final int DRAGGING_THRESHOLD = 20;

    //Font for difficulty labels
    private static final ApplicationFontManager.FontType_NEW DIFFICULTY_FONT_STYLE = ApplicationFontManager.FontType_NEW.BOLD_18;

    //Progress status font
    private static final ApplicationFontManager.FontType_NEW PROGRESS_STATUS_FONT  = ApplicationFontManager.FontType_NEW.BOLD_18;

    private GameDefinition game;
    private GameMenuScreen.GameMenuPageDefinition gameMenuPageDefinition;

    private float diffEasyX, diffMediumX, diffHardX;
    private Actor buttonDifficultyEasy, buttonDifficultyMedium, buttonDifficultyHard;

    private Stack diffStack;
    private Switch.DragSwitchListener dragSwitchListener;

    //hack to keep paralax layers out of picture before scrolled for first time
    private float scrollXOffset = TablexiaSettings.getWorldSize();

    //PagedScrollPane, which is parent of this menu page has Y offset, which we have to counter
    //to position all elements correctly
    private float pagedScrollPaneOffsetY;

    private Music introSound;

    //Enabled playing sound steps on difficulty change
    private boolean stepSoundsEnabled = false;

    //Step sounds assets
    private Music step_easy_medium, step_medium_easy;
    private Music step_medium_hard, step_hard_medium;

    private int lastDragDifficultyStep = 0;

    //Is this page the currently active page ?
    private boolean isPageActive = false;

    //Is this first display of this page (and should we play sound effects?)
    private boolean isFirstLogin = false;

    private TablexiaProgressBar gameMenuPageProgress;

    private TablexiaLabel progressLabel;

	private TextureRegion background;

    private TextureRegion midLayer;
    private TextureRegion foregroundLayer;

    boolean updateGameDifficultyInDB = true;

    public GameMenuPage(GameMenuScreen screen, GameDefinition game) {
        super(screen);
        this.game = game;
        this.gameMenuPageDefinition = GameMenuScreen.GameMenuPageDefinition.getGameMenuPageForGameDefinition(game);

        //offset is needed to properly calculate position of elements
        pagedScrollPaneOffsetY = screen.getPagedScrollPane().getY();

        prepareSounds();
		prepareBackground();
        prepareMidLayer();
        prepareForegroundLayer();
		prepareTitle();
        prepareStartButton();
        prepareDifficulty();
        prepareGameProgress();

		// init event bus handlers
		ApplicationBus.getInstance().subscribe(this);
	}

	private void prepareBackground() {
		background = getScreen().getScreenTextureRegion(   GameMenuScreen.GameMenuPageDefinition.getResourcePath(game, GameMenuAssets.AssetType.BACK));
	}

    private void prepareMidLayer() {
        this.midLayer = getScreen().getScreenTextureRegion(
            GameMenuScreen.GameMenuPageDefinition.getResourcePath(game, GameMenuAssets.AssetType.MID)
        );
    }

    private void prepareForegroundLayer() {
        this.foregroundLayer = getScreen().getScreenTextureRegion(
				GameMenuScreen.GameMenuPageDefinition.getResourcePath(game, GameMenuAssets.AssetType.FORE)
		);
    }

	/**
     *  Prepares/Loads Sounds
     */
    private void prepareSounds() {
        step_easy_medium = getScreen().getMusic(GameMenuAssets.STEP_EASY_MEDIUM);
        step_medium_easy = getScreen().getMusic(GameMenuAssets.STEP_MEDIUM_EASY);

        step_medium_hard = getScreen().getMusic(GameMenuAssets.STEP_MEDIUM_HARD);
        step_hard_medium = getScreen().getMusic(GameMenuAssets.STEP_HARD_MEDIUM);
    }

    /**
     * Creates and repares Title
     */
    public void prepareTitle() {
        //Resources
        TextureRegion title = getScreen().getScreenTextureRegion(GameMenuScreen.GameMenuPageDefinition.getResourcePath(game, GameMenuAssets.AssetType.TITLE));
        final Image titleImage = new Image(title);

        //Size
        int titleWidth  = (int) (getScreen().getSceneWidth() * TITLE_WIDTH);
        int titleHeight = (int) (titleWidth * ((float) title.getRegionHeight() / (float) title.getRegionWidth()));

		//Makes sure Height is always between min and max height
		float tempHeight = MathUtils.clamp(titleHeight, TITLE_MIN_HEIGHT * getScreen().getSceneInnerHeight(), TITLE_MAX_HEIGHT* getScreen().getSceneInnerHeight());
		if(tempHeight != titleHeight) {
			titleHeight = (int) tempHeight;
			titleWidth = (int) (titleHeight * (title.getRegionWidth() / (float) title.getRegionHeight()));
		}

        //Position
        int titleX = (int) (getScreen().getSceneWidth()/2 - titleWidth/2);
        int titleY = (int) (getScreen().getSceneInnerTopY() - titleHeight - (getScreen().getSceneInnerHeight() * TITLE_TOP_OFFSET) - pagedScrollPaneOffsetY);

        titleImage.setPosition(titleX, titleY);
        titleImage.setSize(titleWidth, titleHeight);

        addActor(titleImage);
    }

    /**
     * Initializes StartButton Pressed and Released
     */
    public void prepareStartButton() {
        // Prepare Start Button UP/Released
        TextureRegion startPressed  = getScreen().getScreenTextureRegion(GameMenuScreen.GameMenuPageDefinition.getResourcePath(game, GameMenuAssets.AssetType.STARTBUTTON));
        TextureRegion startReleased = getScreen().getScreenTextureRegion(GameMenuScreen.GameMenuPageDefinition.getResourcePath(game, GameMenuAssets.AssetType.STARTBUTTON_PRESSED));

        TablexiaButton startButton = new TablexiaButton(null, false, startPressed, startReleased, null, null)
                .adaptiveSize(true);

		// Size
        int startWidth = (int) (getScreen().getSceneWidth() * START_BUTTON_WIDTH);
        int startHeight = (int) (startWidth * (startPressed.getRegionHeight() / (float) startPressed.getRegionWidth()));


        //Makes sure Height is always between min and max height
        float tempHeight = MathUtils.clamp(startHeight, START_BUTTON_MIN_HEIGHT * getScreen().getSceneInnerHeight(), START_BUTTON_MAX_HEIGHT* getScreen().getSceneInnerHeight());
        if(tempHeight != startHeight) {
            startHeight = (int) tempHeight;
            startWidth = (int) (startHeight * (startPressed.getRegionWidth() / (float) startPressed.getRegionHeight()));
        }

        // Position
        int startX = (int) getScreen().getSceneWidth() / 2 - startWidth / 2;
        int startY = (int) ( (getScreen().getSceneInnerHeight() * START_BUTTON_BOTTOM_OFFSET) - pagedScrollPaneOffsetY);

		startButton.setSize(startWidth, startHeight);
		startButton.setPosition(startX, startY);
        
        startButton.setInputListener(new InputListener(){
            
            private boolean dragged;
            private float startX;
            
            @Override
            public boolean touchDown(InputEvent event, float x, float y, int pointer, int button) {
                super.touchDown(event, x, y, pointer, button);
                dragged = false;
                startX = x;
                return true;
            }
            
            @Override
            public void touchUp(InputEvent event, float x, float y, int pointer, int button) {
                super.touchUp(event, x, y, pointer, button);
                if (!dragged){
                        ApplicationBus.getInstance().publishAsync(new Tablexia.ChangeScreenEvent(GameMenuPage.this.game.getScreenClass(),
                                TablexiaApplication.ScreenTransaction.FADE));
                }
            }

            @Override
            public void touchDragged(InputEvent event, float x, float y, int pointer) {
                super.touchDragged(event, x, y, pointer);
                if (Math.abs(startX-x)>DRAGGING_THRESHOLD){
                    dragged =true;
                }
            }
        });

        addActor(startButton);
    }

    /**
     * Creates and prepares Difficulty slider, Difficulty labels and Buttons
     */
    public void prepareDifficulty() {
        stepSoundsEnabled = false;

        final float difficultyY = getScreen().getSceneInnerHeight() * DIFFICULTY_BOTTOM_OFFSET - pagedScrollPaneOffsetY;

        // Difficulty bar
        TextureRegion diffBar = getScreen().getScreenTextureRegion(GameMenuAssets.DIFF_BAR);
        Image diffBarImage = new Image(diffBar);

        int diffBarWidth  = (int) (getScreen().getSceneWidth() * DIFFICULTY_BAR_WIDTH);
        int diffBarHeight = (int) (diffBarWidth * (diffBar.getRegionHeight() / (float) diffBar.getRegionWidth()));

        int diffBarX = (int) getScreen().getSceneWidth() / 2 - diffBarWidth / 2;
        int diffBarY = (int) ( difficultyY + (getScreen().getSceneWidth() * DIFFICULTY_BAR_BOTTOM_OFFSET) );

        diffBarImage.setPosition(diffBarX, diffBarY);
        diffBarImage.setSize(diffBarWidth, diffBarHeight);
        addActor(diffBarImage);

        //Difficulty Thumb
        final TextureRegion diff = getScreen().getScreenTextureRegion(GameMenuAssets.DIFF_THUMB_EASY);

        float diffWidth  = getScreen().getSceneWidth() * DIFFICULTY_THUMB_WIDTH;
        float diffHeight = (int) (diffWidth * (diff.getRegionHeight() / (float) diff.getRegionWidth()));

        //Centers thumb difficulty button to difficulty bar
        float diffY = diffBarY + (diffBarHeight/2)  - (diffHeight/2);

        float centerDiffX = diffBarX + (diffBarWidth / 2);
        diffMediumX = centerDiffX - (diffWidth / 2);
        diffEasyX = centerDiffX - (diffBarWidth / 2) - (diffWidth / 2) + (diffBarWidth * DIFFICULTY_BAR_EDGE_OFFSET); // + (diffWidth * 0.1f);
        diffHardX = centerDiffX + (diffBarWidth / 2) - (diffWidth / 2) - (diffBarWidth * DIFFICULTY_BAR_EDGE_OFFSET);// - (diffWidth * 0.1f);

        // Labels for diffuculty slider
        TablexiaLabel.TablexiaLabelStyle labelStyle = new TablexiaLabel.TablexiaLabelStyle(DIFFICULTY_FONT_STYLE, Color.BLACK);

        TablexiaLabel easy = new TablexiaLabel(GameDifficulty.EASY.getTextDescription(), labelStyle);
        TablexiaLabel medium = new TablexiaLabel(GameDifficulty.MEDIUM.getTextDescription(), labelStyle);
        TablexiaLabel hard = new TablexiaLabel(GameDifficulty.HARD.getTextDescription(), labelStyle);

        easy.setAlignment(Align.center);
        medium.setAlignment(Align.center);
        hard.setAlignment(Align.center);

        float labelY = difficultyY;
        float labelAddX = diffWidth / 2;

        easy.setPosition(diffEasyX + labelAddX - easy.getWidth() / 2, labelY);
        medium.setPosition(diffMediumX + labelAddX - medium.getWidth() / 2, labelY);
        hard.setPosition(diffHardX + labelAddX - hard.getWidth() / 2, labelY);

        addActor(easy);
        addActor(medium);
        addActor(hard);

        float buttonAddX = diffWidth/2.0f;

        //Invisible buttons for easier difficulty choosing
        buttonDifficultyEasy = new Actor();
        buttonDifficultyEasy.setSize(diffWidth, diffWidth);
        buttonDifficultyEasy.setPosition( diffEasyX + buttonAddX - buttonDifficultyEasy.getWidth() / 2, diffY);

        buttonDifficultyMedium = new Actor();
        buttonDifficultyMedium.setSize(diffWidth, diffWidth);
        buttonDifficultyMedium.setPosition( diffMediumX + buttonAddX - buttonDifficultyMedium.getWidth() / 2, diffY);

        buttonDifficultyHard = new Actor();
        buttonDifficultyHard.setSize(diffWidth, diffWidth);
        buttonDifficultyHard.setPosition(diffHardX + buttonAddX - buttonDifficultyHard.getWidth() / 2, diffY);

        buttonDifficultyEasy.addListener(new ClickListener() {
            @Override
            public void clicked(InputEvent event, float x, float y) {
                super.clicked(event, x, y);
                updateDifficultySeekBar(GameDifficulty.EASY);
            }
        });

        buttonDifficultyMedium.addListener(new ClickListener() {
            @Override
            public void clicked(InputEvent event, float x, float y) {
                super.clicked(event, x, y);
                updateDifficultySeekBar(GameDifficulty.MEDIUM);
            }
        });

        buttonDifficultyHard.addListener(new ClickListener() {
            @Override
            public void clicked(InputEvent event, float x, float y) {
                super.clicked(event, x, y);
                updateDifficultySeekBar(GameDifficulty.HARD);
            }
        });

        addActor(buttonDifficultyEasy);
        addActor(buttonDifficultyMedium);
        addActor(buttonDifficultyHard);

        // Difficulty button
        diffStack = new Stack();

        // Creates and adds Difficulty thumbs
        diffStack.addActor(new Image(diff));
        diffStack.addActor(new Image(getScreen().getScreenTextureRegion(GameMenuAssets.DIFF_THUMB_MEDIUM)));
        diffStack.addActor(new Image(getScreen().getScreenTextureRegion(GameMenuAssets.DIFF_THUMB_HARD)));

        float diffX = diffEasyX;
        diffStack.setSize(diffWidth, diffHeight);
        diffStack.setPosition(diffX, diffY);
        addActor(diffStack);

        diffStack.addListener(dragSwitchListener = new Switch.DragSwitchListener(diffStack, diffEasyX, diffHardX, diffMediumX));
        dragSwitchListener.setSwitchMovedListener(new Switch.DragSwitchListener.SwitchMovedListener() {
            @Override
            public void movedToStep(int step) {
                for (int i = 0; i < diffStack.getChildren().size; i++) {
                    diffStack.getChildren().get(i).setVisible(i == step);
                    if(i == step) {
                        if(lastDragDifficultyStep != step) {
                            GameDifficulty difficulty = GameDifficulty.getGameDifficultyForDifficultyNumber(step + 1);
                            GameDifficulty previous = GameDifficulty.getGameDifficultyForDifficultyNumber(lastDragDifficultyStep + 1);

                            if (stepSoundsEnabled) {
                                playStepSound(previous, difficulty);
                            }

                            lastDragDifficultyStep = step;
                        }
                    }
                }
            }
        });
        dragSwitchListener.setSwitchSelectedListener(new Switch.DragSwitchListener.SwitchSelectedListener() {
            @Override
            public void stepSelected(int step) {
                if (updateGameDifficultyInDB) {
                    UserDifficultySettingsDAO.saveSettingsForUser(TablexiaSettings.getInstance().getSelectedUser().getId(), game.getGameNumber(), GameDifficulty.getVisibleGameDifficultyList().get(step).getDifficultyNumber());
                }
            }
        });
		updateGameDifficultyFromDB();

        stepSoundsEnabled = true;
    }

    private void updateGameDifficultyFromDB() {
        updateDifficultySeekBar(UserDifficultySettingsDAO.getUserSettingsByGame(TablexiaSettings.getInstance().getSelectedUser().getId(), game.getGameNumber()), false);
    }

    private void updateDifficultySeekBar(GameDifficulty gameDifficulty) {
        updateDifficultySeekBar(gameDifficulty, true);
    }

    private void updateDifficultySeekBar(GameDifficulty gameDifficulty, boolean updateGameDifficultyInDB) {
        this.updateGameDifficultyInDB = updateGameDifficultyInDB;
        dragSwitchListener.switchToStep(gameDifficulty.getDifficultyNumber() - 1);
        this.updateGameDifficultyInDB = true;
    }

    public void prepareGameProgress() {
        this.gameMenuPageProgress = new TablexiaProgressBar(
                getScreen().getScreenTextureRegion(GameMenuScreen.GameMenuPageDefinition.getResourcePath(game, GameMenuAssets.AssetType.PROGRESSBAR)),
                getScreen().getScreenTextureRegion(GameMenuAssets.PROGRESSBAR_FILL)
        );

        this.gameMenuPageProgress.setWidth(getScreen().getSceneWidth() * PROGRESS_BAR_WIDTH);
        this.gameMenuPageProgress.setPosition(
                getScreen().getSceneWidth() / 2 - gameMenuPageProgress.getWidth() / 2,
                getScreen().getSceneInnerBottomY() + (getScreen().getSceneOuterHeight() * PROGRESS_BAR_BOTTOM_OFFSET) - pagedScrollPaneOffsetY
        );

        UserRankManager.RankProgress rankProgress = UserRankManager.getInstance().getGameRankProgress(TablexiaSettings.getInstance().getSelectedUser(), game);

        this.progressLabel = new TablexiaLabel(getProgressLabelText(rankProgress), new TablexiaLabel.TablexiaLabelStyle(PROGRESS_STATUS_FONT, Color.BLACK));
        this.progressLabel.setPosition(
                gameMenuPageProgress.getX() + gameMenuPageProgress.getWidth() + PROGRESS_STATUS_OFFSET_X * getScreen().getSceneWidth(),
                gameMenuPageProgress.getY() + gameMenuPageProgress.getHeight()/2 - progressLabel.getHeight()/2
        );

        this.gameMenuPageProgress.addListener(new ClickListenerWithSound() {
            @Override
            public void onClick(InputEvent event, float x, float y) {
                ApplicationBus.getInstance().publish(new Tablexia.ChangeScreenEvent(ProfileScreen.class, TablexiaApplication.ScreenTransaction.FADE));
                gameMenuPageProgress.setTouchable(Touchable.disabled); //Makes sure you wont click twice...
            }
        });

        updateGameProgress();

        addActor(gameMenuPageProgress);
        addActor(progressLabel);
    }

    public void updateGameProgress() {
        UserRankManager.RankProgress rankProgress = UserRankManager.getInstance().getGameRankProgress(TablexiaSettings.getInstance().getSelectedUser(), game);

        if(gameMenuPageProgress == null) return;
        this.gameMenuPageProgress.setPercent(rankProgress.getPercentDone());

        if(progressLabel == null) return;
        progressLabel.setText(getProgressLabelText(rankProgress));
    }

    private String getProgressLabelText(UserRankManager.RankProgress rankProgress) {
        return (rankProgress.getCurrXP() - rankProgress.getMinXP()) + Character.toString(PROGRESS_STATUS_SEPARATOR) + (rankProgress.getNextXP() - rankProgress.getMinXP());
    }

    private void playStepSound(GameDifficulty previous, GameDifficulty current) {
        switch (previous) {
            case EASY:
                step_easy_medium.play();
                break;
            case MEDIUM:
                if(current == GameDifficulty.EASY)
                    step_medium_easy.play();
                else
                    step_medium_hard.play();
                break;
            case HARD:
                step_hard_medium.play();
                break;
            case TUTORIAL:
                break;
        }
    }

    /**
     * Custom draw method
     * @param batch
     * @param parentAlpha
     */
    @Override
    public void draw(Batch batch, float parentAlpha) {
        // Paralax layers
        // Back layer travels half the speed of scroll
        // needed to make fade work
        batch.setColor(getColor());
        batch.getColor().a *= parentAlpha;

        //Background rendering has been moved to PagedScrollPane class

        batch.enableBlending();

        // Mid layer travels the same speed
        batch.draw(
            midLayer,
            getScreen().getSceneLeftX() + getX(),
            getScreen().getSceneOuterBottomY()  - pagedScrollPaneOffsetY,
            getScreen().getSceneWidth(),
            midLayer.getRegionHeight()
        );

        batch.enableBlending();

        // Foreground layer travels faster and could be wider than screen. Needs to be clipped
        batch.draw(
            //Foreground texture region
            foregroundLayer,
            //X Pos
            getScreen().getSceneLeftX() + getX() + (scrollXOffset / 2) //X
                + getScreen().getSceneWidth()/2 - foregroundLayer.getRegionWidth()/2 //Center
                + (gameMenuPageDefinition.getForegroundOffset().x * getScreen().getSceneWidth()), // + Offset Y
            //Y Pos
            getScreen().getSceneInnerBottomY() - pagedScrollPaneOffsetY //Y
                + (gameMenuPageDefinition.getForegroundOffset().y * getScreen().getSceneOuterHeight()), // + Offset Y

            foregroundLayer.getRegionWidth(), //Width
            foregroundLayer.getRegionHeight() //Height
        );

        super.draw(batch, parentAlpha);
    }

    @Override
    public void onScroll(float offset) {
        scrollXOffset = offset;
    }

    @Override
    public void onScrolledToPage() {
        ApplicationBus.getInstance().post(new GameMenuPageScrolledEvent(gameMenuPageDefinition)).asynchronously();

        //Was scrolled to this page from different page
        if(!isPageActive && !isFirstLogin) {
            //Loads music if needed
            introSound = getScreen().getMusic(GameMenuScreen.GameMenuPageDefinition.getResourcePath(game, GameMenuAssets.AssetType.INTRO));
            introSound.setVolume(1.0f);

            //Waits for INTRO_SOUND_DELAY seconds and plays the music
            addAction(new SequenceAction(Actions.delay(INTRO_SOUND_DELAY), Actions.run(new Runnable() {
                @Override
                public void run() {
                    if (isPageActive) {
                        introSound.play();
                    }
                }
            })));
        }

        if(!isFirstLogin) isPageActive = true;
    }

    @Override
    public void onScrollPageExit() {
        isPageActive = false;
        MusicUtil.fadeOut(introSound, 1);
    }

    @Override
    public void onResize() { }

    @Override
    public void dispose() {
        disposeSounds();
    }

    @Override
    public TextureRegion getBackgroundTextureRegion() {
        return background;
    }

    @Override
    public boolean isSnapToAllowed() {
        return true;
    }

    public static class GameMenuPageScrolledEvent implements ApplicationBus.ApplicationEvent {
        private GameMenuScreen.GameMenuPageDefinition gameMenuPageDefinition;

        GameMenuPageScrolledEvent(GameMenuScreen.GameMenuPageDefinition gameMenuPageDefinition) {
            this.gameMenuPageDefinition = gameMenuPageDefinition;
        }

        public GameMenuScreen.GameMenuPageDefinition getGameMenuPageDefinition() {
            return gameMenuPageDefinition;
        }
    }

    public void disposeSounds() {
        if(introSound != null)
            introSound.dispose();
    }

    public void setIsFirstLogin(boolean isFirstLogin) {
        this.isFirstLogin = isFirstLogin;
    }

	@Handler
	public void handleSelectedUserEvent(final TablexiaSettings.SelectedUserEvent selectedUserEvent) {
		Gdx.app.postRunnable(new Runnable() {
			@Override
			public void run() {
				if (selectedUserEvent.isUserSelected()) {
                    stepSoundsEnabled = false;
                    updateGameDifficultyFromDB();
                    stepSoundsEnabled = true;

                    updateGameProgress();
				}
			}
		});
	}
}
