package cz.nic.tablexia.screen.loader;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.audio.Music;
import com.badlogic.gdx.graphics.Color;
import com.badlogic.gdx.graphics.Texture;
import com.badlogic.gdx.graphics.g2d.NinePatch;
import com.badlogic.gdx.scenes.scene2d.actions.Actions;
import com.badlogic.gdx.scenes.scene2d.actions.ParallelAction;
import com.badlogic.gdx.scenes.scene2d.actions.SequenceAction;
import com.badlogic.gdx.scenes.scene2d.ui.Image;

import net.engio.mbassy.listener.Handler;

import java.util.Map;

import cz.nic.tablexia.Tablexia;
import cz.nic.tablexia.TablexiaSettings;
import cz.nic.tablexia.bus.ApplicationBus;
import cz.nic.tablexia.loader.TablexiaAbstractFileManager;
import cz.nic.tablexia.loader.TablexiaTextureManager;
import cz.nic.tablexia.loader.application.ApplicationAtlasManager;
import cz.nic.tablexia.loader.zip.ZipAssetLoader;
import cz.nic.tablexia.menu.AbstractMenu;
import cz.nic.tablexia.menu.user.UserMenu;
import cz.nic.tablexia.screen.AbstractTablexiaScreen;
import cz.nic.tablexia.util.Log;
import cz.nic.tablexia.util.MusicUtil;
import cz.nic.tablexia.util.ui.TablexiaNoBlendingImage;

public class LoaderScreen extends AbstractTablexiaScreen<Void> {
    //Tower Clock
    //How long does it take for the big hand to rotate by 360 degrees
    private static final int CLOCK_BIG_HAND_SPEED        = 8;
    private static final float CLOCK_HANDS_OFFSET_X      = 220;
    private static final float CLOCK_HANDS_OFFSET_Y      = 266;
    private static final int CLOCK_HANDS_WIDTH           = 10;
    private static final int CLOCK_HANDS_HEIGHT          = 40;
    private static final int CLOCK_HANDS_ORIGIN_OFFSET_Y = 5;

    //Launch Logo
    private static final float LAUNCH_LOGO_WIDTH            = 0.35f;
    private static final float LAUNCH_LOGO_Y_TOP_OFFSET		= 0.10f;

    //Loading Bar
    private static final Color LOADING_BAR_SHADOW_COLOR     = Color.BLACK;
    private static final float LOADING_BAR_WIDTH            = 0.52f;
    private static final float LOADING_BAR_OFFSET_Y         = 0.16f;
    private static final float LOADING_BAR_FADE_IN_TIME		= 0.5f;
    private static final float LOADING_BAR_FADE_OUT_TIME	= 0.5f;

	//Loading Status Text
	private static final float STATUS_TEXT_FADE_IN_TIME		= 0.5f;
    private static final float STATUS_TEXT_FADE_OUT_TIME	= 0.5f;
    private static final float LOADING_STATUS_TOP_OFFSET    = 2f;

	//Please Login Text
	private static final float LOGIN_TEXT_WIDTH 			= 0.43f;
	private static final float LOGIN_TEXT_FADE_IN_TIME		= 0.5f;
	private static final float LOGIN_TEXT_FADE_OUT_TIME		= 0.5f;
    private static final float LOGIN_MUSIC_FADE_OUT_TIME    = 1;

    public static final String LOGIN_SPEECH                 = MFX_PATH + "login.mp3";

    private TablexiaTextureManager textureManager;
    private String backgroundAsset = LoaderAssets.LOADER_BACKGROUND;

	private LoadingBar loadingBar;
    private LoadingStatusText loadingStatusText;

    private Image launchLogo;
	private Image loginText;

    private String currentLanguage;
	private boolean isDownloadingFile = false;

    public LoaderScreen() {
        this.textureManager = new TablexiaTextureManager(TablexiaAbstractFileManager.AssetsStorageType.INTERNAL);
    }

    @Override
    public boolean canBePaused() { return false; }

    @Override
    public boolean canResetState() {
        return false;
    }

    @Override
    protected String prepareScreenAtlasPath(String screenAssetsPath, String screenName) {
        // no atlas loading
        return null;
    }

    @Override
    protected String prepareScreenTextResourcesAssetName() {
        // no text loading
        return null;
    }

    @Override
    protected void screenLoaded(Map<String, String> screenState) {
        currentLanguage = TablexiaSettings.getInstance().getLocale().getLanguage();
		ApplicationBus.getInstance().subscribe(this);

        prepareTextures();

        prepareBackground();
		prepareLoginText();
        prepareClockHands();
        prepareLaunchLogo();
        prepareLoadingBar();
        prepareLoadingStatusText();
    }

    ////////////////// HANDLERS

    /**
     * Called when loading finished
     * @param applicationLoadingCompleteEvent
     */
    @Handler
    public void onLoadingCompleteEvent(Tablexia.ApplicationLoadingCompleteEvent applicationLoadingCompleteEvent) {
		//Hide loadingBar and statusBar
		hideLoadingBar(LOADING_BAR_FADE_OUT_TIME);
		hideStatusText(STATUS_TEXT_FADE_OUT_TIME);
	}

	@Handler
	public void onAssetsCheckStartEvent(ZipAssetLoader.AssetsCheckStartEvent assetsCheckStartEvent) {
		showLoadingBar(LOADING_BAR_FADE_IN_TIME);
		loadingStatusText.setStatusText(LoadingStatusText.StatusText.ASSETS_CHECK);
		showStatusText(STATUS_TEXT_FADE_IN_TIME);
	}

	@Handler
	public void onAssetsExtractingEvent(ZipAssetLoader.AssetsExtractingEvent assetsExtractingEvent) {
		loadingStatusText.changeText(LoadingStatusText.StatusText.EXTRACTING_ASSETS, 1);
		loadingBar.changeActivity(LoadingBar.LoadingBarActivity.IDLE);
	}

    /**
     * Called whenever there is anything new with the download
     * @param event
     */
    @Handler
    public void onFileDownloadEvent(ZipAssetLoader.FileDownloadEvent event) {
        switch (event.getType()) {
            case Started:
                //Don't active the loading bar when downloading checksum file
                if(event.getFileName().toLowerCase().contains("checksum"))
                    return;

                isDownloadingFile = true;

                loadingBar.changeActivity(LoadingBar.LoadingBarActivity.PROGRESS);

				//Shows loading bar if needed
                showLoadingBar(LOADING_BAR_FADE_IN_TIME);

                loadingStatusText.changeText(LoadingStatusText.StatusText.DOWNLOADING_ASSETS, 1);
                break;

            case Update:
                if(isDownloadingFile)
                    loadingBar.setCurrentPercentage(event.getProgress());
                break;

            case Finished:
				if(isDownloadingFile) {
                    isDownloadingFile = false;
                    loadingBar.changeActivity(LoadingBar.LoadingBarActivity.IDLE);
                }
                break;

            case Failed:
                Log.err(LoaderScreen.class, "Failed to download file: " + event.getFileName());
                break;
        }
    }

    @Handler
    public void onMenuEvent(AbstractMenu.MenuEvent menuEvent) {
        if(menuEvent.getTargetMenu() == UserMenu.class) {
            if(menuEvent.getMenuEventType() == AbstractMenu.MenuEvent.MenuEventType.Open) {
                showLoginText();
            }
            else {
                hideLoginText();
            }
        }
    }


    ////////////////// PREPARING

    private void prepareTextures () {
		backgroundAsset = TablexiaSettings.getInstance().isWinterMode() ? LoaderAssets.LOADER_BACKGROUND_WINTER : LoaderAssets.LOADER_BACKGROUND;
		textureManager.loadTexture(backgroundAsset);

        textureManager.loadTexture(LoaderAssets.LOADER_SMALL_HAND);
        textureManager.loadTexture(LoaderAssets.LOADER_BIG_HAND);

        //Localized textures
        textureManager.loadTexture(LoaderAssets.LOADER_LAUNCH_LOGO + "_" + currentLanguage + ".png");
		textureManager.loadTexture(LoaderAssets.LOADER_LOGIN_TEXT + "_" + currentLanguage + ".png");

        textureManager.loadTexture(LoaderAssets.LOADER_BAR_BACKGROUND);
        textureManager.loadTexture(LoaderAssets.LOADER_BAR_FOREGROUND);
        textureManager.finishLoading();
    }

	private void prepareLoginText() {
		Texture texture = textureManager.getTexture(LoaderAssets.LOADER_LOGIN_TEXT + "_" + currentLanguage + ".png");
		texture.setFilter(Texture.TextureFilter.Linear, Texture.TextureFilter.Linear);

		loginText = new Image(texture);
		loginText.setSize(getSceneWidth() * LOGIN_TEXT_WIDTH, (getSceneWidth() * LOGIN_TEXT_WIDTH) * (texture.getHeight() / (float) texture.getWidth()));

		loginText.setPosition(getSceneWidth() / 2 - loginText.getWidth() / 2,
                getSceneInnerBottomY() + getSceneInnerHeight() / 2 - loginText.getHeight() / 2
        );

		loginText.addAction(Actions.alpha(0));
		getStage().addActor(loginText);
    }

    /**
     * Waits for LOGIN_TEXT_SHOW_DELAY seconds
     * and then fades in the text and plays login speech
     */
    private void showLoginText() {
        loginText.addAction(
            new ParallelAction(
                    Actions.fadeIn(LOGIN_TEXT_FADE_IN_TIME),
                    Actions.run(new Runnable() {
                        @Override
                        public void run() {
                            Music music = getMusic(LOGIN_SPEECH);
                            if (music != null) music.play();
                        }
                    })
            ));
    }

    private void hideLoginText() {
        loginText.addAction(Actions.fadeOut(LOGIN_TEXT_FADE_OUT_TIME));

        Music music = getMusic(LOGIN_SPEECH);
        if (music != null && music.isPlaying())
            MusicUtil.fadeOut(music, LOGIN_MUSIC_FADE_OUT_TIME, true);
    }

    private void prepareBackground() {
        Texture texture = textureManager.getTexture(backgroundAsset);
        texture.setFilter(Texture.TextureFilter.Linear, Texture.TextureFilter.Linear);

        TablexiaNoBlendingImage background = new TablexiaNoBlendingImage(texture);
        setActorToFullScene(background);

        getStage().addActor(background);
    }

    private void prepareClockHands() {
        //Preparing small hand
        Texture smallHandTexture = textureManager.getTexture(LoaderAssets.LOADER_SMALL_HAND);
        smallHandTexture.setFilter(Texture.TextureFilter.Linear, Texture.TextureFilter.Linear);

        Image smallHand = new Image(smallHandTexture);

        smallHand.setPosition(CLOCK_HANDS_OFFSET_X + getSceneLeftX(), CLOCK_HANDS_OFFSET_Y + getSceneInnerBottomY());
        smallHand.setSize(CLOCK_HANDS_WIDTH, CLOCK_HANDS_HEIGHT);
        smallHand.setOrigin(smallHand.getWidth() / 2, CLOCK_HANDS_ORIGIN_OFFSET_Y);

        getStage().addActor(smallHand);


        //Preparing big hand
        Texture bigHandTexture = textureManager.getTexture(LoaderAssets.LOADER_BIG_HAND);
        bigHandTexture.setFilter(Texture.TextureFilter.Linear, Texture.TextureFilter.Linear);

        Image bigHand = new Image(bigHandTexture);

        bigHand.setPosition(CLOCK_HANDS_OFFSET_X + getSceneLeftX(), CLOCK_HANDS_OFFSET_Y + getSceneInnerBottomY());
        bigHand.setSize(CLOCK_HANDS_WIDTH, CLOCK_HANDS_HEIGHT);
        bigHand.setOrigin(bigHand.getWidth() / 2, CLOCK_HANDS_ORIGIN_OFFSET_Y);

        getStage().addActor(bigHand);

        //Make 'em both rotate
        smallHand.addAction(Actions.forever(Actions.rotateBy(-360, CLOCK_BIG_HAND_SPEED)));
        bigHand.addAction(Actions.forever(Actions.rotateBy(-360, CLOCK_BIG_HAND_SPEED / 12.0f)));
    }

    public void prepareLaunchLogo() {
        Texture launchLogoTexture = textureManager.getTexture(LoaderAssets.LOADER_LAUNCH_LOGO + "_" + currentLanguage + ".png");
        launchLogoTexture.setFilter(Texture.TextureFilter.Linear, Texture.TextureFilter.Linear);
        launchLogo = new Image(launchLogoTexture);

        int width  = (int) (getSceneWidth() * LAUNCH_LOGO_WIDTH);
        int height = (int) (width * (launchLogoTexture.getHeight() / (float)(launchLogoTexture.getWidth())));

        launchLogo.setSize(width, height);
        launchLogo.setPosition(	getSceneWidth() / 2 - width / 2,
								getSceneInnerTopY() - height - (getSceneInnerHeight() * LAUNCH_LOGO_Y_TOP_OFFSET));

        getStage().addActor(launchLogo);
    }

    @Override
    public void backButtonPressed() {
        Gdx.app.exit();
    }

    public void prepareLoadingBar() {
        //Skin for LoadingBar
        LoadingBar.LoadingBarSkin skin = new LoadingBar.LoadingBarSkin(
                textureManager.getTexture(LoaderAssets.LOADER_BAR_FOREGROUND),
                textureManager.getTexture(LoaderAssets.LOADER_BAR_BACKGROUND),
                LOADING_BAR_SHADOW_COLOR
        );

        ApplicationAtlasManager.getInstance().getColorTexture(Color.BLACK);

        //Creates loading bar
        this.loadingBar = new LoadingBar(0, skin);

        //Set Size
        loadingBar.setWidth(LOADING_BAR_WIDTH * getSceneWidth());
        loadingBar.setHeight(loadingBar.getWidth() * (skin.getForegroundImage().getHeight() / (float) (skin.getForegroundImage().getWidth())));

        //Set position
        loadingBar.setPosition(getSceneWidth() / 2 - loadingBar.getWidth() / 2, LOADING_BAR_OFFSET_Y * getSceneInnerHeight() + getSceneInnerBottomY());
        loadingBar.setVisible(false);
		getStage().addActor(loadingBar);
    }

    public void prepareLoadingStatusText() {
        loadingStatusText = new LoadingStatusText();
        loadingStatusText.setBounds(
            loadingBar.getX(),
            loadingBar.getY() - LOADING_STATUS_TOP_OFFSET,
            loadingBar.getWidth(),
            loadingBar.getHeight()
        );
        getStage().addActor(loadingStatusText);
    }

	private void showLoadingBar(float time) {
		if(!loadingBar.isVisible()) {
			loadingBar.addAction(Actions.alpha(0));
			loadingBar.setVisible(true);
			loadingBar.addAction(Actions.fadeIn(time));
		}
	}

	private void hideLoadingBar(float time) {
		if(loadingBar.isVisible()) {
			loadingBar.addAction(new SequenceAction(Actions.fadeOut(time), Actions.visible(false)));
		}
	}

	private void showStatusText(float time) {
        if(loadingStatusText.isVisible())
            return;

        loadingStatusText.setVisible(true);
        loadingStatusText.addAction(Actions.fadeIn(time));
    }

	private void hideStatusText(float time) {
        if(!loadingStatusText.isVisible())
            return;

        loadingStatusText.addAction(new SequenceAction(Actions.fadeOut(time), Actions.visible(false)));
    }

    @Override
    public boolean hasSoftBackButton() {
        return false;
    }

    @Override
    protected void screenDisposed() {
        super.screenDisposed();
        loadingBar.dispose();
    }
}