package cz.nic.tablexia.util.ui.dialog;

import com.badlogic.gdx.scenes.scene2d.InputListener;
import com.badlogic.gdx.scenes.scene2d.Stage;
import com.badlogic.gdx.utils.Disposable;
import com.badlogic.gdx.utils.Scaling;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import cz.nic.tablexia.TablexiaSettings;
import cz.nic.tablexia.loader.application.ApplicationAtlasManager;
import cz.nic.tablexia.util.Log;
import cz.nic.tablexia.util.ui.dialog.components.AdaptiveSizeDialogComponent;
import cz.nic.tablexia.util.ui.dialog.components.AlertOnShowDialogComponent;
import cz.nic.tablexia.util.ui.dialog.components.BackButtonHideComponent;
import cz.nic.tablexia.util.ui.dialog.components.CenterPositionDialogComponent;
import cz.nic.tablexia.util.ui.dialog.components.DimmerDialogComponent;
import cz.nic.tablexia.util.ui.dialog.components.FixedSpaceContentDialogComponent;
import cz.nic.tablexia.util.ui.dialog.components.ImageContentDialogComponent;
import cz.nic.tablexia.util.ui.dialog.components.PositiveNegativeButtonContentDialogComponent;
import cz.nic.tablexia.util.ui.dialog.components.ResizableSpaceContentDialogComponent;
import cz.nic.tablexia.util.ui.dialog.components.TablexiaDialogComponentAdapter;
import cz.nic.tablexia.util.ui.dialog.components.TextContentDialogComponent;
import cz.nic.tablexia.util.ui.dialog.components.TouchCloseDialogComponent;
import cz.nic.tablexia.util.ui.dialog.components.ViewportMaximumSizeComponent;

/**
 * Created by Matyáš Latner.
 */
public class TablexiaComponentDialogFactory implements Disposable {

    public static final int         WARNING_DIALOG_WIDTH        = 400;
    public static final int         WARNING_DIALOG_HEIGHT       = 300;
    private static final Float      WARNING_DIALOG_TOP_PAD      = 1f/10;
    private static final Float      WARNING_DIALOG_BOT_PAD      = 1f/40;
    private static final float      WARNING_DIALOG_WIDTH_RATIO  = 3f/5;
    private static final float      WARNING_DIALOG_HEIGHT_RATIO = 1f/5;
    private static final Scaling    WARNING_DIALOG_SCALING      = Scaling.fill;

    private static final TablexiaComponentDialog.TablexiaDialogType DEFAULT_DIALOG_TYPE = TablexiaComponentDialog.TablexiaDialogType.DIALOG_SQUARE_BORDERS;

    private static TablexiaComponentDialogFactory instance;

    private Stage stage;

    private TablexiaComponentDialogFactory(Stage stage) {
        this.stage = stage;
    }

    public static TablexiaComponentDialogFactory getInstance() {
        if (instance == null) {
            String exceptionMessage = TablexiaComponentDialogFactory.class.getSimpleName() + " is not initialized!";
            Log.err(TablexiaSettings.class, exceptionMessage);
            throw new IllegalStateException(exceptionMessage);
        }
        return instance;
    }

    public static void init(Stage stage) {
        if (instance != null) {
            String exceptionMessage = TablexiaComponentDialogFactory.class.getSimpleName() + " already initialized!";
            Log.err(TablexiaSettings.class, exceptionMessage);
            throw new IllegalStateException(exceptionMessage);
        }

        instance = new TablexiaComponentDialogFactory(stage);
    }

    public static Stage getDialogStage() {
        return getInstance().stage;
    }

    @Override
    public void dispose() {
        instance = null;
    }

    public TablexiaComponentDialog createWarningYesNoDialog(String text, InputListener yesInputListener, boolean hideOnBackButton) {
        return createWarningYesNoDialog(text, yesInputListener, null, hideOnBackButton);
    }

    public TablexiaComponentDialog createWarningYesNoDialog(String text, InputListener yesInputListener, InputListener noInputListener, boolean hideOnBackButton) {
        ArrayList<TablexiaDialogComponentAdapter> adapters = new ArrayList<TablexiaDialogComponentAdapter>();

        adapters.add(new FixedSpaceContentDialogComponent(WARNING_DIALOG_TOP_PAD));
        adapters.add(new ImageContentDialogComponent(ApplicationAtlasManager.getInstance().getTextureRegion(ApplicationAtlasManager.DIALOG_WARNING), WARNING_DIALOG_SCALING, WARNING_DIALOG_WIDTH_RATIO, WARNING_DIALOG_HEIGHT_RATIO));

		if (hideOnBackButton) {
			adapters.add(new BackButtonHideComponent());
		}

        return createYesNoDialog(adapters, text, yesInputListener, noInputListener);
    }

    public TablexiaComponentDialog createYesNoDialog(String text, InputListener yesInputListener, InputListener noInputListener) {
        return createYesNoDialog(null, text, yesInputListener, noInputListener);
    }

    public TablexiaComponentDialog createYesNoDialog(List<TablexiaDialogComponentAdapter> additionalComponents, String text, InputListener yesInputListener, InputListener noInputListener) {
        ArrayList<TablexiaDialogComponentAdapter> adapters = new ArrayList<TablexiaDialogComponentAdapter>();

        if (additionalComponents != null) adapters.addAll(additionalComponents);

        adapters.add(new ResizableSpaceContentDialogComponent());
        adapters.add(new TextContentDialogComponent(text));

        adapters.add(new ResizableSpaceContentDialogComponent());
        adapters.add(new PositiveNegativeButtonContentDialogComponent(yesInputListener,true, noInputListener, true));
        adapters.add(new FixedSpaceContentDialogComponent(WARNING_DIALOG_BOT_PAD));

        adapters.add(new AlertOnShowDialogComponent());
        adapters.add(new AdaptiveSizeDialogComponent());
        adapters.add(new CenterPositionDialogComponent());
        adapters.add(new DimmerDialogComponent());

        return createDialog(adapters.toArray(new TablexiaDialogComponentAdapter[]{}));
    }

    public TablexiaComponentDialog createStandardDialog(TablexiaDialogComponentAdapter... componentDefinitions) {

        List<TablexiaDialogComponentAdapter> standardComponents = new ArrayList<TablexiaDialogComponentAdapter>() {
            {
                add(new CenterPositionDialogComponent());
                add(new DimmerDialogComponent());
                add(new TouchCloseDialogComponent());
                add(new ViewportMaximumSizeComponent());
                add(new AdaptiveSizeDialogComponent());
            }
        };

        List<TablexiaDialogComponentAdapter> components = new ArrayList<TablexiaDialogComponentAdapter>(Arrays.asList(componentDefinitions));
        components.addAll(standardComponents);

        return createDialog(components.toArray(new TablexiaDialogComponentAdapter[] {}));
    }

    public TablexiaComponentDialog createDialog(TablexiaDialogComponentAdapter... componentDefinitions) {
        return createDialog(DEFAULT_DIALOG_TYPE, componentDefinitions);
    }

    public TablexiaComponentDialog createDialog(TablexiaComponentDialog.TablexiaDialogType dialogType, TablexiaDialogComponentAdapter ... componentDefinitions) {
        return createDialog(stage, dialogType, componentDefinitions);
    }

    public TablexiaComponentDialog createDialog(Stage stage, TablexiaComponentDialog.TablexiaDialogType dialogType, TablexiaDialogComponentAdapter ... componentDefinitions) {
        return new TablexiaComponentDialog(stage, dialogType, componentDefinitions);
    }
}
